Package: xml2
Title: Parse XML
Version: 1.3.8
Authors@R: c(
    person("Hadley", "Wickham", role = "aut"),
    person("Jim", "Hester", role = "aut"),
    person("Jeroen", "Ooms", email = "jeroenooms@gmail.com", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd")),
    person("R Foundation", role = "ctb",
           comment = "Copy of R-project homepage cached as example")
  )
Description: Bindings to 'libxml2' for working with XML data using a simple, 
    consistent interface based on 'XPath' expressions. Also supports XML schema
    validation; for 'XSLT' transformations see the 'xslt' package.
License: MIT + file LICENSE
URL: https://xml2.r-lib.org, https://r-lib.r-universe.dev/xml2
BugReports: https://github.com/r-lib/xml2/issues
Depends: R (>= 3.6.0)
Imports: cli, methods, rlang (>= 1.1.0)
Suggests: covr, curl, httr, knitr, magrittr, mockery, rmarkdown,
        testthat (>= 3.2.0), xslt
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Encoding: UTF-8
RoxygenNote: 7.2.3
SystemRequirements: libxml2: libxml2-dev (deb), libxml2-devel (rpm)
Collate: 'S4.R' 'as_list.R' 'xml_parse.R' 'as_xml_document.R'
        'classes.R' 'format.R' 'import-standalone-obj-type.R'
        'import-standalone-purrr.R' 'import-standalone-types-check.R'
        'init.R' 'nodeset_apply.R' 'paths.R' 'utils.R' 'xml2-package.R'
        'xml_attr.R' 'xml_children.R' 'xml_document.R' 'xml_find.R'
        'xml_missing.R' 'xml_modify.R' 'xml_name.R' 'xml_namespaces.R'
        'xml_node.R' 'xml_nodeset.R' 'xml_path.R' 'xml_schema.R'
        'xml_serialize.R' 'xml_structure.R' 'xml_text.R' 'xml_type.R'
        'xml_url.R' 'xml_write.R' 'zzz.R'
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2025-03-14 14:37:00 UTC; jeroen
Author: Hadley Wickham [aut],
  Jim Hester [aut],
  Jeroen Ooms [aut, cre],
  Posit Software, PBC [cph, fnd],
  R Foundation [ctb] (Copy of R-project homepage cached as example)
Maintainer: Jeroen Ooms <jeroenooms@gmail.com>
Repository: CRAN
Date/Publication: 2025-03-14 19:30:02 UTC
Built: R 4.4.0; amd64-portbld-midnightbsd4.0.2; 2026-02-20 04:47:33 UTC; unix

Package: xml2
Title: Parse XML
Version: 1.5.2
Authors@R: c(
    person("Hadley", "Wickham", role = "aut"),
    person("Jim", "Hester", role = "aut"),
    person("Jeroen", "Ooms", email = "jeroenooms@gmail.com", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd")),
    person("R Foundation", role = "ctb",
           comment = "Copy of R-project homepage cached as example")
  )
Description: Bindings to 'libxml2' for working with XML data using a simple, 
    consistent interface based on 'XPath' expressions. Also supports XML schema
    validation; for 'XSLT' transformations see the 'xslt' package.
License: MIT + file LICENSE
URL: https://xml2.r-lib.org, https://r-lib.r-universe.dev/xml2
BugReports: https://github.com/r-lib/xml2/issues
Depends: R (>= 3.6.0)
Imports: cli, methods, rlang (>= 1.1.0)
Suggests: covr, curl, httr, knitr, mockery, rmarkdown, testthat (>=
        3.2.0), xslt
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Encoding: UTF-8
RoxygenNote: 7.3.3
SystemRequirements: libxml2: libxml2-dev (deb), libxml2-devel (rpm)
Collate: 'S4.R' 'as_list.R' 'xml_parse.R' 'as_xml_document.R'
        'classes.R' 'format.R' 'import-standalone-obj-type.R'
        'import-standalone-purrr.R' 'import-standalone-types-check.R'
        'init.R' 'nodeset_apply.R' 'paths.R' 'utils.R' 'xml2-package.R'
        'xml_attr.R' 'xml_children.R' 'xml_document.R' 'xml_find.R'
        'xml_missing.R' 'xml_modify.R' 'xml_name.R' 'xml_namespaces.R'
        'xml_node.R' 'xml_nodeset.R' 'xml_path.R' 'xml_schema.R'
        'xml_serialize.R' 'xml_structure.R' 'xml_text.R' 'xml_type.R'
        'xml_url.R' 'xml_write.R' 'zzz.R'
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-01-17 13:03:05 UTC; jeroen
Author: Hadley Wickham [aut],
  Jim Hester [aut],
  Jeroen Ooms [aut, cre],
  Posit Software, PBC [cph, fnd],
  R Foundation [ctb] (Copy of R-project homepage cached as example)
Maintainer: Jeroen Ooms <jeroenooms@gmail.com>
Repository: CRAN
Date/Publication: 2026-01-17 16:10:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("[",xml_missing)
S3method("[",xml_nodeset)
S3method("[[",xml_missing)
S3method("xml_attr<-",xml_missing)
S3method("xml_attr<-",xml_node)
S3method("xml_attr<-",xml_nodeset)
S3method("xml_attrs<-",xml_missing)
S3method("xml_attrs<-",xml_node)
S3method("xml_attrs<-",xml_nodeset)
S3method("xml_name<-",xml_missing)
S3method("xml_name<-",xml_node)
S3method("xml_name<-",xml_nodeset)
S3method("xml_text<-",xml_missing)
S3method("xml_text<-",xml_node)
S3method("xml_text<-",xml_nodeset)
S3method(as.character,xml_document)
S3method(as.character,xml_missing)
S3method(as.character,xml_node)
S3method(as.character,xml_nodeset)
S3method(as_list,xml_document)
S3method(as_list,xml_missing)
S3method(as_list,xml_node)
S3method(as_list,xml_nodeset)
S3method(as_xml_document,character)
S3method(as_xml_document,connection)
S3method(as_xml_document,list)
S3method(as_xml_document,raw)
S3method(as_xml_document,response)
S3method(as_xml_document,xml_document)
S3method(as_xml_document,xml_node)
S3method(as_xml_document,xml_nodeset)
S3method(format,xml_node)
S3method(is.na,xml_missing)
S3method(is.na,xml_node)
S3method(is.na,xml_nodeset)
S3method(nodeset_apply,xml_document)
S3method(nodeset_apply,xml_missing)
S3method(nodeset_apply,xml_node)
S3method(nodeset_apply,xml_nodeset)
S3method(print,xml_document)
S3method(print,xml_missing)
S3method(print,xml_namespace)
S3method(print,xml_node)
S3method(print,xml_nodeset)
S3method(read_html,default)
S3method(read_html,response)
S3method(read_xml,character)
S3method(read_xml,connection)
S3method(read_xml,raw)
S3method(read_xml,response)
S3method(tree_structure,xml_missing)
S3method(tree_structure,xml_node)
S3method(tree_structure,xml_nodeset)
S3method(write_html,xml_document)
S3method(write_html,xml_missing)
S3method(write_html,xml_node)
S3method(write_html,xml_nodeset)
S3method(write_xml,xml_document)
S3method(write_xml,xml_missing)
S3method(write_xml,xml_node)
S3method(write_xml,xml_nodeset)
S3method(xml_add_child,xml_document)
S3method(xml_add_child,xml_missing)
S3method(xml_add_child,xml_node)
S3method(xml_add_child,xml_nodeset)
S3method(xml_add_parent,xml_missing)
S3method(xml_add_parent,xml_node)
S3method(xml_add_parent,xml_nodeset)
S3method(xml_add_sibling,xml_missing)
S3method(xml_add_sibling,xml_node)
S3method(xml_add_sibling,xml_nodeset)
S3method(xml_find_all,xml_missing)
S3method(xml_find_all,xml_node)
S3method(xml_find_all,xml_nodeset)
S3method(xml_find_chr,xml_missing)
S3method(xml_find_chr,xml_node)
S3method(xml_find_chr,xml_nodeset)
S3method(xml_find_first,xml_missing)
S3method(xml_find_first,xml_node)
S3method(xml_find_first,xml_nodeset)
S3method(xml_find_int,xml_missing)
S3method(xml_find_int,xml_node)
S3method(xml_find_int,xml_nodeset)
S3method(xml_find_lgl,xml_missing)
S3method(xml_find_lgl,xml_node)
S3method(xml_find_lgl,xml_nodeset)
S3method(xml_find_num,xml_missing)
S3method(xml_find_num,xml_node)
S3method(xml_find_num,xml_nodeset)
S3method(xml_ns,xml_document)
S3method(xml_ns,xml_missing)
S3method(xml_ns,xml_node)
S3method(xml_ns,xml_nodeset)
S3method(xml_parent,xml_missing)
S3method(xml_parent,xml_node)
S3method(xml_parent,xml_nodeset)
S3method(xml_remove,xml_missing)
S3method(xml_remove,xml_node)
S3method(xml_remove,xml_nodeset)
S3method(xml_replace,xml_missing)
S3method(xml_replace,xml_node)
S3method(xml_replace,xml_nodeset)
S3method(xml_serialize,xml_document)
S3method(xml_serialize,xml_node)
S3method(xml_serialize,xml_nodeset)
S3method(xml_set_attr,xml_missing)
S3method(xml_set_attr,xml_node)
S3method(xml_set_attr,xml_nodeset)
S3method(xml_set_attrs,xml_missing)
S3method(xml_set_attrs,xml_node)
S3method(xml_set_attrs,xml_nodeset)
S3method(xml_set_name,xml_missing)
S3method(xml_set_name,xml_node)
S3method(xml_set_name,xml_nodeset)
S3method(xml_url,xml_missing)
S3method(xml_url,xml_node)
S3method(xml_url,xml_nodeset)
S3method(xml_validate,xml_document)
export("xml_attr<-")
export("xml_attrs<-")
export("xml_name<-")
export("xml_text<-")
export(as_list)
export(as_xml_document)
export(download_html)
export(download_xml)
export(html_structure)
export(read_html)
export(read_xml)
export(url_absolute)
export(url_escape)
export(url_parse)
export(url_relative)
export(url_unescape)
export(write_html)
export(write_xml)
export(xml2_example)
export(xml_add_child)
export(xml_add_parent)
export(xml_add_sibling)
export(xml_attr)
export(xml_attrs)
export(xml_cdata)
export(xml_child)
export(xml_children)
export(xml_comment)
export(xml_contents)
export(xml_double)
export(xml_dtd)
export(xml_find_all)
export(xml_find_chr)
export(xml_find_first)
export(xml_find_int)
export(xml_find_lgl)
export(xml_find_num)
export(xml_find_one)
export(xml_has_attr)
export(xml_integer)
export(xml_length)
export(xml_missing)
export(xml_name)
export(xml_new_document)
export(xml_new_root)
export(xml_ns)
export(xml_ns_rename)
export(xml_ns_strip)
export(xml_parent)
export(xml_parents)
export(xml_path)
export(xml_remove)
export(xml_replace)
export(xml_root)
export(xml_serialize)
export(xml_set_attr)
export(xml_set_attrs)
export(xml_set_name)
export(xml_set_namespace)
export(xml_set_text)
export(xml_siblings)
export(xml_structure)
export(xml_text)
export(xml_type)
export(xml_unserialize)
export(xml_url)
export(xml_validate)
exportClasses(xml_document)
exportClasses(xml_missing)
exportClasses(xml_node)
exportClasses(xml_nodeset)
import(rlang)
importFrom(methods,setOldClass)
useDynLib(xml2, .registration = TRUE)

# xml2
[![Build Status](https://travis-ci.org/r-lib/xml2.svg?branch=master)](https://travis-ci.org/r-lib/xml2)
[![Coverage Status](https://img.shields.io/codecov/c/github/r-lib/xml2/master.svg)](https://codecov.io/github/r-lib/xml2?branch=master)
[![AppVeyor Build Status](https://ci.appveyor.com/api/projects/status/github/r-lib/xml2?branch=master&svg=true)](https://ci.appveyor.com/project/r-lib/xml2)

The xml2 package is a binding to [libxml2](http://xmlsoft.org), making it easy to work with HTML and XML from R. The API is somewhat inspired by [jQuery](http://jquery.com).

## Installation

You can install xml2 from CRAN, 

```R
install.packages("xml2")
```

or you can install the development version from github, using `devtools`:

```R
# install.packages("devtools")
devtools::install_github("r-lib/xml2")
```

## Usage

```R
library("xml2")
x <- read_xml("<foo> <bar> text <baz/> </bar> </foo>")
x

xml_name(x)
xml_children(x)
xml_text(x)
xml_find_all(x, ".//baz")

h <- read_html("<html><p>Hi <b>!")
h
xml_name(h)
xml_text(h)
```

There are three key classes:

* `xml_node`: a single node in a document.

* `xml_doc`: the complete document. Acting on a document is usually the same 
  as acting on the root node of the document.

* `xml_nodeset`: a __set__ of nodes within the document. Operations on 
  `xml_nodeset`s are vectorised, apply the operation over each node in the set.

## Compared to the XML package

xml2 has similar goals to the XML package. The main differences are:

* xml2 takes care of memory management for you. It will automatically
  free the memory used by an XML document as soon as the last reference
  to it goes away.

* xml2 has a very simple class hierarchy so don't need to think about exactly 
  what type of object you have, xml2 will just do the right thing.

* More convenient handling of namespaces in Xpath expressions - see `xml_ns()` 
  and `xml_ns_strip()` to get started.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# xml2

<!-- badges: start -->

[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/xml2)](https://cran.r-project.org/package=xml2)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/xml2/branch/master/graph/badge.svg)](https://app.codecov.io/gh/r-lib/xml2?branch=main)
[![R build
status](https://github.com/r-lib/xml2/workflows/R-CMD-check/badge.svg)](https://github.com/r-lib/xml2/actions)
[![R-CMD-check](https://github.com/r-lib/xml2/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/xml2/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

The xml2 package is a binding to [libxml2](http://xmlsoft.org), making
it easy to work with HTML and XML from R. The API is somewhat inspired
by [jQuery](https://jquery.com).

## Installation

You can install xml2 from CRAN,

``` r
install.packages("xml2")
```

or you can install the development version from github, using
`pak`:

``` r
# install.packages("pak")
pak::pak("r-lib/xml2")
```

## Usage

``` r
library(xml2)
x <- read_xml("<foo> <bar> text <baz/> </bar> </foo>")
x

xml_name(x)
xml_children(x)
xml_text(x)
xml_find_all(x, ".//baz")

h <- read_html("<html><p>Hi <b>!")
h
xml_name(h)
xml_text(h)
```

There are three key classes:

- `xml_node`: a single node in a document.

- `xml_doc`: the complete document. Acting on a document is usually the
  same as acting on the root node of the document.

- `xml_nodeset`: a **set** of nodes within the document. Operations on
  `xml_nodeset`s are vectorised, apply the operation over each node in
  the set.

## Compared to the XML package

xml2 has similar goals to the XML package. The main differences are:

- xml2 takes care of memory management for you. It will automatically
  free the memory used by an XML document as soon as the last reference
  to it goes away.

- xml2 has a very simple class hierarchy so you don’t need to think
  about exactly what type of object you have, xml2 will just do the
  right thing.

- More convenient handling of namespaces in Xpath expressions - see
  `xml_ns()` and `xml_ns_strip()` to get started.

## Code of Conduct

Please note that the xml2 project is released with a [Contributor Code
of Conduct](https://xml2.r-lib.org/CODE_OF_CONDUCT.html). By
contributing to this project, you agree to abide by its terms.

#' Modify a tree by inserting, replacing or removing nodes
#'
#' \code{xml_add_sibling()} and \code{xml_add_child()} are used to insert a node
#' as a sibling or a child. \code{xml_add_parent()} adds a new parent in
#' between the input node and the current parent. \code{xml_replace()}
#' replaces an existing node with a new node. \code{xml_remove()} removes a
#' node from the tree.
#'
#' @details Care needs to be taken when using \code{xml_remove()},
#' @param .x a document, node or nodeset.
#' @param .copy whether to copy the \code{.value} before replacing. If this is \code{FALSE}
#'   then the node will be moved from it's current location.
#' @param .where to add the new node, for \code{xml_add_child} the position
#' after which to add, use \code{0} for the first child. For
#' \code{xml_add_sibling} either \sQuote{"before"} or \sQuote{"after"}
#' indicating if the new node should be before or after \code{.x}.
#' @param ... If named attributes or namespaces to set on the node, if unnamed
#' text to assign to the node.
#' @param .value node to insert.
#' @param free When removing the node also free the memory used for that node.
#' Note if you use this option you cannot use any existing objects pointing to
#' the node or its children, it is likely to crash R or return garbage.
#' @export
xml_replace <- function(.x, .value, ..., .copy = TRUE) {
  UseMethod("xml_replace")
}

#' @export
xml_replace.xml_node <- function(.x, .value, ..., .copy = TRUE) {

  node <- create_node(.value, .x, .copy = .copy, ...)

  .x$node <- node_replace(.x$node, node$node)
  node
}

#' @export
xml_replace.xml_nodeset <- function(.x, .value, ..., .copy = TRUE) {

  if (length(.x) == 0) {
    return(.x)
  }

  # Need to wrap this in a list if a bare xml_node so it is recycled properly
  if (inherits(.value, "xml_node")) {
    .value <- list(.value)
  }

  Map(xml_replace, .x, .value, ..., .copy = .copy)
}

#' @export
xml_replace.xml_missing <- function(.x, .value, ..., .copy = TRUE) {
  .x
}

#' @rdname xml_replace
#' @export
xml_add_sibling <- function(.x, .value, ..., .where = c("after", "before"), .copy = TRUE) {
  UseMethod("xml_add_sibling")
}

#' @export
xml_add_sibling.xml_node <- function(.x, .value, ..., .where = c("after", "before"), .copy = inherits(.value, "xml_node")) {
  .where <- match.arg(.where)

  node <- create_node(.value, .x, .copy = .copy, ...)

  .x$node <- switch(.where,
    before = node_prepend_sibling(.x$node, node$node),
    after = node_append_sibling(.x$node, node$node))

  invisible(.x)
}

#' @export
xml_add_sibling.xml_nodeset <- function(.x, .value, ..., .where = c("after", "before"), .copy = TRUE) {
  if (length(.x) == 0) {
    return(.x)
  }

  .where <- match.arg(.where)

  # Need to wrap this in a list if a bare xml_node so it is recycled properly
  if (inherits(.value, "xml_node")) {
    .value <- list(.value)
  }

  invisible(Map(xml_add_sibling, rev(.x), rev(.value), ..., .where = .where, .copy = .copy))
}

#' @export
xml_add_sibling.xml_missing <- function(.x, .value, ..., .where = c("after", "before"), .copy = TRUE) {
  .x
}

# Helper function used in the xml_add* methods
create_node <- function(.value, parent, .copy, ...) {
  if (inherits(.value, "xml_node")) {
    if (isTRUE(.copy)) {
      .value$node <- node_copy(.value$node)
    }
    return(.value)
  }

  if (inherits(.value, "xml_cdata")) {
    return(xml_node(node_cdata_new(parent$doc, .value), doc = parent$doc))
  }

  if (inherits(.value, "xml_comment")) {
    return(xml_node(node_comment_new(.value), doc = parent$doc))
  }

  if (inherits(.value, "xml_dtd")) {
    node_new_dtd(parent$doc, .value$name, .value$external_id, .value$system_id)
    return()
  }

  if (!is.character(.value)) {
    stop("`.value` must be a character", call. = FALSE)
  }

  parts <- strsplit(.value, ":")[[1]]
  if (length(parts) == 2) {
    namespace <- ns_lookup(parent$doc, parent$node, parts[[1]])
    node <- structure(list(node = node_new_ns(parts[[2]], namespace), doc = parent$doc), class = "xml_node")
  } else {
    node <- structure(list(node = node_new(.value), doc = parent$doc), class = "xml_node")
  }

  args <- list(...)
  named <- has_names(args)
  xml_attrs(node) <- args[named]
  xml_text(node) <- paste(args[!named], collapse = "")

  node
}

#' @rdname xml_replace
#' @export
xml_add_child <- function(.x, .value, ..., .where = length(xml_children(.x)), .copy = TRUE) {
  UseMethod("xml_add_child")
}

#' @export
xml_add_child.xml_node <- function(.x, .value, ..., .where = length(xml_children(.x)), .copy = inherits(.value, "xml_node")) {

  node <- create_node(.value, .x, .copy = .copy, ...)

  if (.where == 0L) {
    if(node_has_children(.x$node))
      node_prepend_child(.x$node, node$node)
    else
      node_append_child(.x$node, node$node)
  } else {
    num_children <- length(xml_children(.x))
    if (.where >= num_children) {
      node_append_child(.x$node, node$node)
    } else
      node_append_sibling(xml_child(.x, search = .where)$node, node$node)
  }

  invisible(node)
}

#' @export
xml_add_child.xml_document <- function(.x, .value, ..., .where = length(xml_children(.x)), .copy = inherits(.value, "xml_node")) {
  if (inherits(.x, "xml_node")) {
    NextMethod("xml_add_child")
  } else {
    node <- create_node(.value, .x, .copy = .copy, ...)
    if (!is.null(node)) {
      if (!doc_has_root(.x$doc)) {
        doc_set_root(.x$doc, node$node)
      }
      node_append_child(doc_root(.x$doc), node$node)
    }
    invisible(xml_document(.x$doc))
  }
}

#' @export
xml_add_child.xml_nodeset <- function(.x, .value, ..., .where = length(xml_children(.x)), .copy = TRUE) {
  if (length(.x) == 0) {
    return(.x)
  }

  # Need to wrap this in a list if a bare xml_node so it is recycled properly
  if (inherits(.value, "xml_node")) {
    .value <- list(.value)
  }

  res <- Map(xml_add_child, .x, .value, ..., .where = .where, .copy = .copy)
  invisible(make_nodeset(res, res[[1]]$doc))
}

#' @export
xml_add_child.xml_missing <- function(.x, .value, ..., .copy = TRUE) {
  .x
}

#' @rdname xml_replace
#' @export
xml_add_parent <- function(.x, .value, ...) {
  UseMethod("xml_add_parent")
}

#' @export
xml_add_parent.xml_node <- function(.x, .value, ...) {
  new_parent <- xml_replace(.x, .value = .value, ..., .copy = FALSE)
  node <- xml_add_child(new_parent, .value = .x, .copy = FALSE)

  invisible(node)
}

#' @export
xml_add_parent.xml_nodeset <- function(.x, .value, ...) {
  if (length(.x) == 0) {
    return(.x)
  }

  # Need to wrap this in a list if a bare xml_node so it is recycled properly
  if (inherits(.value, "xml_node")) {
    .value <- list(.value)
  }

  res <- Map(xml_add_parent, .x, .value, ...)
  invisible(make_nodeset(res, res[[1]]$doc))
}

#' @export
xml_add_parent.xml_missing <- function(.x, .value, ..., .copy = TRUE) {
  invisible(.x)
}

#' @rdname xml_replace
#' @export
xml_remove <- function(.x, free = FALSE) {
   UseMethod("xml_remove")
}

#' @export
xml_remove.xml_node <- function(.x, free = FALSE) {
  node_remove(.x$node, free = free)

  invisible(.x)
}

#' @export
xml_remove.xml_nodeset <- function(.x, free = FALSE) {
  if (length(.x) == 0) {
    return(invisible(.x))
  }

  invisible(Map(xml_remove, rev(.x), free = free))
}

#' @export
xml_remove.xml_missing <- function(.x, free = FALSE) {
  invisible(.x)
}

#' Set the node's namespace
#'
#' The namespace to be set must be already defined in one of the node's
#' ancestors.
#' @param .x a node
#' @param prefix The namespace prefix to use
#' @param uri The namespace URI to use
#' @return the node (invisibly)
#' @export
xml_set_namespace <- function(.x, prefix = "", uri = "") {
  stopifnot(inherits(.x, "xml_node"))

  if (nzchar(uri)) {
    node_set_namespace_uri(.x$doc, .x$node, uri)
  } else {
    node_set_namespace_prefix(.x$doc, .x$node, prefix)
  }
  invisible(.x)
}

#' Create a new document, possibly with a root node
#'
#' \code{xml_new_document} creates only a new document without a root node. In
#' most cases you should instead use \code{xml_new_root}, which creates a new
#' document and assigns the root node in one step.
#' @param version The version number of the document.
#' @param encoding The character encoding to use in the document. The default
#' encoding is \sQuote{UTF-8}. Available encodings are specified at
#' \url{http://xmlsoft.org/html/libxml-encoding.html#xmlCharEncoding}.
#' @return A \code{xml_document} object.
#' @export
# TODO: jimhester 2016-12-16 Deprecate this in the future?
xml_new_document <- function(version = "1.0", encoding = "UTF-8") {
  doc <- doc_new(version)
  structure(list(doc = doc), class = "xml_document")
}

#' @param .version The version number of the document, passed to \code{xml_new_document(version)}.
#' @param .encoding The encoding of the document, passed to \code{xml_new_document(encoding)}.
#' @inheritParams xml_add_child
#' @rdname xml_new_document
#' @export
xml_new_root <- function(.value, ..., .copy = inherits(.value, "xml_node"), .version = "1.0", .encoding = "UTF-8") {
  xml_add_child(xml_new_document(version = .version, encoding = .encoding), .value = .value, ... = ..., .copy = .copy)
}

#' Strip the default namespaces from a document
#'
#' @inheritParams xml_name
#' @examples
#' x <- read_xml(
#'  "<foo xmlns = 'http://foo.com'>
#'    <baz/>
#'    <bar xmlns = 'http://bar.com'>
#'      <baz/>
#'    </bar>
#'   </foo>")
#' # Need to specify the default namespaces to find the baz nodes
#' xml_find_all(x, "//d1:baz")
#' xml_find_all(x, "//d2:baz")
#'
#' # After stripping the default namespaces you can find both baz nodes directly
#' xml_ns_strip(x)
#' xml_find_all(x, "//baz")
#' @export
xml_ns_strip <- function(x) {

  # //namespace::*[name()=''] finds all the namespace definition nodes with no
  # prefix (default namespaces).
  # What we actually want is the element node the definitions are contained in
  # so return the parent (/parent::*)
  namespace_element_nodes <- xml_find_all(x, "//namespace::*[name()='']/parent::*")
  xml_attr(namespace_element_nodes, "xmlns") <- NULL
  invisible(x)
}

# XML parsing tests ------------------------------------------------------------

test_that("multiple default namespaces given unique names", {
  ns <- unclass(xml_ns(read_xml(test_path("ns-multiple-default.xml"))))
  expect_equal(ns, c(d1 = "http://foo.com", d2 = "http://bar.com"))
})

test_that("repeated prefixes given unique names", {
  ns <- unclass(xml_ns(read_xml(test_path("ns-multiple-prefix.xml"))))
  expect_equal(ns, c(b = "http://baz.com", b1 = "http://bar.com"))
})

test_that("aliased prefixes retained", {
  ns <- unclass(xml_ns(read_xml(test_path("ns-multiple-aliases.xml"))))
  expect_equal(ns, c(b = "http://bar.com", c = "http://bar.com"))
})


# Low-level character vector tests ---------------------------------------------

test_that("unique prefix-url combo unchanged", {
  x <- c(blah = "http://blah.com", rah = "http://rah.com")
  expect_equal(.Call(unique_ns, x), x)
})

test_that("all prefixs kept", {
  x <- c(blah = "http://blah.com", rah = "http://blah.com")
  expect_named(.Call(unique_ns, x), c("blah", "rah"))
})

test_that("multiple default namespaces can be stripped", {
  x <- read_xml(test_path("ns-multiple-default.xml"))
  ns <- unclass(xml_ns(x))
  expect_equal(ns, c(d1 = "http://foo.com", d2 = "http://bar.com"))
  expect_length(xml_find_all(x, "//bar"), 0)

  xml_ns_strip(x)
  ns <- unclass(xml_ns(x))

  expect_equal(unname(ns), character())
  expect_length(xml_find_all(x, "//bar"), 2)
})

test_that("modifying nodes works", {
  x <- read_xml("<x><y/></x>")
  node <- xml_find_first(x, "//x")

  expect_equal(xml_name(node), "x")

  .Call(node_set_name, node$node, "y")
  expect_equal(xml_name(node), "y")

  expect_equal(xml_text(node), "")

  .Call(node_set_content, node$node, "test")
  expect_equal(xml_text(node), "test")
})

test_that("xml_text<- only modifies text content", {
  x <- read_xml("<node>Text1<subnode/>text2</node>")

  expect_equal(xml_text(x), "Text1text2")

  # will only change the first text by default
  xml_text(x) <- "new_text1"
  expect_equal(xml_text(x), "new_text1text2")

  # You can change the second by explicitly selecting it
  text_node <- xml_find_first(x, "//text()[2]")
  xml_text(text_node) <- "new_text2"
  expect_equal(xml_text(x), "new_text1new_text2")
})

test_that("xml_text<- creates new text nodes if needed", {
  x <- read_xml("<node><subnode/></node>")
  xml_text(x) <- "test"

  expect_equal(xml_text(x), "test")
})

test_that("xml_remove removes nodes", {
  x <- read_xml("<parent><child>1</child><child>2<child>3</child></child></parent>")
  children <- xml_children(x)
  t1 <- children[[1]]
  xml_remove(children, free = TRUE)
  expect_equal(xml_text(x), "")
})

test_that("xml_replace replaces nodes", {
  x <- read_xml("<parent><child>1</child><child>2<child>3</child></child></parent>")
  children <- xml_children(x)
  t1 <- children[[1]]
  t2 <- children[[2]]
  t3 <- xml_children(children[[2]])[[1]]
  expect_equal(xml_text(x), "123")

  xml_replace(t1, t3)
  expect_equal(xml_text(x), "323")

  first_child <- xml_children(x)[[1]]
  xml_replace(first_child, t1, .copy = FALSE)
  expect_equal(xml_text(x), "123")
  xml_remove(first_child, free = TRUE)

  first_child <- xml_children(x)[[1]]
  xml_replace(first_child, t3, .copy = FALSE)
  expect_equal(xml_text(x), "32")
  xml_remove(first_child, free = TRUE)
})

test_that("xml_replace works with nodesets", {
  x <- read_xml("<parent><child>1</child><child>2<child>3</child></child></parent>")
  children <- xml_children(x)
  t1 <- children[[1]]
  xml_replace(children, t1)
  expect_equal(xml_text(x), "11")
})

test_that("xml_sibling adds a sibling node", {
  x <- read_xml("<parent><child>1</child><child>2<child>3</child></child></parent>")
  children <- xml_children(x)
  t1 <- children[[1]]
  t2 <- children[[2]]
  t3 <- xml_children(children[[2]])[[1]]

  xml_add_sibling(t1, t3)
  expect_length(xml_siblings(t1), 2)
  expect_equal(xml_text(x), "1323")

  xml_add_sibling(t1, t3, .where = "before")
  expect_length(xml_siblings(t1), 3)
  expect_equal(xml_text(x), "31323")

  children <- xml_children(x)
  xml_add_sibling(children, t1)
  expect_equal(xml_text(x), "311131231")
})

test_that("xml_add_child adds a child node", {
  x <- read_xml("<parent><child>1</child><child>2<child>3</child></child></parent>")
  children <- xml_children(x)
  t1 <- children[[1]]
  t2 <- children[[2]]
  t3 <- xml_children(children[[2]])[[1]]

  expect_length(xml_children(t1), 0)

  xml_add_child(t1, t3, .copy = TRUE)
  expect_length(xml_children(t1), 1)
  expect_equal(xml_text(x), "1323")

  children <- xml_children(x)
  xml_add_child(children, t1)
  expect_equal(xml_text(x), "1313231313")
})

test_that("xml_add_child can create a new default namespace", {
  x <- xml_root(xml_add_child(xml_new_document(), "foo", xmlns = "bar"))

  expect_equal(unclass(xml_ns(x)), c(d1 = "bar"))
})

test_that("xml_add_child can create a new prefixed namespace", {
  x <- xml_root(xml_add_child(xml_new_document(), "foo", "xmlns:bar" = "baz"))

  expect_equal(unclass(xml_ns(x)), c(bar = "baz"))
})

test_that("xml_add_child can create a new attribute", {
  x <- xml_add_child(xml_new_document(), "foo", "bar" = "baz")

  expect_equal(xml_attr(x, "bar"), "baz")
})

test_that("xml_add_child can create new text", {
  x <- xml_add_child(xml_new_document(), "foo", "bar")

  expect_equal(xml_text(x), "bar")
})

test_that("xml_add_child can create a new node with the specified prefix", {
  x <- xml_root(xml_add_child(xml_new_document(), "foo", "xmlns:bar" = "baz"))

  t1 <- xml_add_child(x, "bar:qux")
  expect_equal(xml_name(t1), "qux")
  expect_equal(xml_name(t1, xml_ns(x)), "bar:qux")
})

test_that("xml_add_child can create a new node with the specified prefix", {
  x <- xml_root(xml_add_child(xml_new_document(), "foo", "xmlns:bar" = "baz"))

  expect_error(xml_add_child(x, "bar2:qux"), "No namespace with prefix `bar2` found")
})

# https://github.com/r-lib/xml2/issues/284
test_that("xml_add_child can create a new node with attribute named 'par'", {
  x <- xml_new_root("foo")
  t1 <- xml_add_child(x, "a", "par" = "b")
  expect_true(xml_has_attr(t1, "par"))
})

test_that("xml_add_parent works with xml_node input", {
  x <- read_xml("<x><y/></x>")
  y <- xml_find_first(x, ".//y")
  xml_add_parent(y, "z")

  expect_equal(xml_name(xml_parent(y)), "z")
  expect_equal(xml_name(xml_child(x)), "z")
})

test_that("xml_add_parent works with xml_nodeset input", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_all(x, ".//y")
  xml_add_parent(y, "z")

  expect_equal(xml_name(xml_parent(y)), c("z", "z"))
  expect_equal(xml_name(xml_child(x)), "z")
})

test_that("xml_add_parent works with xml_missing input", {
  x <- read_xml("<body>
    <p>Some <b>text</b>.</p>
    <p>Some <b>other</b>.</p>
    <p>No bold text</p>
    </body>")

  y <- xml_find_all(x, ".//p")
  z <- xml_find_first(y, ".//b")
  xml_add_parent(z, "em")

  expect_equal(xml_name(xml_parent(z)), c("em", "em"))
  expect_equal(xml_name(xml_children(y)), c("em", "em"))
})

test_that("xml_new_document adds a default character encoding", {
  x <- read_xml("<root>\u00E1\u00FC\u00EE</root>")
  expect_equal(as.character(x), "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<root>\u00E1\u00FC\u00EE</root>\n")

  x2 <- xml_new_document()
  xml_add_child(x2, "root", "\u00E1\u00FC\u00EE")
  expect_equal(as.character(x2), "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<root>\u00E1\u00FC\u00EE</root>\n")
})

test_that("xml_new_root is equivalent to using xml_new_document xml_add_child", {
  x1 <- xml_add_child(xml_new_document(), "foo", "bar")

  x2 <- xml_new_root("foo", "bar")

  expect_identical(as.character(x1), as.character(x2))
})

test_that("xml_add_child can insert anywhere in the child list", {
  x <- read_xml("<a/>")

  xml_add_child(x, "z")
  expect_equal(c("z"), xml_name(xml_children(x)))

  xml_add_child(x, "w", .where = 0)
  expect_equal(c("w", "z"), xml_name(xml_children(x)))

  xml_add_child(x, "y", .where = 1)
  expect_equal(c("w", "y", "z"), xml_name(xml_children(x)))

  xml_add_child(x, "x", .where = 1)
  expect_equal(c("w", "x", "y", "z"), xml_name(xml_children(x)))
})

test_that("xml_add_child can insert anywhere in a nodeset", {
  x <- read_xml("<body>
    <p>Some <b>text</b>.</p>
    <p>Some <b>other</b>.</p>
    <p>No bold text</p>
    </body>")

  y <- xml_find_all(x, ".//p")
  z <- xml_find_first(y, ".//b")

  xml_add_child(z, "bar")
  xml_add_child(z, "foo", .where = 0)

  expect_equal(c("foo", "bar", "foo", "bar"), xml_name(xml_children(z)))
})

test_that("Can write root nodes with namespaces", {
  x <- xml_new_root("foo:bar", "xmlns:foo" = "http://foo/bar")
  expect_equal(unclass(xml_ns(x)), c(foo = "http://foo/bar"))
  expect_equal(as.character(xml_find_first(x, "/*")), "<foo:bar xmlns:foo=\"http://foo/bar\"/>")
})

context("modify nodes")

test_that("modifying nodes works", {
  x <- read_xml("<x><y/></x>")
  node <- xml_find_first(x, "//x")

  expect_equal(xml_name(node), "x")

  node_set_name(node$node, "y")
  expect_equal(xml_name(node), "y")

  expect_equal(xml_text(node), "")

  node_set_content(node$node, "test")
  expect_equal(xml_text(node), "test")
})

test_that("xml_text<- only modifies text content", {
  x <- read_xml("<node>Text1<subnode/>text2</node>")

  expect_equal(xml_text(x), "Text1text2")

  # will only change the first text by default
  xml_text(x) <- "new_text1"
  expect_equal(xml_text(x), "new_text1text2")

  # You can change the second by explicitly selecting it
  text_node <- xml_find_first(x, "//text()[2]")
  xml_text(text_node) <- "new_text2"
  expect_equal(xml_text(x), "new_text1new_text2")
})

test_that("xml_text<- creates new text nodes if needed", {
  x <- read_xml("<node><subnode/></node>")
  xml_text(x) <- "test"

  expect_equal(xml_text(x), "test")
})

test_that("xml_remove removes nodes", {

  x <- read_xml("<parent><child>1</child><child>2<child>3</child></child></parent>")
  children <- xml_children(x)
  t1 <- children[[1]]
  xml_remove(children, free = TRUE)
  expect_equal(xml_text(x), "")
})

test_that("xml_replace replaces nodes", {

  x <- read_xml("<parent><child>1</child><child>2<child>3</child></child></parent>")
  children <- xml_children(x)
  t1 <- children[[1]]
  t2 <- children[[2]]
  t3 <- xml_children(children[[2]])[[1]]
  expect_equal(xml_text(x), "123")

  xml_replace(t1, t3)
  expect_equal(xml_text(x), "323")

  first_child <- xml_children(x)[[1]]
  xml_replace(first_child, t1, .copy = FALSE)
  expect_equal(xml_text(x), "123")
  xml_remove(first_child, free = TRUE)

  first_child <- xml_children(x)[[1]]
  xml_replace(first_child, t3, .copy = FALSE)
  expect_equal(xml_text(x), "32")
  xml_remove(first_child, free = TRUE)
})

test_that("xml_replace works with nodesets", {

  x <- read_xml("<parent><child>1</child><child>2<child>3</child></child></parent>")
  children <- xml_children(x)
  t1 <- children[[1]]
  xml_replace(children, t1)
  expect_equal(xml_text(x), "11")
})

test_that("xml_sibling adds a sibling node", {
  x <- read_xml("<parent><child>1</child><child>2<child>3</child></child></parent>")
  children <- xml_children(x)
  t1 <- children[[1]]
  t2 <- children[[2]]
  t3 <- xml_children(children[[2]])[[1]]

  xml_add_sibling(t1, t3)
  expect_length(xml_siblings(t1), 2)
  expect_equal(xml_text(x), "1323")

  xml_add_sibling(t1, t3, .where = "before")
  expect_length(xml_siblings(t1), 3)
  expect_equal(xml_text(x), "31323")

  children <- xml_children(x)
  xml_add_sibling(children, t1)
  expect_equal(xml_text(x), "311131231")
})

test_that("xml_add_child adds a child node", {
  x <- read_xml("<parent><child>1</child><child>2<child>3</child></child></parent>")
  children <- xml_children(x)
  t1 <- children[[1]]
  t2 <- children[[2]]
  t3 <- xml_children(children[[2]])[[1]]

  expect_length(xml_children(t1), 0)

  xml_add_child(t1, t3, .copy = TRUE)
  expect_length(xml_children(t1), 1)
  expect_equal(xml_text(x), "1323")

  children <- xml_children(x)
  xml_add_child(children, t1)
  expect_equal(xml_text(x), "1313231313")
})

test_that("xml_add_child can create a new default namespace", {
  x <- xml_root(xml_add_child(xml_new_document(), "foo", xmlns = "bar"))

  expect_equal(unclass(xml_ns(x)), c(d1 = "bar"))
})

test_that("xml_add_child can create a new prefixed namespace", {
  x <- xml_root(xml_add_child(xml_new_document(), "foo", "xmlns:bar" = "baz"))

  expect_equal(unclass(xml_ns(x)), c(bar = "baz"))
})

test_that("xml_add_child can create a new attribute", {
  x <- xml_add_child(xml_new_document(), "foo", "bar" = "baz")

  expect_equal(xml_attr(x, "bar"), "baz")
})

test_that("xml_add_child can create new text", {
  x <- xml_add_child(xml_new_document(), "foo", "bar")

  expect_equal(xml_text(x), "bar")
})

test_that("xml_add_child can create a new node with the specified prefix", {
  x <- xml_root(xml_add_child(xml_new_document(), "foo", "xmlns:bar" = "baz"))

  t1 <- xml_add_child(x, "bar:qux")
  expect_equal(xml_name(t1), "qux")
  expect_equal(xml_name(t1, xml_ns(x)), "bar:qux")
})

test_that("xml_add_child can create a new node with the specified prefix", {
  x <- xml_root(xml_add_child(xml_new_document(), "foo", "xmlns:bar" = "baz"))

  expect_error(xml_add_child(x, "bar2:qux"), "No namespace with prefix `bar2` found")
})

test_that("xml_add_parent works with xml_node input", {
  x <- read_xml("<x><y/></x>")
  y <- xml_find_first(x, ".//y")
  xml_add_parent(y, "z")

  expect_equal(xml_name(xml_parent(y)), "z")
  expect_equal(xml_name(xml_child(x)), "z")
})

test_that("xml_add_parent works with xml_nodeset input", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_all(x, ".//y")
  xml_add_parent(y, "z")

  expect_equal(xml_name(xml_parent(y)), c("z", "z"))
  expect_equal(xml_name(xml_child(x)), "z")
})

test_that("xml_add_parent works with xml_missing input", {
  x <- read_xml("<body>
    <p>Some <b>text</b>.</p>
    <p>Some <b>other</b>.</p>
    <p>No bold text</p>
    </body>")

    y <- xml_find_all(x, ".//p")
    z <- xml_find_first(y, ".//b")
    xml_add_parent(z, "em")

    expect_equal(xml_name(xml_parent(z)), c("em", "em"))
    expect_equal(xml_name(xml_children(y)), c("em", "em"))
})

test_that("xml_new_document adds a default character encoding", {

  x <- read_xml("<root>\u00E1\u00FC\u00EE</root>")
  expect_equal(as.character(x), "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<root>\u00E1\u00FC\u00EE</root>\n")

  x2 <- xml_new_document()
  xml_add_child(x2, "root", "\u00E1\u00FC\u00EE")
  expect_equal(as.character(x2), "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<root>\u00E1\u00FC\u00EE</root>\n")
})

test_that("xml_new_root is equivalent to using xml_new_document xml_add_child", {
  x1 <- xml_add_child(xml_new_document(), "foo", "bar")

  x2 <- xml_new_root("foo", "bar")

  expect_identical(as.character(x1), as.character(x2))
})

test_that("xml_add_child can insert anywhere in the child list", {
  x <- read_xml("<a/>")

  xml_add_child(x, "z")
  expect_equal(c("z"), xml_name(xml_children(x)))

  xml_add_child(x, "w", .where = 0)
  expect_equal(c("w", "z"), xml_name(xml_children(x)))

  xml_add_child(x, "y", .where = 1)
  expect_equal(c("w", "y", "z"), xml_name(xml_children(x)))

  xml_add_child(x, "x", .where = 1)
  expect_equal(c("w", "x", "y", "z"), xml_name(xml_children(x)))
})

test_that("xml_add_child can insert anywhere in a nodeset", {
  x <- read_xml("<body>
    <p>Some <b>text</b>.</p>
    <p>Some <b>other</b>.</p>
    <p>No bold text</p>
    </body>")

    y <- xml_find_all(x, ".//p")
    z <- xml_find_first(y, ".//b")

    xml_add_child(z, "bar")
    xml_add_child(z, "foo", .where = 0)

    expect_equal(c("foo", "bar", "foo", "bar"), xml_name(xml_children(z)))
})

context("comment")

test_that("Comment creation works", {
  x <- xml_new_root("root")
  xml_add_child(x, xml_comment("Hello!"))
  expect_identical("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<root><!--Hello!--></root>\n", as.character(x, options = ""))
})

context("Namespaces")

# XML parsing tests ------------------------------------------------------------

test_that("multiple default namespaces given unique names", {
  ns <- unclass(xml_ns(read_xml("ns-multiple-default.xml")))
  expect_equal(ns, c(d1 = "http://foo.com", d2 = "http://bar.com"))
})

test_that("repeated prefixes given unique names", {
  ns <- unclass(xml_ns(read_xml("ns-multiple-prefix.xml")))
  expect_equal(ns, c(b = "http://baz.com", b1 = "http://bar.com"))
})

test_that("aliased prefixes retained", {
  ns <- unclass(xml_ns(read_xml("ns-multiple-aliases.xml")))
  expect_equal(ns, c(b = "http://bar.com", c = "http://bar.com"))
})


# Low-level character vector tests ---------------------------------------------

test_that("unique prefix-url combo unchanged", {
  x <- c(blah = "http://blah.com", rah = "http://rah.com")
  expect_equal(unique_ns(x), x)
})

test_that("all prefixs kept", {
  x <- c(blah = "http://blah.com", rah = "http://blah.com")
  expect_equal(names(unique_ns(x)), c("blah", "rah"))
})

test_that("multiple default namespaces can be stripped", {
  x <- read_xml("ns-multiple-default.xml")
  ns <- unclass(xml_ns(x))
  expect_equal(ns, c(d1 = "http://foo.com", d2 = "http://bar.com"))
  expect_equal(length(xml_find_all(x, "//bar")), 0)

  xml_ns_strip(x)
  ns <- unclass(xml_ns(x))

  expect_equivalent(ns, character())
  expect_equal(length(xml_find_all(x, "//bar")), 2)
})

context("dtd")

test_that("xml_dtd works", {

  r <- xml_new_root(xml_dtd(name = "html", external_id = "-//W3C//DTD XHTML 1.0 Transitional//EN", system_id = "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"))
  expect_identical("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n", as.character(r))

  no_name <- xml_new_root(xml_dtd(external_id = "-//W3C//DTD XHTML 1.0 Transitional//EN", system_id = "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"))
  expect_identical("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE  PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n", as.character(no_name))

  no_name_external_id <- xml_new_root(xml_dtd(system_id = "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"))
  expect_identical("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE  SYSTEM \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n", as.character(no_name_external_id))

  no_name_external_id_internal_id <- xml_new_root(xml_dtd())
  expect_identical("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE >\n", as.character(no_name_external_id_internal_id))
})

context("cdata")

test_that("CDATA creation works", {
  x <- xml_new_root("root")
  xml_add_child(x, xml_cdata("<d/>"))
  expect_identical(as.character(x),  "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<root><![CDATA[<d/>]]></root>\n")
})

xml_document <- function(doc) {
  if (.Call(doc_has_root, doc)) {
    x <- xml_node(.Call(doc_root, doc), doc)
    class(x) <- c("xml_document", class(x))
    x
  } else {
    out <- list(doc = doc)
    class(out) <- "xml_document"
    out
  }
}

doc_type <- function(x) {
  if (is.null(x$doc)) {
    return("xml")
  }
  if (.Call(doc_is_html, x$doc)) {
    "html"
  } else {
    "xml"
  }
}

#' @export
print.xml_document <- function(x, width = getOption("width"), max_n = 20, ...) {
  doc <- xml_document(x$doc)
  cat("{", doc_type(x), "_document}\n", sep = "")
  if (inherits(doc, "xml_node")) {
    cat(format(doc), "\n", sep = "")
    show_nodes(xml_children(doc), width = width, max_n = max_n)
  }
}

#' @export
as.character.xml_document <- function(
  x,
  ...,
  options = "format",
  encoding = "UTF-8"
) {
  options <- parse_options(options, xml_save_options())
  .Call(doc_write_character, x$doc, encoding, options)
}

#' XML namespaces.
#'
#' \code{xml_ns} extracts all namespaces from a document, matching each
#' unique namespace url with the prefix it was first associated with. Default
#' namespaces are named \code{d1}, \code{d2} etc. Use \code{xml_ns_rename}
#' to change the prefixes. Once you have a namespace object, you can pass it to
#' other functions to work with fully qualified names instead of local names.
#'
#' @export
#' @inheritParams xml_name
#' @param old,... An existing xml_namespace object followed by name-value
#'   (old prefix-new prefix) pairs to replace.
#' @return A character vector with class \code{xml_namespace} so the
#'   default display is a little nicer.
#' @examples
#' x <- read_xml('
#'  <root>
#'    <doc1 xmlns = "http://foo.com"><baz /></doc1>
#'    <doc2 xmlns = "http://bar.com"><baz /></doc2>
#'  </root>
#' ')
#' xml_ns(x)
#'
#' # When there are default namespaces, it's a good idea to rename
#' # them to give informative names:
#' ns <- xml_ns_rename(xml_ns(x), d1 = "foo", d2 = "bar")
#' ns
#'
#' # Now we can pass ns to other xml function to use fully qualified names
#' baz <- xml_children(xml_children(x))
#' xml_name(baz)
#' xml_name(baz, ns)
#'
#' xml_find_all(x, "//baz")
#' xml_find_all(x, "//foo:baz", ns)
#'
#' str(as_list(x))
#' str(as_list(x, ns))
xml_ns <- function(x) {
  UseMethod("xml_ns")
}

#' @export
xml_ns.xml_document <- function(x) {
  if (length(x) == 0) {
    return(character())
  }

  stopifnot(inherits(x, "xml_document"))
  doc <- x$doc
  x <- doc_namespaces(doc)

  # Number default namespaces
  is_default <- names(x) == ""
  names(x)[is_default] <- paste0("d", seq_len(sum(is_default)))

  # Make prefixes unique
  names(x) <- make.unique(names(x), "")

  class(x) <- "xml_namespace"

  x
}

#' @export
xml_ns.xml_node <- function(x) {
  xml_ns(xml_root(x))
}

#' @export
xml_ns.xml_nodeset <- function(x) {
  if (length(x) == 0) {
    return(character())
  }
  xml_ns(x[[1]])
}

#' @export
xml_ns.xml_missing <- function(x) {
  character()
}

#' @export
print.xml_namespace <- function(x, ...) {
  prefix <- format(names(x))

  cat(paste0(prefix, " <-> ", x, collapse = "\n"), "\n", sep = "")
}

#' @export
#' @rdname xml_ns
xml_ns_rename <- function(old, ...) {
  new <- c(...)

  m <- match(names(new), names(old))
  if (any(is.na(m))) {
    missing <- paste(names(new)[is.na(m)], collapse = ", ")
    stop("Some prefixes [", missing, "] don't already exist.", call. = FALSE)
  }

  names(old)[m] <- new
  old
}

#' The (tag) name of an xml element.
#'
#' @param x A document, node, or node set.
#' @param ns Optionally, a named vector giving prefix-url pairs, as produced
#'   by [xml_ns()]. If provided, all names will be explicitly
#'   qualified with the ns prefix, i.e. if the element `bar` is defined
#'   in namespace `foo`, it will be called `foo:bar`. (And
#'   similarly for attributes). Default namespaces must be given an explicit
#'   name. The ns is ignored when using [xml_name<-()] and
#'   [xml_set_name()].
#' @return A character vector.
#' @export
#' @examples
#' x <- read_xml("<bar>123</bar>")
#' xml_name(x)
#'
#' y <- read_xml("<bar><baz>1</baz>abc<foo /></bar>")
#' z <- xml_children(y)
#' xml_name(xml_children(y))
xml_name <- function(x, ns = character()) {
  .Call(node_name, x, ns)
}

#' Modify the (tag) name of an element
#'
#' @param value a character vector with replacement name.
#' @rdname xml_name
#' @export
`xml_name<-` <- function(x, ns = character(), value) {
  UseMethod("xml_name<-")
}

#' @export
`xml_name<-.xml_node` <- function(x, ns = character(), value) {
  .Call(node_set_name, x$node, value)
  x
}

#' @export
`xml_name<-.xml_nodeset` <- function(x, ns = character(), value) {
  if (length(x) == 0) {
    return(x)
  }
  if (!is.list(ns)) {
    ns <- list(ns)
  }
  Map(`xml_name<-`, x, ns, value)
  x
}

#' @export
`xml_name<-.xml_missing` <- function(x, ns = character(), value) {
  x
}

set_name <- function(x, value, ns = character()) {
  xml_name(x = x, ns = ns) <- value
  x
}

#' @rdname xml_name
#' @export
xml_set_name <- function(x, value, ns = character()) {
  UseMethod("xml_set_name")
}

#' @export
xml_set_name.xml_node <- set_name

#' @export
xml_set_name.xml_nodeset <- set_name

#' @export
xml_set_name.xml_missing <- set_name

test_that("xml_name() returns the name", {
  x <- read_xml("<body>
    <p>Some <b>text</b>.</p>
    <p>Some <i>other</i>.</p>
    <p>No bold text</p>
    </body>")

  children <- xml_children(x)
  x <- xml_find_first(children, ".//b|.//i")

  expect_equal(xml_name(x[[1]]), "b")
  expect_equal(xml_name(x[[2]]), "i")
  expect_equal(xml_name(x[[3]]), NA_character_)

  expect_equal(xml_name(x), c("b", "i", NA_character_))
})

test_that("qualified names returned when ns given", {
  x <- read_xml(test_path("ns-multiple-default.xml"))
  ns <- xml_ns(x)

  bars <- xml_children(xml_children(x))
  expect_equal(xml_name(bars), c("bar", "bar"))
  expect_equal(xml_name(bars, ns), c("d1:bar", "d2:bar"))
})

test_that("error if missing ns spec", {
  x <- read_xml(test_path("ns-multiple-default.xml"))
  ns <- xml_ns(x)[1]

  bars <- xml_children(xml_children(x))
  expect_snapshot(error = TRUE, xml_name(bars, ns))
})

test_that("xml_name<- modifies the name", {
  x <- read_xml(test_path("ns-multiple-default.xml"))
  ns <- xml_ns(x)

  bars <- xml_children(xml_children(x))
  bar <- bars[[1]]

  xml_name(bar) <- "foo"
  expect_equal(xml_name(bar), "foo")
  expect_equal(xml_name(bar, ns), "d1:foo")

  # ns is ignored
  xml_name(bar, ns) <- "bar"
  expect_equal(xml_name(bar), "bar")
  expect_equal(xml_name(bar, ns), "d1:bar")

  xml_name(bars) <- "foo"
  expect_equal(xml_name(bars), c("foo", "foo"))

  old_mss <- mss <- xml_missing()
  xml_name(mss) <- "foo"
  expect_identical(old_mss, mss)
})

test_that("xml_set_name modifies the name", {
  x <- read_xml(test_path("ns-multiple-default.xml"))
  ns <- xml_ns(x)

  bars <- xml_children(xml_children(x))
  bar <- bars[[1]]

  xml_set_name(bar, "foo")
  expect_equal(xml_name(bar), "foo")
  expect_equal(xml_name(bar, ns), "d1:foo")

  # ns is ignored
  xml_set_name(bar, "bar", ns)
  expect_equal(xml_name(bar), "bar")
  expect_equal(xml_name(bar, ns), "d1:bar")

  xml_set_name(bars, "foo")
  expect_equal(xml_name(bars), c("foo", "foo"))

  old_mss <- mss <- xml_missing()
  xml_set_name(mss, "foo")
  expect_identical(old_mss, mss)
})

test_that("xml_name() returns the name", {
  x <- read_xml(
    "<body>
    <p>Some <b>text</b>.</p>
    <p>Some <i>other</i>.</p>
    <p>No bold text</p>
    </body>"
  )

  children <- xml_children(x)
  x <- xml_find_first(children, ".//b|.//i")

  expect_equal(xml_name(x[[1]]), "b")
  expect_equal(xml_name(x[[2]]), "i")
  expect_equal(xml_name(x[[3]]), NA_character_)

  expect_equal(xml_name(x), c("b", "i", NA_character_))
})

test_that("qualified names returned when ns given", {
  x <- read_xml(test_path("ns-multiple-default.xml"))
  ns <- xml_ns(x)

  bars <- xml_children(xml_children(x))
  expect_equal(xml_name(bars), c("bar", "bar"))
  expect_equal(xml_name(bars, ns), c("d1:bar", "d2:bar"))
})

test_that("error if missing ns spec", {
  x <- read_xml(test_path("ns-multiple-default.xml"))
  ns <- xml_ns(x)[1]

  bars <- xml_children(xml_children(x))
  expect_snapshot(error = TRUE, xml_name(bars, ns))
})

test_that("xml_name<- modifies the name", {
  x <- read_xml(test_path("ns-multiple-default.xml"))
  ns <- xml_ns(x)

  bars <- xml_children(xml_children(x))
  bar <- bars[[1]]

  xml_name(bar) <- "foo"
  expect_equal(xml_name(bar), "foo")
  expect_equal(xml_name(bar, ns), "d1:foo")

  # ns is ignored
  xml_name(bar, ns) <- "bar"
  expect_equal(xml_name(bar), "bar")
  expect_equal(xml_name(bar, ns), "d1:bar")

  xml_name(bars) <- "foo"
  expect_equal(xml_name(bars), c("foo", "foo"))

  old_mss <- mss <- xml_missing()
  xml_name(mss) <- "foo"
  expect_identical(old_mss, mss)
})

test_that("xml_set_name modifies the name", {
  x <- read_xml(test_path("ns-multiple-default.xml"))
  ns <- xml_ns(x)

  bars <- xml_children(xml_children(x))
  bar <- bars[[1]]

  xml_set_name(bar, "foo")
  expect_equal(xml_name(bar), "foo")
  expect_equal(xml_name(bar, ns), "d1:foo")

  # ns is ignored
  xml_set_name(bar, "bar", ns)
  expect_equal(xml_name(bar), "bar")
  expect_equal(xml_name(bar, ns), "d1:bar")

  xml_set_name(bars, "foo")
  expect_equal(xml_name(bars), c("foo", "foo"))

  old_mss <- mss <- xml_missing()
  xml_set_name(mss, "foo")
  expect_identical(old_mss, mss)
})

#' Coerce xml nodes to a list.
#'
#' This turns an XML document (or node or nodeset) into the equivalent R
#' list. Note that this is \code{as_list()}, not \code{as.list()}:
#' \code{lapply()} automatically calls \code{as.list()} on its inputs, so
#' we can't override the default.
#'
#' \code{as_list} currently only handles the four most common types of
#' children that an element might have:
#'
#' \itemize{
#'   \item Other elements, converted to lists.
#'   \item Attributes, stored as R attributes. Attributes that have special meanings in R
#'           (\code{\link{class}}, \code{\link{comment}}, \code{\link{dim}},
#'           \code{\link{dimnames}}, \code{\link{names}}, \code{\link{row.names}} and
#'           \code{\link{tsp}}) are escaped with '.'
#'   \item Text, stored as a character vector.
#' }
#'
#' @inheritParams xml_name
#' @param ... Needed for compatibility with generic. Unused.
#' @export
#' @examples
#' as_list(read_xml("<foo> a <b /><c><![CDATA[<d></d>]]></c></foo>"))
#' as_list(read_xml("<foo> <bar><baz /></bar> </foo>"))
#' as_list(read_xml("<foo id = 'a'></foo>"))
#' as_list(read_xml("<foo><bar id='a'/><bar id='b'/></foo>"))
as_list <- function(x, ns = character(), ...) {
  UseMethod("as_list")
}

#' @export
as_list.xml_missing <- function(x, ns = character(), ...) {
  list()
}

#' @export
as_list.xml_document <- function(x, ns = character(), ...) {
  if (!inherits(x, "xml_node")) {
    return(list())
  }

  out <- list(NextMethod())
  names(out) <- xml_name(x)
  out
}

#' @export
as_list.xml_node <- function(x, ns = character(), ...) {
  contents <- xml_contents(x)
  if (length(contents) == 0) {
    # Base case - contents
    type <- xml_type(x)

    if (type %in% c("text", "cdata"))
      return(xml_text(x))
    if (type != "element" && type != "document")
      return(paste("[", type, "]"))

    out <- list()
  } else {
    out <- lapply(seq_along(contents), function(i) as_list(contents[[i]], ns = ns))

    nms <- ifelse(xml_type(contents) == "element", xml_name(contents, ns = ns), "")
    if (any(nms != "")) {
      names(out) <- nms
    }
  }

  # Add xml attributes as R attributes
  attributes(out) <- c(list(names = names(out)), xml_to_r_attrs(xml_attrs(x, ns = ns)))

  out
}

#' @export
as_list.xml_nodeset <- function(x, ns = character(), ...) {
  lapply(seq_along(x), function(i) as_list(x[[i]], ns = ns))
}

special_attributes <- c("class", "comment", "dim", "dimnames", "names", "row.names", "tsp")

xml_to_r_attrs <- function(x) {
  if (length(x) == 0) {
    return(NULL)
  }
  # escape special names
  special <- names(x) %in% special_attributes
  names(x)[special] <- paste0(".", names(x)[special])
  as.list(x)
}

r_attrs_to_xml <- function(x) {
  if (length(x) == 0) {
    return(NULL)
  }

  # Drop R special attributes
  x <- x[!names(x) %in% special_attributes]

  # Rename any xml attributes needed
  special <- names(x) %in% paste0(".", special_attributes)

  names(x)[special] <- sub("^\\.", "", names(x)[special])
  x
}

roundtrip_xml <- function(x) {
  xml <- read_xml(x)
  lst <- as_list(xml)
  xml2 <- as_xml_document(lst)
  expect_equal(as.character(xml), as.character(xml2))
}

test_that("roundtrips with single children", {
  roundtrip_xml("<a><b/></a>")

  roundtrip_xml("<a><b><c/></b></a>")

  roundtrip_xml("<a><b>foo<c/></b></a>")

  roundtrip_xml("<a><b>foo<c>bar</c></b></a>")

  roundtrip_xml("<a x = '1'><b y = '2'>foo<c z = '3'>bar</c></b></a>")
})

test_that("roundtrips with multi children", {
  roundtrip_xml("<a><b1/><b2/></a>")

  roundtrip_xml("<a><b><c1/><c2/></b></a>")

  roundtrip_xml("<a><b1>foo<c/></b1><b2>bar<c/></b2></a>")

  roundtrip_xml("<a><b>foo<c>bar</c><c>baz</c></b></a>")

  roundtrip_xml("<a x = '1'><b y = '2'>foo<c z = '3'>bar</c></b></a>")
  roundtrip_xml("<a x = '1'><b y = '2'>foo<c z = '3'>bar</c></b><c zz = '4'>baz</c></a>")
})

test_that("rountrips with special attributes", {
  roundtrip_xml("<a names = 'test'><b/></a>")
})

test_that("more than one root node is an error", {
  expect_error(as_xml_document(list(a = list(), b = list())), "Root nodes must be of length 1")
})

test_that("Can convert nodes with leading and trailing text", {
  roundtrip_xml("<a>foo<b>bar</b>baz</a>")
})

list_xml <- function(x) as_list(read_xml(x))

test_that("empty elements become empty lists", {
  expect_equal(list_xml("<x></x>"), list(x = list()))
  expect_equal(list_xml("<x><y/></x>"), list(x = list(y = list())))
  expect_equal(list_xml("<x><y><z/></y></x>"), list(x = list(y = list(z = list()))))
})

test_that("text nodes become character vectors", {
  expect_equal(list_xml("<x>a</x>"), list(x = list("a")))
  expect_equal(list_xml("<x><y>a</y></x>"), list(x = list(y = list("a"))))
})

test_that("cdata nodes become character vectors", {
  expect_equal(list_xml("<x><![CDATA[<y/>]]></x>"), list(x = list("<y/>")))
})

test_that("xml attributes become R attibutes", {
  expect_equal(list_xml("<x a='1' b='2'></x>"), list(x = structure(list(), a = "1", b = "2")))
})

test_that("xml names are preserved when attributes exist", {
  expect_equal(
    list_xml("<x a='1' b='2'><y>3</y><z>4</z></x>"),
    list(x = structure(list(y = list("3"), z = list("4")), a = "1", b = "2"))
  )
})

test_that("special attributes are escaped", {
  expect_equal(
    list_xml("<x a='1' b='2' names='esc'><y>3</y><z>4</z></x>"),
    list(x = structure(list(y = list("3"), z = list("4")), a = "1", b = "2", .names = "esc"))
  )
})

test_that("attributes in child nodes", {
  expect_equal(
    list_xml("<w aa = '0'><x a='1' b='2' names='esc'><y>3</y><z>4</z></x></w>"),
    list(w = structure(list(x = structure(list(y = list("3"), z = list("4")), a = "1", b = "2", .names = "esc")), aa = "0"))
  )
})

context("as_xml_document")

roundtrip_xml <- function(x) {
  xml <- read_xml(x)
  lst <- as_list(xml)
  xml2 <- as_xml_document(lst)
  expect_equal(as.character(xml), as.character(xml2))
}

test_that("roundtrips with single children", {
  roundtrip_xml("<a><b/></a>")

  roundtrip_xml("<a><b><c/></b></a>")

  roundtrip_xml("<a><b>foo<c/></b></a>")

  roundtrip_xml("<a><b>foo<c>bar</c></b></a>")

  roundtrip_xml("<a x = '1'><b y = '2'>foo<c z = '3'>bar</c></b></a>")
})

test_that("roundtrips with multi children", {
  roundtrip_xml("<a><b1/><b2/></a>")

  roundtrip_xml("<a><b><c1/><c2/></b></a>")

  roundtrip_xml("<a><b1>foo<c/></b1><b2>bar<c/></b2></a>")

  roundtrip_xml("<a><b>foo<c>bar</c><c>baz</c></b></a>")

  roundtrip_xml("<a x = '1'><b y = '2'>foo<c z = '3'>bar</c></b></a>")
  roundtrip_xml("<a x = '1'><b y = '2'>foo<c z = '3'>bar</c></b><c zz = '4'>baz</c></a>")
})

test_that("rountrips with special attributes", {
  roundtrip_xml("<a names = 'test'><b/></a>")
})

test_that("more than one root node is an error", {
  expect_error(as_xml_document(list(a = list(), b = list())), "Root nodes must be of length 1")
})

list_xml <- function(x) as_list(read_xml(x))

test_that("empty elements become empty lists", {
  expect_equal(list_xml("<x></x>"), list(x = list()))
  expect_equal(list_xml("<x><y/></x>"), list(x = list(y = list())))
  expect_equal(
    list_xml("<x><y><z/></y></x>"),
    list(x = list(y = list(z = list())))
  )
})

test_that("text nodes become character vectors", {
  expect_equal(list_xml("<x>a</x>"), list(x = list("a")))
  expect_equal(list_xml("<x><y>a</y></x>"), list(x = list(y = list("a"))))
})

test_that("cdata nodes become character vectors", {
  expect_equal(list_xml("<x><![CDATA[<y/>]]></x>"), list(x = list("<y/>")))
})

test_that("xml attributes become R attibutes", {
  expect_equal(
    list_xml("<x a='1' b='2'></x>"),
    list(x = structure(list(), a = "1", b = "2"))
  )
})

test_that("xml names are preserved when attributes exist", {
  expect_equal(
    list_xml("<x a='1' b='2'><y>3</y><z>4</z></x>"),
    list(x = structure(list(y = list("3"), z = list("4")), a = "1", b = "2"))
  )
})

test_that("special attributes are escaped", {
  expect_equal(
    list_xml("<x a='1' b='2' names='esc'><y>3</y><z>4</z></x>"),
    list(
      x = structure(
        list(y = list("3"), z = list("4")),
        a = "1",
        b = "2",
        .names = "esc"
      )
    )
  )
})

test_that("attributes in child nodes", {
  expect_equal(
    list_xml("<w aa = '0'><x a='1' b='2' names='esc'><y>3</y><z>4</z></x></w>"),
    list(
      w = structure(
        list(
          x = structure(
            list(y = list("3"), z = list("4")),
            a = "1",
            b = "2",
            .names = "esc"
          )
        ),
        aa = "0"
      )
    )
  )
})

#' Construct an missing xml object
#' @export
#' @keywords internal
xml_missing <- function() {
  out <- list()
  class(out) <- "xml_missing"
  out
}

format.xml_missing <- function(x, ...) {
  "<NA>"
}

#' @export
print.xml_missing <- function(x, width = getOption("width"), max_n = 20, ...) {
  cat("{xml_missing}\n")
  cat(format(x), "\n", sep = "")
}

#' @export
as.character.xml_missing <- function(x, ...) {
  NA_character_
}

# These mimic the behavior of NA[[1]], NA[[2]], NA[1], NA[2]

#' @export
`[.xml_missing` <- function(x, i, ...) x

#' @export
`[[.xml_missing` <- function(x, i, ...) {
  if (i == 1L) x else cli::cli_abort("subscript out of bounds")
}

#' @export
is.na.xml_missing <- function(x) {
  TRUE
}

#' @export
format.xml_node <- function(x, ...) {
  attrs <- xml_attrs(x)
  paste0(
    "<",
    paste(
      c(
        xml_name(x),
        format_attributes(attrs)
      ),
      collapse = " "
    ),
    ">"
  )
}

format_attributes <- function(x) {
  if (length(x) == 0) {
    character(0)
  } else {
    paste(names(x), quote_str(x), sep = "=")
  }
}

xml_nodeset <- function(nodes = list(), deduplicate = TRUE) {
  if (isTRUE(deduplicate)) {
    nodes <- nodes[!.Call(nodes_duplicated, nodes)]
  }
  class(nodes) <- "xml_nodeset"
  nodes
}

#' @param nodes A list (possible nested) of external pointers to nodes
#' @return a nodeset
#' @noRd
make_nodeset <- function(nodes, doc) {
  nodes <- unlist(nodes, recursive = FALSE)

  xml_nodeset(lapply(nodes, xml_node, doc = doc))
}

#' @export
print.xml_nodeset <- function(x, width = getOption("width"), max_n = 20, ...) {
  n <- length(x)
  cat("{", doc_type(x), "_nodeset (", n, ")}\n", sep = "")

  if (n > 0) {
    show_nodes(x, width = width, max_n = max_n)
  }
}

#' @export
as.character.xml_nodeset <- function(x, ...) {
  vapply(x, as.character, FUN.VALUE = character(1))
}

#' @export
`[.xml_nodeset` <- function(x, i, ...) {
  if (length(x) == 0) {
    return(x)
  }
  xml_nodeset(NextMethod())
}

#' Wrapper for encodeString() that takes width into consideration
#'
#' encodeString() is relatively expensive to run (see #366), so
#'   avoid doing so to very wide inputs by first trimming inputs
#'   to approximately the correct width, then encoding. A second
#'   round of truncation occurs after encoding to account for
#'   any newly-inserted characters bumping an input too wide.
#' @noRd
encode_with_width <- function(x, width) {
  truncate_raw <- nchar(x) > width
  x[truncate_raw] <- substr(x[truncate_raw], 1L, width - 3L)
  x <- encodeString(x)
  truncate_encoded <- truncate_raw | nchar(x) > width
  x[truncate_encoded] <- paste(
    substr(x[truncate_encoded], 1L, width - 3L),
    "..."
  )
  x
}

show_nodes <- function(x, width = getOption("width"), max_n = 20) {
  stopifnot(inherits(x, "xml_nodeset"))

  n <- length(x)
  if (n == 0) {
    return()
  }

  trunc <- n > max_n
  if (trunc) {
    n <- max_n
    x <- x[seq_len(n)]
  }

  label <- format(paste0("[", seq_len(n), "]"), justify = "right")
  contents <- vapply(x, as.character, FUN.VALUE = character(1L))

  desc <- encode_with_width(paste(label, contents), width)

  cat(desc, sep = "\n")
  if (trunc) {
    cat("...\n")
  }
  invisible()
}

#' @export
is.na.xml_nodeset <- function(x) {
  vapply(x, is.na, logical(1))
}

#' Read HTML or XML.
#'
#' @section Setting the "user agent" header:
#'
#' When performing web scraping tasks it is both good practice --- and often required ---
#' to set the [user agent](https://en.wikipedia.org/wiki/User_agent) request header
#' to a specific value. Sometimes this value is assigned to emulate a browser in order
#' to have content render in a certain way (e.g. `Mozilla/5.0 (Windows NT 5.1; rv:52.0)
#' Gecko/20100101 Firefox/52.0` to emulate more recent Windows browsers). Most often,
#' this value should be set to provide the web resource owner information on who you are
#' and the intent of your actions like this Google scraping bot user agent identifier:
#' `Googlebot/2.1 (+http://www.google.com/bot.html)`.
#'
#' You can set the HTTP user agent for URL-based requests using [httr::set_config()] and [httr::user_agent()]:
#'
#' `httr::set_config(httr::user_agent("me@@example.com; +https://example.com/info.html"))`
#'
#' [httr::set_config()] changes the configuration globally,
#' [httr::with_config()] can be used to change configuration temporarily.
#'
#' @param x A string, a connection, or a raw vector.
#'
#'   A string can be either a path, a url or literal xml. Urls will
#'   be converted into connections either using `base::url` or, if
#'   installed, `curl::curl`. Local paths ending in `.gz`,
#'   `.bz2`, `.xz`, `.zip` will be automatically uncompressed.
#'
#'   If a connection, the complete connection is read into a raw vector before
#'   being parsed.
#' @param encoding Specify a default encoding for the document. Unless
#'   otherwise specified XML documents are assumed to be in UTF-8 or
#'   UTF-16. If the document is not UTF-8/16, and lacks an explicit
#'   encoding directive, this allows you to supply a default.
#' @param ... Additional arguments passed on to methods.
#' @param as_html Optionally parse an xml file as if it's html.
#' @param base_url When loading from a connection, raw vector or literal
#'   html/xml, this allows you to specify a base url for the document. Base
#'   urls are used to turn relative urls into absolute urls.
#' @param n If `file` is a connection, the number of bytes to read per
#'    iteration. Defaults to 64kb.
#' @param verbose When reading from a slow connection, this prints some
#'    output on every iteration so you know its working.
#' @param options Set parsing options for the libxml2 parser. Zero or more of
#' \Sexpr[results=rd, stage=build]{xml2:::describe_options(xml2:::xml_parse_options())}
#' @return An XML document. HTML is normalised to valid XML - this may not
#'   be exactly the same transformation performed by the browser, but it's
#'   a reasonable approximation.
#' @export
#' @examples
#' # Literal xml/html is useful for small examples
#' read_xml("<foo><bar /></foo>")
#' read_html("<html><title>Hi<title></html>")
#' read_html("<html><title>Hi")
#'
#' # From a local path
#' read_html(system.file("extdata", "r-project.html", package = "xml2"))
#'
#' \dontrun{
#' # From a url
#' cd <- read_xml(xml2_example("cd_catalog.xml"))
#' me <- read_html("http://had.co.nz")
#' }
read_xml <- function(
  x,
  encoding = "",
  ...,
  as_html = FALSE,
  options = "NOBLANKS"
) {
  UseMethod("read_xml")
}

#' @export
#' @rdname read_xml
read_html <- function(
  x,
  encoding = "",
  ...,
  options = c("RECOVER", "NOERROR", "NOBLANKS", "HUGE")
) {
  UseMethod("read_html")
}

#' @export
read_html.default <- function(
  x,
  encoding = "",
  ...,
  options = c("RECOVER", "NOERROR", "NOBLANKS", "HUGE")
) {
  options <- parse_options(options, xml_parse_options())

  suppressWarnings(read_xml(
    x,
    encoding = encoding,
    ...,
    as_html = TRUE,
    options = options
  ))
}

#' @export
read_html.response <- function(
  x,
  encoding = "",
  options = c("RECOVER", "NOERROR", "NOBLANKS"),
  ...
) {
  check_installed("httr")

  options <- parse_options(options, xml_parse_options())

  httr::stop_for_status(x)
  content <- httr::content(x, as = "raw")
  xml2::read_html(content, encoding = encoding, options = options, ...)
}

#' @export
#' @rdname read_xml
read_xml.character <- function(
  x,
  encoding = "",
  ...,
  as_html = FALSE,
  options = "NOBLANKS"
) {
  check_string(x)

  options <- parse_options(options, xml_parse_options())
  if (grepl("<|>", x)) {
    read_xml.raw(
      charToRaw(enc2utf8(x)),
      "UTF-8",
      ...,
      as_html = as_html,
      options = options
    )
  } else {
    con <- path_to_connection(x)
    if (inherits(con, "connection")) {
      read_xml.connection(
        con,
        encoding = encoding,
        ...,
        as_html = as_html,
        base_url = x,
        options = options
      )
    } else {
      doc <- .Call(
        doc_parse_file,
        con,
        encoding = encoding,
        as_html = as_html,
        options = options
      )
      xml_document(doc)
    }
  }
}

#' @export
#' @rdname read_xml
read_xml.raw <- function(
  x,
  encoding = "",
  base_url = "",
  ...,
  as_html = FALSE,
  options = "NOBLANKS"
) {
  options <- parse_options(options, xml_parse_options())

  doc <- .Call(doc_parse_raw, x, encoding, base_url, as_html, options)
  xml_document(doc)
}

#' @export
#' @rdname read_xml
read_xml.connection <- function(
  x,
  encoding = "",
  n = 64 * 1024,
  verbose = FALSE,
  ...,
  base_url = "",
  as_html = FALSE,
  options = "NOBLANKS"
) {
  options <- parse_options(options, xml_parse_options())

  if (!isOpen(x)) {
    open(x, "rb")
    on.exit(close(x))
  }

  raw <- .Call(read_connection_, x, n)
  read_xml.raw(
    raw,
    encoding = encoding,
    base_url = base_url,
    as_html = as_html,
    options = options
  )
}

#' @export
read_xml.response <- function(
  x,
  encoding = "",
  base_url = "",
  ...,
  as_html = FALSE,
  options = "NOBLANKS"
) {
  check_installed("httr")

  options <- parse_options(options, xml_parse_options())
  httr::stop_for_status(x)
  content <- httr::content(x, as = "raw")
  xml2::read_xml(
    content,
    encoding = encoding,
    base_url = if (nzchar(base_url)) base_url else x$url,
    as_html = as_html,
    option = options,
    ...
  )
}

#' @export
read_xml.textConnection <- function(x, encoding = "", ...) {
  s <- paste(readLines(x), collapse = "\n")
  read_xml.character(s, ...)
}

#' Download a HTML or XML file
#'
#' Libcurl implementation of `C_download` (the "internal" download method)
#' with added support for https, ftps, gzip, etc. Default behavior is identical
#' to [download.file()], but request can be fully configured by passing
#' a custom [curl::handle()].
#' @inherit curl::curl_download
#' @param file A character string with the name where the downloaded file is
#'   saved.
#' @seealso [curl_download][curl::curl_download]
#' @export
#' @examples
#' \dontrun{
#' download_html("http://tidyverse.org/index.html")
#' }
download_xml <- function(
  url,
  file = basename(url),
  quiet = TRUE,
  mode = "wb",
  handle = curl::new_handle()
) {
  check_installed("curl", "to use `download_xml()`.")

  curl::curl_download(url, file, quiet = quiet, mode = mode, handle = handle)

  invisible(file)
}

#' @export
#' @rdname download_xml
download_html <- download_xml

test_that("xml schema validates", {
  doc <- read_xml(system.file("extdata/order-doc.xml", package = "xml2"))
  schema <- read_xml(system.file("extdata/order-schema.xml", package = "xml2"))
  expect_true(xml_validate(doc, schema))
})

test_that("xml schema errors", {
  str <- readLines(system.file("extdata/order-doc.xml", package = "xml2"))
  str <- sub("<quantity>1", "<quantity>", str)
  str <- sub("95819", "ABC95819", str)
  str <- sub('partNum="926-AA"', "", str)
  doc <- read_xml(paste(str, collapse = "\n"))
  schema <- read_xml(system.file("extdata/order-schema.xml", package = "xml2"))
  out <- xml_validate(doc, schema)
  expect_false(out)
  errors <- attr(out, "errors")
  expect_type(errors, "character")
  expect_length(errors, 4)
})

x <- read_xml("<a>
  <b><c>123</c></b>
  <b><c>456</c></b>
  </a>")

test_that("xml_serialize and xml_unserialize work with xml_document input", {
  out <- xml_unserialize(xml_serialize(x, NULL))
  expect_identical(as.character(x), as.character(out))

  f <- tempfile()
  on.exit(unlink(f))

  xml_serialize(x, f)
  expect_identical(as.character(xml_unserialize(f)), as.character(x))
})

test_that("xml_serialize and xml_unserialize work with xml_node input", {
  b <- xml_find_first(x, "//b")
  out <- xml_unserialize(xml_serialize(b, NULL))
  expect_identical(as.character(b), as.character(out))

  f <- tempfile()
  on.exit(unlink(f))

  xml_serialize(b, f)
  expect_identical(as.character(xml_unserialize(f)), as.character(b))
})

test_that("xml_serialize and xml_unserialize work with xml_nodeset input", {
  b <- xml_find_all(x, "//b")
  out <- xml_unserialize(xml_serialize(b, NULL))
  expect_identical(as.character(b), as.character(out))

  f <- tempfile()
  on.exit(unlink(f))

  xml_serialize(b, f)
  expect_identical(as.character(xml_unserialize(f)), as.character(b))
})

test_that("xml_serialize and xml_unserialize work with HTML-based xml_document input", {
  file <- system.file("extdata", "r-project.html", package = "xml2")
  x <- read_html(file)

  out <- xml_unserialize(xml_serialize(x, NULL))
  expect_identical(as.character(x), as.character(out))

  f <- tempfile()
  on.exit(unlink(f))

  xml_serialize(x, f)
  expect_identical(as.character(xml_unserialize(f)), as.character(x))
})

test_that("xml_unserialize throws an error if given a invalid object", {
  expect_error(xml_unserialize(serialize(1, NULL)), "Not a serialized xml2 object")
})

test_that("xml_type() works", {
  x <- read_xml("<body>
    <p>Some <b>text</b>.</p>
    <p>Some <i>other</i>.</p>
    <p>No bold text</p>
    </body>")

  children <- xml_children(x)
  x <- xml_find_first(children, ".//b|.//i")

  expect_equal(xml_type(x[[1]]), "element")
  expect_equal(xml_type(x[[3]]), NA_character_)

  expect_equal(xml_type(x), c("element", "element", NA))

  empty <- xml_children(x)
  expect_identical(xml_type(empty), character())
})

test_that("format.xml_node prints attributes for root nodes", {
  x <- read_xml("<parent foo = 'bar' />")
  expect_equal(format(x), "<parent foo=\"bar\">")
})
test_that("format.xml_node prints namespaces for root nodes", {
  x <- read_xml("<parent/>")
  expect_equal(format(x), "<parent>")

  y <- read_xml("<parent xmlns = 'tag:james.f.hester@gmail.com,2016:bar' />")
  expect_equal(format(y), "<parent xmlns=\"tag:james.f.hester@gmail.com,2016:bar\">")

  z <- read_xml("<parent xmlns:foo = 'tag:james.f.hester@gmail.com,2016:bar' />")
  expect_equal(format(z), "<parent xmlns:foo=\"tag:james.f.hester@gmail.com,2016:bar\">")
})

test_that("write_xml errors for incorrect directory and with invalid inputs", {
  x <- read_xml("<x/>")
  filename <- "does_not_exist/test.xml"
  expect_error(write_xml(x, filename), "'does_not_exist' does not exist in current working directory")

  expect_snapshot(error = TRUE, {
    write_xml(x, c("test.xml", "foo"))
  })
})

test_that("write_xml works with relative file paths", {
  x <- read_xml("<x/>")

  filename <- "../test.xml"
  on.exit(unlink(filename))
  write_xml(x, filename, options = "no_declaration")
  expect_identical(readChar(filename, 1000L), "<x/>\n")
})

test_that("write_xml works with no options", {
  x <- read_xml("<x/>")

  filename <- "../test.xml"
  on.exit(unlink(filename))
  write_xml(x, filename, options = NULL)
  expect_identical(readChar(filename, 1000L), "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<x/>\n")
})

test_that("write_xml works with an explicit connections", {
  x <- read_xml("<x/>")

  filename <- "../test.xml"
  file <- file(filename, "wb")
  on.exit(unlink(filename))
  write_xml(x, file, options = "no_declaration")
  close(file)
  expect_identical(readChar(filename, 1000L), "<x/>\n")
})

test_that("write_xml works with an implicit connections", {
  x <- read_xml("<x/>")

  filename <- "../test.xml.gz"
  write_xml(x, filename, options = "no_declaration")
  file <- gzfile(filename, "rb")
  on.exit({
    unlink(filename)
    close(file)
  })
  expect_identical(readChar(file, 1000L), "<x/>\n")
})

test_that("write_xml works with nodeset input and files", {
  x <- read_xml("<x><y/><y><z/></y></x>")
  y <- xml_find_all(x, "//y")

  filename <- "../test.xml"
  on.exit(unlink(filename))
  expect_error(
    write_xml(y, filename, options = "no_declaration"),
    "Can only save length 1 node sets"
  )

  write_xml(y[1], filename, options = "no_declaration")
  expect_identical(readChar(filename, 1000L), "<y/>")
})

test_that("write_xml works with nodeset input and connections", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_all(x, "//y")

  filename <- "../test.xml.gz"
  expect_error(
    write_xml(y, filename, options = "no_declaration"),
    "Can only save length 1 node sets"
  )

  expect_snapshot(error = TRUE, {
    write_xml(y[1], c(filename, "foo"))
  })

  write_xml(y[1], filename, options = "no_declaration")
  file <- gzfile(filename, "rb")
  on.exit({
    unlink(filename)
    close(file)
  })
  expect_identical(readChar(file, 1000L), "<y/>")
})

test_that("write_xml works with node input and files", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_first(x, "//y")

  filename <- "../test.xml"
  expect_snapshot(error = TRUE, write_xml(y, c(filename, "foo")))

  write_xml(y, filename, options = "no_declaration")
  on.exit(unlink(filename))
  expect_identical(readChar(filename, 1000L), "<y/>")
})

test_that("write_xml works with node input and connections", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_first(x, "//y")

  filename <- "../test.xml.gz"
  write_xml(y, filename, options = "no_declaration")
  file <- gzfile(filename, "rb")
  on.exit({
    unlink(filename)
    close(file)
  })
  expect_identical(readChar(file, 1000L), "<y/>")
})

test_that("write_html work with html input", {
  x <- read_html("<html><title>Foo</title></html>")

  filename <- "../test.html.gz"
  write_html(x, filename)
  file <- gzfile(filename, "rb")
  on.exit({
    unlink(filename)
    close(file)
  })
  expect_identical(
    readChar(file, 1000L),
    "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\" \"http://www.w3.org/TR/REC-html40/loose.dtd\">\n<html><head>\n<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">\n<title>Foo</title>\n</head></html>\n"
  )
})

test_that("write_xml returns invisibly", {
  x <- read_xml("<x>foo</x>")
  tf <- tempfile()
  on.exit(unlink(tf))

  res <- withVisible(write_xml(x, tf))

  expect_null(res$value)
  expect_false(res$visible)
})

test_that("download_xml fails if curl is not installed", {
  skip("how to test error with `check_installed()`?")
  mockery::stub(download_xml, "requireNamespace", function(...) FALSE)

  expect_error(
    download_xml("http://httpbin.org/xml"),
    "`curl` must be installed to use `download_xml\\(\\)`"
  )
})

test_that("read_xml errors with an empty document", {
  expect_snapshot(error = TRUE, {
    read_xml(character())
  })

  tf <- tempfile()
  file.create(tf)
  on.exit(unlink(tf))

  expect_error(read_xml(tf), "Document is empty")
})

test_that("read_html correctly parses malformed document", {
  lego <- read_html(test_path("lego.html.bz2"))
  expect_length(xml_find_all(lego, ".//p"), 39)
})

test_that("parse_options errors when given an invalid option", {
  expect_error(
    parse_options("INVALID", xml_parse_options()),
    '`options` "INVALID" is not a valid option'
  )

  expect_snapshot(error = TRUE,
    read_html(test_path("lego.html.bz2"), options = "INVALID")
  )

  # Empty inputs returned as 0
  expect_identical(0L, parse_options("", xml_parse_options()))
  expect_identical(0L, parse_options(NULL, xml_parse_options()))

  # Numerics returned as integers
  expect_identical(12L, parse_options(12L, xml_parse_options()))
  expect_identical(12L, parse_options(12, xml_parse_options()))

  # Multiple inputs summed
  expect_identical(3L, parse_options(c("RECOVER", "NOENT"), xml_parse_options()))
})

test_that("read_html properly passes parser arguments", {
  skip_if_not(libxml2_version() >= "2.9.2")

  blanks <- read_html(xml2_example("cd_catalog.xml"), options = c("RECOVER", "NOERROR"))
  expect_equal(
    as_list(blanks)$html$body$catalog$cd[[1]],
    "\r\n    "
  )

  no_blanks <- read_html(xml2_example("cd_catalog.xml"), options = c("RECOVER", "NOERROR", "NOBLANKS"))

  expect_equal(
    as_list(no_blanks)$html$body$catalog$cd[[1]],
    list("Empire Burlesque")
  )
})

test_that("read_xml works with httr response objects", {
  skip("httpbin is unreliable")
  x <- read_xml(httr::GET("http://httpbin.org/xml"))

  expect_s3_class(x, "xml_document")

  expect_length(xml_find_all(x, "//slide"), 2)
})

test_that("read_xml and read_html fail for bad status codes", {
  skip("httpbin is unreliable")

  expect_error(
    read_xml(httr::GET("http://httpbin.org/status/404")),
    class = "http_404"
  )

  expect_error(
    read_html(httr::GET("http://httpbin.org/status/404")),
    class = "http_404"
  )
})

test_that("read_xml works with raw inputs", {
  x <- read_xml("<foo/>")
  expect_equal(xml_url(x), NA_character_)
})

test_that("read_html works with non-ASCII encodings", {
  tmp <- tempfile()
  on.exit(unlink(tmp))

  writeLines("<html><body>\U2019</body></html>", tmp, useBytes = TRUE)
  res <- read_html(tmp, encoding = "UTF-8")

  expect_equal(
    as.character(res, options = ""),
    "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\" \"http://www.w3.org/TR/REC-html40/loose.dtd\">\n<html><body>\U2019</body></html>\n"
  )
})

test_that("read_xml and read_html fail with > 1 input", {
  expect_snapshot(error = TRUE, {
    read_xml(c("foo", "bar"))
    read_html(c("foo", "bar"))
  })
})

test_that("xml_structure", {
  expect_output(
    xml_structure(read_xml("<a><b><c/><c/></b><d/></a>")),
    "<a>
  <b>
    <c>
    <c>
  <d>"
  )

  expect_output(
    xml_structure(read_xml("<a><b><c/><c/></b><d/></a>"), indent = 0L),
    "<a>
<b>
<c>
<c>
<d>"
  )
})

test_that("xml_structure can write to a file (#244)", {
  tmp <- tempfile()
  xml_structure(read_xml("<a><b><c/><c/></b><d/></a>"), file = tmp)
  expect_equal(readLines(tmp), c("<a>", "  <b>", "    <c>", "    <c>", "  <d>"))

  # repeated calls erase existing content
  xml_structure(read_xml("<a><b><c/><c/></b><d/></a>"), file = tmp)
  expect_equal(readLines(tmp), c("<a>", "  <b>", "    <c>", "    <c>", "  <d>"))
})

test_that("xml_structure is correct", {
  x <- read_html(test_path("lego.html.bz2"))

  quicklinks <- xml_find_first(x, "//div[contains(@div, 'quicklinks')]")
  expect_snapshot(html_structure(quicklinks))
})

context("write_xml")

test_that("write_xml errors for incorrect directory and with invalid inputs", {
  x <- read_xml("<x/>")
  filename <- ".../test.xml"
  expect_error(write_xml(x, filename), "'...' does not exist in current working directory")


  expect_error(write_xml(x, c("test.xml", "foo")), "`file` must be a non-zero character of length 1")
})

test_that("write_xml works with relative file paths", {
  x <- read_xml("<x/>")

  filename <- "../test.xml"
  on.exit(unlink(filename))
  write_xml(x, filename, options = "no_declaration")
  expect_identical(readChar(filename, 1000L), "<x/>\n")
})

test_that("write_xml works with no options", {
  x <- read_xml("<x/>")

  filename <- "../test.xml"
  on.exit(unlink(filename))
  write_xml(x, filename, options = NULL)
  expect_identical(readChar(filename, 1000L), "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<x/>\n")
})

test_that("write_xml works with an explicit connections", {
  x <- read_xml("<x/>")

  filename <- "../test.xml"
  file <- file(filename, "wb")
  on.exit(unlink(filename))
  write_xml(x, file, options = "no_declaration")
  close(file)
  expect_identical(readChar(filename, 1000L), "<x/>\n")
})

test_that("write_xml works with an implicit connections", {
  x <- read_xml("<x/>")

  filename <- "../test.xml.gz"
  write_xml(x, filename, options = "no_declaration")
  file <- gzfile(filename, "rb")
  on.exit({unlink(filename); close(file)})
  expect_identical(readChar(file, 1000L), "<x/>\n")
})

test_that("write_xml works with nodeset input and files", {
  x <- read_xml("<x><y/><y><z/></y></x>")
  y <- xml_find_all(x, "//y")

  filename <- "../test.xml"
  on.exit(unlink(filename))
  expect_error(write_xml(y, filename, options = "no_declaration"),
    "Can only save length 1 node sets")

  write_xml(y[1], filename, options = "no_declaration")
  expect_identical(readChar(filename, 1000L), "<y/>")
})

test_that("write_xml works with nodeset input and connections", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_all(x, "//y")

  filename <- "../test.xml.gz"
  expect_error(write_xml(y, filename, options = "no_declaration"),
    "Can only save length 1 node sets")

  expect_error(write_xml(y[1], c(filename, "foo")), "`file` must be a non-zero character of length 1")

  write_xml(y[1], filename, options = "no_declaration")
  file <- gzfile(filename, "rb")
  on.exit({unlink(filename); close(file)})
  expect_identical(readChar(file, 1000L), "<y/>")
})

test_that("write_xml works with node input and files", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_first(x, "//y")

  filename <- "../test.xml"
  expect_error(write_xml(y, c(filename, "foo")), "`file` must be a non-zero character of length 1")

  write_xml(y, filename, options = "no_declaration")
  on.exit(unlink(filename))
  expect_identical(readChar(filename, 1000L), "<y/>")
})

test_that("write_xml works with node input and connections", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_first(x, "//y")

  filename <- "../test.xml.gz"
  write_xml(y, filename, options = "no_declaration")
  file <- gzfile(filename, "rb")
  on.exit({unlink(filename); close(file)})
  expect_identical(readChar(file, 1000L), "<y/>")
})

test_that("write_html work with html input", {
  x <- read_html("<html><title>Foo</title></html>")

  filename <- "../test.html.gz"
  write_html(x, filename)
  file <- gzfile(filename, "rb")
  on.exit({unlink(filename); close(file)})
  expect_identical(readChar(file, 1000L),
    "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\" \"http://www.w3.org/TR/REC-html40/loose.dtd\">\n<html><head>\n<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">\n<title>Foo</title>\n</head></html>\n")
})

context("read_xml")

test_that("read_html correctly parses malformed document", {
  lego <- read_html("lego.html.bz2")
  expect_equal(length(xml_find_all(lego, ".//p")), 39)
})

test_that("parse_options errors when given an invalid option", {
  expect_error(parse_options("INVALID", xml_parse_options()),
    "`options` 'INVALID' is not a valid option")

  expect_error(read_html("lego.html.bz2", options = "INVALID"),
    "`options` 'INVALID' is not a valid option")

  # Empty inputs returned as 0
  expect_identical(0L, parse_options("", xml_parse_options()))
  expect_identical(0L, parse_options(NULL, xml_parse_options()))

  # Numerics returned as integers
  expect_identical(12L, parse_options(12L, xml_parse_options()))
  expect_identical(12L, parse_options(12, xml_parse_options()))

  # Multiple inputs summed
  expect_identical(3L, parse_options(c("RECOVER", "NOENT"), xml_parse_options()))
})

test_that("read_html properly passes parser arguments", {

  skip_if_not(libxml2_version() >= "2.9.2")

  blanks <- read_html(xml2_example("cd_catalog.xml"), options = c("RECOVER", "NOERROR"))
  expect_equal(as_list(blanks)$html$body$catalog$cd[[1]],
    "\r\n    ")

  no_blanks <- read_html(xml2_example("cd_catalog.xml"), options = c("RECOVER", "NOERROR", "NOBLANKS"))

  expect_equal(as_list(no_blanks)$html$body$catalog$cd[[1]],
    list("Empire Burlesque"))
})

test_that("read_xml works with httr response objects", {
  skip_on_cran()

  x <- read_xml(httr::GET("http://httpbin.org/xml"))
  expect_is(x, "xml_document")

  expect_equal(length(xml_find_all(x, "//slide")), 2)
})

test_that("read_html works with httr response objects", {
  skip_on_cran()

  x <- read_html(httr::GET("http://httpbin.org/xml"))
  expect_is(x, "xml_document")

  expect_equal(length(xml_find_all(x, "//slide")), 2)
})

test_that("download_xml fails if curl is not installed", {
  skip("how to test error with `check_installed()`?")
  mockery::stub(download_xml, "requireNamespace", function(...) FALSE)

  expect_error(
    download_xml("http://httpbin.org/xml"),
    "`curl` must be installed to use `download_xml\\(\\)`"
  )
})

test_that("read_xml errors with an empty document", {
  expect_snapshot(error = TRUE, {
    read_xml(character())
  })

  tf <- tempfile()
  file.create(tf)
  on.exit(unlink(tf))

  expect_error(read_xml(tf), "Document is empty")
})

test_that("read_html correctly parses malformed document", {
  lego <- read_html(test_path("lego.html.bz2"))
  expect_length(xml_find_all(lego, ".//p"), 39)
})

test_that("parse_options errors when given an invalid option", {
  expect_error(
    parse_options("INVALID", xml_parse_options()),
    '`options` "INVALID" is not a valid option'
  )

  expect_snapshot(
    error = TRUE,
    read_html(test_path("lego.html.bz2"), options = "INVALID")
  )

  # Empty inputs returned as 0
  expect_identical(0L, parse_options("", xml_parse_options()))
  expect_identical(0L, parse_options(NULL, xml_parse_options()))

  # Numerics returned as integers
  expect_identical(12L, parse_options(12L, xml_parse_options()))
  expect_identical(12L, parse_options(12, xml_parse_options()))

  # Multiple inputs summed
  expect_identical(
    3L,
    parse_options(c("RECOVER", "NOENT"), xml_parse_options())
  )
})

test_that("read_html properly passes parser arguments", {
  skip_if_not(libxml2_version() >= "2.9.2")

  blanks <- read_html(
    xml2_example("cd_catalog.xml"),
    options = c("RECOVER", "NOERROR")
  )
  expect_equal(
    sub("\r\n", "\n", fixed = TRUE, as_list(blanks)$html$body$catalog$cd[[1]]),
    "\n    "
  )

  no_blanks <- read_html(
    xml2_example("cd_catalog.xml"),
    options = c("RECOVER", "NOERROR", "NOBLANKS")
  )

  expect_equal(
    as_list(no_blanks)$html$body$catalog$cd[[1]],
    list("Empire Burlesque")
  )
})

test_that("read_xml works with httr response objects", {
  skip("httpbin is unreliable")
  x <- read_xml(httr::GET("http://httpbin.org/xml"))

  expect_s3_class(x, "xml_document")

  expect_length(xml_find_all(x, "//slide"), 2)
})

test_that("read_xml and read_html fail for bad status codes", {
  skip("httpbin is unreliable")

  expect_error(
    read_xml(httr::GET("http://httpbin.org/status/404")),
    class = "http_404"
  )

  expect_error(
    read_html(httr::GET("http://httpbin.org/status/404")),
    class = "http_404"
  )
})

test_that("read_xml works with raw inputs", {
  x <- read_xml("<foo/>")
  expect_equal(xml_url(x), NA_character_)
})

test_that("read_html works with non-ASCII encodings", {
  tmp <- tempfile()
  on.exit(unlink(tmp))

  writeLines("<html><body>\U2019</body></html>", tmp, useBytes = TRUE)
  res <- read_html(tmp, encoding = "UTF-8")

  expect_equal(
    as.character(res, options = ""),
    "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\" \"http://www.w3.org/TR/REC-html40/loose.dtd\">\n<html><body>\U2019</body></html>\n"
  )
})

test_that("read_xml and read_html fail with > 1 input", {
  expect_snapshot(error = TRUE, {
    read_xml(c("foo", "bar"))
    read_html(c("foo", "bar"))
  })
})

# Fails in libxml2 2.12
#test_that("Truncated HTML should not error", {
#  res <- read_html('<html><head')
#  expect_s3_class(res, "xml_document")
#})

test_that("read_xml from a textConnection", {
  s <- '<?xml version="1.0" encoding="UTF-8"?>\n<outer>\n<inner>Inner</inner>\n</outer>'
  con <- textConnection(s)
  x <- read_xml(con)
  close(con)
  expect_s3_class(x, "xml_document")
})

context("xml_schema")

test_that("xml schema validates", {
  doc <- read_xml(system.file("extdata/order-doc.xml", package = "xml2"))
  schema <- read_xml(system.file("extdata/order-schema.xml", package = "xml2"))
  expect_true(xml_validate(doc, schema))
})

test_that("xml schema errors", {
  str <- readLines(system.file("extdata/order-doc.xml", package = "xml2"))
  str <- sub("<quantity>1", "<quantity>", str)
  str <- sub("95819", "ABC95819", str)
  str <- sub('partNum="926-AA"', "", str)
  doc <- read_xml(paste(str, collapse = "\n"))
  schema <- read_xml(system.file("extdata/order-schema.xml", package = "xml2"))
  out <- xml_validate(doc, schema)
  expect_false(out)
  errors <- attr(out, "errors")
  expect_is(errors, "character")
  expect_length(errors, 4)
})

context("xml_structure")

test_that("xml_structure", {
  expect_output(xml_structure(read_xml("<a><b><c/><c/></b><d/></a>")),
"<a>
  <b>
    <c>
    <c>
  <d>")

  expect_output(xml_structure(read_xml("<a><b><c/><c/></b><d/></a>"), indent = 0L),
"<a>
<b>
<c>
<c>
<d>")
})

test_that("xml_structure can write to a file (#244)", {
  tmp = tempfile()
  xml_structure(read_xml("<a><b><c/><c/></b><d/></a>"), file = tmp)
  expect_equal(readLines(tmp), c("<a>", "  <b>", "    <c>", "    <c>", "  <d>"))

  # repeated calls erase existing content
  xml_structure(read_xml("<a><b><c/><c/></b><d/></a>"), file = tmp)
  expect_equal(readLines(tmp), c("<a>", "  <b>", "    <c>", "    <c>", "  <d>"))
})

test_that("xml_type() works", {
  x <- read_xml(
    "<body>
    <p>Some <b>text</b>.</p>
    <p>Some <i>other</i>.</p>
    <p>No bold text</p>
    </body>"
  )

  children <- xml_children(x)
  x <- xml_find_first(children, ".//b|.//i")

  expect_equal(xml_type(x[[1]]), "element")
  expect_equal(xml_type(x[[3]]), NA_character_)

  expect_equal(xml_type(x), c("element", "element", NA))

  empty <- xml_children(x)
  expect_identical(xml_type(empty), character())
})

test_that("format.xml_node prints attributes for root nodes", {
  x <- read_xml("<parent foo = 'bar' />")
  expect_equal(format(x), "<parent foo=\"bar\">")
})
test_that("format.xml_node prints namespaces for root nodes", {
  x <- read_xml("<parent/>")
  expect_equal(format(x), "<parent>")

  y <- read_xml("<parent xmlns = 'tag:james.f.hester@gmail.com,2016:bar' />")
  expect_equal(
    format(y),
    "<parent xmlns=\"tag:james.f.hester@gmail.com,2016:bar\">"
  )

  z <- read_xml(
    "<parent xmlns:foo = 'tag:james.f.hester@gmail.com,2016:bar' />"
  )
  expect_equal(
    format(z),
    "<parent xmlns:foo=\"tag:james.f.hester@gmail.com,2016:bar\">"
  )
})

`%||%` <- function(a, b) if (is.null(a)) b else a

is_named <- function(x) {
  all(has_names(x))
}

has_names <- function(x) {
  nms <- names(x)
  if (is.null(nms)) {
    rep(FALSE, length(x))
  } else {
    !(is.na(nms) | nms == "")
  }
}

# non smart quote version of sQuote
quote_str <- function(x, quote = "\"") {
  if (!length(x)) {
    return(character(0))
  }

  paste0(quote, x, quote)
}

# Format the C bitwise flags for display in Rd. The input object is a named
# integer vector with a 'descriptions' character vector attribute that
# corresponds to each flag.
describe_options <- function(x) {
  paste0(
    "\\describe{\n",
    paste0(
      "  \\item{",
      names(x),
      "}{",
      attr(x, "descriptions"),
      "}",
      collapse = "\n"
    ),
    "\n}"
  )
}

s_quote <- function(x) paste0("'", x, "'")

# Similar to match.arg, but returns character() with NULL or empty input and
# errors if any of the inputs are not found (fixing
# https://bugs.r-project.org/bugzilla3/show_bug.cgi?id=16659)
parse_options <- function(arg, options, error_call = caller_env()) {
  if (is.numeric(arg)) {
    return(as.integer(arg))
  }

  if (is.null(arg) || !any(nzchar(arg))) {
    return(0L)
  }

  # set duplicates.ok = TRUE so any duplicates are counted differently than
  # non-matches, then take only unique results
  i <- pmatch(arg, names(options), duplicates.ok = TRUE)
  if (anyNA(i)) {
    cli::cli_abort(
      c(
        x = "{.arg options} {.val {arg[is.na(i)][1L]}} is not a valid option.",
        i = "Valid options are one of {.or {.val {names(options)}}}.",
        i = "See {.help [read_html](xml2::read_html)} for all options."
      ),
      call = error_call
    )
  }
  sum(options[unique(i)])
}

#' Get path to a xml2 example
#'
#' xml2 comes bundled with a number of sample files in its \sQuote{inst/extdata}
#' directory. This function makes them easy to access.
#' @param path Name of file. If `NULL`, the example files will be listed.
#' @export
xml2_example <- function(path = NULL) {
  if (is.null(path)) {
    dir(system.file("extdata", package = "xml2"))
  } else {
    system.file("extdata", path, package = "xml2", mustWork = TRUE)
  }
}

#' Register S4 classes
#'
#' @description
#' Classes are exported so they can be re-used within S4 classes, see [methods::setOldClass()].
#'
#' * `xml_document`: a complete document.
#' * `xml_nodeset`: a _set_ of nodes within a document.
#' * `xml_missing`: a missing object, e.g. for an empty result set.
#' * `xml_node`: a single node in a document.
#'
#' @importFrom methods setOldClass
#' @keywords internal
#' @rdname oldclass
#' @name xml_document-class
#' @exportClass xml_document
setOldClass("xml_document")

#' @name xml_missing-class
#' @exportClass xml_missing
#' @rdname oldclass
setOldClass("xml_missing")

#' @name xml_node-class
#' @exportClass xml_node
#' @rdname oldclass
setOldClass("xml_node")

#' @name xml_nodeset-class
#' @exportClass xml_nodeset
#' @rdname oldclass
setOldClass("xml_nodeset")

#' Coerce a R list to xml nodes.
#'
#' This turns an R list into the equivalent XML document. Not all R lists will
#' produce valid XML, in particular there can only be one root node and all
#' child nodes need to be named (or empty) lists. R attributes become XML
#' attributes and R names become XML node names.
#'
#' @inheritParams as_list
#' @include as_list.R xml_parse.R
#' @export
#' @examples
# empty lists generate empty nodes
#'as_xml_document(list(x = list()))
#'
#'# Nesting multiple nodes
#'as_xml_document(list(foo = list(bar = list(baz = list()))))
#'
#'# attributes are stored as R attributes
#'as_xml_document(list(foo = structure(list(), id = "a")))
#'as_xml_document(list(foo = list(
#'      bar = structure(list(), id = "a"),
#'      bar = structure(list(), id = "b"))))
as_xml_document <- function(x, ...) {
  UseMethod("as_xml_document")
}

#' @export
as_xml_document.character <- read_xml.character

#' @export
as_xml_document.raw <- read_xml.raw

#' @export
as_xml_document.connection <- read_xml.connection

#' @export
as_xml_document.response <- read_xml.response

#' @export
as_xml_document.list <- function(x, ...) {
  if (length(x) > 1) {
    stop("Root nodes must be of length 1", call. = FALSE)
  }


  add_node <- function(x, parent, tag = NULL) {
    if (is.atomic(x)) {
      return(xml_set_text(parent, as.character(x)))
    }
    if (!is.null(tag)) {
      parent <- xml_add_child(parent, tag)
      attr <- r_attrs_to_xml(attributes(x))
      for (i in seq_along(attr)) {
        xml_set_attr(parent, names(attr)[[i]], attr[[i]])
      }
    }
    for (i in seq_along(x)) {
      add_node(x[[i]], parent, names(x)[[i]])
    }
  }

  doc <- xml_new_document()
  add_node(x, doc)
  xml_root(doc)
}

#' @export
as_xml_document.xml_node <- function(x, ...) {
  xml_new_root(.value = x, ..., .copy = TRUE)
}

#' @export
as_xml_document.xml_nodeset <- function(x, root, ...) {
  doc <- xml_new_root(.value = root, ..., .copy = TRUE)
  for (i in seq_along(x)) {
    xml_add_child(doc, x[[i]], .copy = TRUE)
  }
  doc
}

#' @export
as_xml_document.xml_document <- function(x, ...) {
  x
}

#' Validate XML schema
#'
#' Validate an XML document against an XML 1.0 schema.
#'
#' @inheritParams xml_name
#' @return TRUE or FALSE
#' @export
#' @param schema an XML document containing the schema
#' @examples # Example from https://msdn.microsoft.com/en-us/library/ms256129(v=vs.110).aspx
#' doc <- read_xml(system.file("extdata/order-doc.xml", package = "xml2"))
#' schema <- read_xml(system.file("extdata/order-schema.xml", package = "xml2"))
#' xml_validate(doc, schema)
xml_validate <- function(x, schema) {
  UseMethod("xml_validate")
}

#' @export
xml_validate.xml_document <- function(x, schema) {
  stopifnot(inherits(schema, "xml_document"))
  doc_validate(x$doc, schema$doc)
}

#' Navigate around the family tree.
#'
#' `xml_children` returns only elements, `xml_contents` returns
#' all nodes. `xml_length` returns the number of children.
#' `xml_parent` returns the parent node, `xml_parents`
#' returns all parents up to the root. `xml_siblings` returns all nodes
#' at the same level. `xml_child` makes it easy to specify a specific
#' child to return.
#'
#' @inheritParams xml_name
#' @param only_elements For `xml_length`, should it count all children,
#'   or just children that are elements (the default)?
#' @param search For `xml_child`, either the child number to return (by
#'   position), or the name of the child node to return. If there are multiple
#'   child nodes with the same name, the first will be returned
#' @return A node or nodeset (possibly empty). Results are always de-duplicated.
#' @export
#' @examples
#' x <- read_xml("<foo> <bar><boo /></bar> <baz/> </foo>")
#' xml_children(x)
#' xml_children(xml_children(x))
#' xml_siblings(xml_children(x)[[1]])
#'
#' # Note the each unique node only appears once in the output
#' xml_parent(xml_children(x))
#'
#' # Mixed content
#' x <- read_xml("<foo> a <b/> c <d>e</d> f</foo>")
#' # Childen gets the elements, contents gets all node types
#' xml_children(x)
#' xml_contents(x)
#'
#' xml_length(x)
#' xml_length(x, only_elements = FALSE)
#'
#' # xml_child makes it easier to select specific children
#' xml_child(x)
#' xml_child(x, 2)
#' xml_child(x, "baz")
xml_children <- function(x) {
  nodeset_apply(x, function(x) .Call(node_children, x, TRUE))
}

#' @export
#' @rdname xml_children
xml_child <- function(x, search = 1, ns = xml_ns(x)) {
  if (length(search) != 1) {
    cli::cli_abort("{.arg {search}} must be of length 1.")
  }

  if (is.numeric(search)) {
    xml_children(x)[[search]]
  } else if (is.character(search)) {
    xml_find_first(x, xpath = paste0("./", search), ns = ns)
  } else {
    cli::cli_abort("{.arg search} must be `numeric` or `character`.")
  }
}

#' @export
#' @rdname xml_children
xml_contents <- function(x) {
  nodeset_apply(x, function(x) .Call(node_children, x, FALSE))
}

#' @export
#' @rdname xml_children
xml_parents <- function(x) {
  nodeset_apply(x, function(x) .Call(node_parents, x))
}

#' @export
#' @rdname xml_children
xml_siblings <- function(x) {
  nodeset_apply(x, function(x) .Call(node_siblings, x, TRUE))
}

#' @export
#' @rdname xml_children
xml_parent <- function(x) {
  UseMethod("xml_parent")
}

#' @export
xml_parent.xml_missing <- function(x) {
  xml_missing()
}

#' @export
xml_parent.xml_node <- function(x) {
  xml_node(.Call(node_parent, x$node), x$doc)
}

#' @export
xml_parent.xml_nodeset <- function(x) {
  nodeset_apply(x, function(x) .Call(node_parent, x))
}


#' @export
#' @rdname xml_children
xml_length <- function(x, only_elements = TRUE) {
  .Call(node_length, x, only_elements)
}

#' @export
#' @rdname xml_children
xml_root <- function(x) {
  stopifnot(inherits(x, c("xml_node", "xml_document", "xml_nodeset")))

  if (inherits(x, "xml_nodeset")) {
    if (length(x) == 0) {
      return(NULL)
    } else {
      return(xml_root(x[[1]]))
    }
  }
  if (!.Call(doc_has_root, x$doc)) {
    xml_missing()
  } else {
    xml_document(x$doc)
  }
}

data <- read_xml(test_path("ns-multiple.xml"))
tf <- tempfile()
on.exit(unlink(tf))
saveRDS(data, file = tf)
x <- readRDS(tf)

test_that("accessors all fail rather than crash with NULL Xptrs", {
  expect_error(as_list(x), "external pointer is not valid")

  expect_error(html_structure(x), "external pointer is not valid")

  expect_error(xml_add_child(x, x), "external pointer is not valid")
  expect_error(xml_add_sibling(x, x), "external pointer is not valid")

  expect_error(xml_attr(x, "foo"), "external pointer is not valid")
  expect_error(xml_attr(x, "foo") <- "bar", "external pointer is not valid")

  expect_error(xml_attrs(x), "external pointer is not valid")
  expect_error(xml_attrs(x) <- list(), "external pointer is not valid")

  expect_error(xml_child(x), "external pointer is not valid")
  expect_error(xml_children(x), "external pointer is not valid")

  expect_error(xml_contents(x), "external pointer is not valid")

  expect_error(xml_double(x), "external pointer is not valid")

  expect_error(xml_find_all(x, ""), "external pointer is not valid")
  expect_error(xml_find_chr(x, ""), "external pointer is not valid")
  expect_error(xml_find_first(x, ""), "external pointer is not valid")
  expect_error(xml_find_lgl(x, ""), "external pointer is not valid")
  expect_error(xml_find_num(x, ""), "external pointer is not valid")

  expect_error(xml_has_attr(x, ""), "external pointer is not valid")

  expect_error(xml_integer(x), "external pointer is not valid")

  expect_error(xml_length(x), "external pointer is not valid")

  expect_error(xml_name(x), "external pointer is not valid")
  expect_error(xml_name(x) <- "foo", "external pointer is not valid")

  expect_error(xml_ns(x), "external pointer is not valid")
  expect_error(xml_ns_strip(x), "external pointer is not valid")

  expect_error(xml_parent(x), "external pointer is not valid")
  expect_error(xml_parents(x), "external pointer is not valid")

  expect_error(xml_path(x), "external pointer is not valid")

  expect_error(xml_remove(x), "external pointer is not valid")

  expect_error(xml_replace(x, x), "external pointer is not valid")

  expect_error(xml_set_namespace(x, "foo"), "external pointer is not valid")

  expect_error(xml_siblings(x), "external pointer is not valid")

  expect_error(xml_structure(x), "external pointer is not valid")

  expect_error(xml_text(x), "external pointer is not valid")
  expect_error(xml_text(x) <- "test", "external pointer is not valid")

  expect_error(xml_type(x), "external pointer is not valid")

  expect_error(xml_url(x), "external pointer is not valid")
})

x <- read_xml("<foo> <bar><boo /></bar> <baz/> </foo>")

test_that("xml_child() returns the proper child", {
  expect_identical(xml_child(x), xml_children(x)[[1L]])

  expect_identical(xml_child(x, 2), xml_children(x)[[2L]])
})

test_that("xml_child() returns child by name", {
  expect_identical(xml_child(x, "baz"), xml_find_first(x, "./baz"))
})

test_that("xml_child() errors if more than one search is given", {
  expect_snapshot(error = TRUE, xml_child(x, 1:2))
})

test_that("xml_child() errors if search is not numeric or character", {
  expect_error(xml_child(x, TRUE), "`search` must be `numeric` or `character`")
  expect_error(xml_child(x, as.factor("test")), "`search` must be `numeric` or `character`")
  expect_error(xml_child(x, raw(1)), "`search` must be `numeric` or `character`")
  expect_error(xml_child(x, list(1)), "`search` must be `numeric` or `character`")
})

test_that("xml_length", {
  expect_equal(xml_length(x), 2)
  all <- xml_find_all(x, "//*")
  expect_equal(xml_length(all), c(2, 1, 0, 0))
})

test_that("xml_parent", {
  expect_identical(unclass(xml_parent(xml_child(x))), unclass(x))
})

test_that("xml_parents", {
  expect_equal(
    xml_name(xml_parents(xml_find_first(x, "//boo"))),
    c("bar", "foo")
  )
})

test_that("xml_root", {
  doc <- xml_new_document()

  expect_s3_class(xml_root(doc), "xml_missing")

  a <- xml_add_child(doc, "a")
  b <- xml_add_child(doc, "b")

  expect_equal(xml_name(xml_root(b)), "a")
  expect_equal(xml_name(xml_root(doc)), "a")
})

context("Null XPtr")

data <- read_xml("ns-multiple.xml")
tf <- tempfile()
on.exit(unlink(tf))
saveRDS(data, file = tf)
x <- readRDS(tf)

test_that("accessors all fail rather than crash with NULL Xptrs", {

  expect_error(as_list(x), "external pointer is not valid")

  expect_error(html_structure(x), "external pointer is not valid")

  expect_error(xml_add_child(x, x), "external pointer is not valid")
  expect_error(xml_add_sibling(x, x), "external pointer is not valid")

  expect_error(xml_attr(x, "foo"), "external pointer is not valid")
  expect_error(xml_attr(x, "foo") <- "bar", "external pointer is not valid")

  expect_error(xml_attrs(x), "external pointer is not valid")
  expect_error(xml_attrs(x) <- list(), "external pointer is not valid")

  expect_error(xml_child(x), "external pointer is not valid")
  expect_error(xml_children(x), "external pointer is not valid")

  expect_error(xml_contents(x), "external pointer is not valid")

  expect_error(xml_double(x), "external pointer is not valid")

  expect_error(xml_find_all(x, ""), "external pointer is not valid")
  expect_error(xml_find_chr(x, ""), "external pointer is not valid")
  expect_error(xml_find_first(x, ""), "external pointer is not valid")
  expect_error(xml_find_lgl(x, ""), "external pointer is not valid")
  expect_error(xml_find_num(x, ""), "external pointer is not valid")

  expect_error(xml_has_attr(x, ""), "external pointer is not valid")

  expect_error(xml_integer(x), "external pointer is not valid")

  expect_error(xml_length(x), "external pointer is not valid")

  expect_error(xml_name(x), "external pointer is not valid")
  expect_error(xml_name(x) <- "foo", "external pointer is not valid")

  expect_error(xml_ns(x), "external pointer is not valid")
  expect_error(xml_ns_strip(x), "external pointer is not valid")

  expect_error(xml_parent(x), "external pointer is not valid")
  expect_error(xml_parents(x), "external pointer is not valid")

  expect_error(xml_path(x), "external pointer is not valid")

  expect_error(xml_remove(x), "external pointer is not valid")

  expect_error(xml_replace(x, x), "external pointer is not valid")

  expect_error(xml_set_namespace(x, "foo"), "external pointer is not valid")

  expect_error(xml_siblings(x), "external pointer is not valid")

  expect_error(xml_structure(x), "external pointer is not valid")

  expect_error(xml_text(x), "external pointer is not valid")
  expect_error(xml_text(x) <- "test", "external pointer is not valid")

  expect_error(xml_type(x), "external pointer is not valid")

  expect_error(xml_url(x), "external pointer is not valid")
})

context("xml_children")

x <- read_xml("<foo> <bar><boo /></bar> <baz/> </foo>")

test_that("xml_child() returns the proper child", {
  expect_identical(xml_child(x), xml_children(x)[[1L]])

  expect_identical(xml_child(x, 2), xml_children(x)[[2L]])
})

test_that("xml_child() returns child by name", {
  expect_identical(xml_child(x, "baz"), xml_find_first(x, "./baz"))
})

test_that("xml_child() errors if more than one search is given", {
  expect_error(xml_child(x, 1:2), "`search` must be of length 1")
})

test_that("xml_child() errors if search is not numeric or character", {
  expect_error(xml_child(x, TRUE), "`search` must be `numeric` or `character`")
  expect_error(xml_child(x, as.factor("test")), "`search` must be `numeric` or `character`")
  expect_error(xml_child(x, raw(1)), "`search` must be `numeric` or `character`")
  expect_error(xml_child(x, list(1)), "`search` must be `numeric` or `character`")
})

test_that("xml_length", {
  expect_equal(xml_length(x), 2)
  all <- xml_find_all(x, "//*")
  expect_equal(xml_length(all), c(2, 1, 0, 0))
})

test_that("xml_parent", {
  expect_identical(unclass(xml_parent(xml_child(x))), unclass(x))
})

test_that("xml_parents", {
  expect_equal(
    xml_name(xml_parents(xml_find_first(x, "//boo"))),
    c("bar", "foo"))
})

test_that("xml_root", {
  doc <- xml_new_document()

  expect_is(xml_root(doc), "xml_missing")

  a <- xml_add_child(doc, "a")
  b <- xml_add_child(doc, "b")

  expect_that(xml_name(xml_root(b)), equals("a"))
  expect_that(xml_name(xml_root(doc)), equals("a"))
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

read_connection_ <- function(con, chunk_size = 64 * 1024L) {
    .Call(`_xml2_read_connection_`, con, chunk_size)
}

xml_parse_options <- function() {
    .Call(`_xml2_xml_parse_options`)
}

doc_parse_file <- function(path, encoding = "", as_html = FALSE, options = 0L) {
    .Call(`_xml2_doc_parse_file`, path, encoding, as_html, options)
}

doc_parse_raw <- function(x, encoding, base_url = "", as_html = FALSE, options = 0L) {
    .Call(`_xml2_doc_parse_raw`, x, encoding, base_url, as_html, options)
}

doc_root <- function(x) {
    .Call(`_xml2_doc_root`, x)
}

doc_has_root <- function(x) {
    .Call(`_xml2_doc_has_root`, x)
}

doc_url <- function(x) {
    .Call(`_xml2_doc_url`, x)
}

doc_new <- function(version, encoding = "UTF-8") {
    .Call(`_xml2_doc_new`, version, encoding)
}

doc_set_root <- function(doc, root) {
    .Call(`_xml2_doc_set_root`, doc, root)
}

init_libxml2 <- function() {
    invisible(.Call(`_xml2_init_libxml2`))
}

libxml2_version <- function() {
    .Call(`_xml2_libxml2_version`)
}

unique_ns <- function(ns) {
    .Call(`_xml2_unique_ns`, ns)
}

doc_namespaces <- function(doc) {
    .Call(`_xml2_doc_namespaces`, doc)
}

ns_lookup_uri <- function(doc, node, uri) {
    .Call(`_xml2_ns_lookup_uri`, doc, node, uri)
}

ns_lookup <- function(doc, node, prefix) {
    .Call(`_xml2_ns_lookup`, doc, node, prefix)
}

node_name <- function(node, nsMap) {
    .Call(`_xml2_node_name`, node, nsMap)
}

node_set_name <- function(node, value) {
    invisible(.Call(`_xml2_node_set_name`, node, value))
}

node_text <- function(node) {
    .Call(`_xml2_node_text`, node)
}

node_attr <- function(node, name, missing, nsMap) {
    .Call(`_xml2_node_attr`, node, name, missing, nsMap)
}

node_attrs <- function(node_, nsMap) {
    .Call(`_xml2_node_attrs`, node_, nsMap)
}

node_set_attr <- function(node_, name, value, nsMap) {
    invisible(.Call(`_xml2_node_set_attr`, node_, name, value, nsMap))
}

node_remove_attr <- function(node_, name, nsMap) {
    invisible(.Call(`_xml2_node_remove_attr`, node_, name, nsMap))
}

node_children <- function(node, onlyNode = TRUE) {
    .Call(`_xml2_node_children`, node, onlyNode)
}

node_length <- function(node, onlyNode = TRUE) {
    .Call(`_xml2_node_length`, node, onlyNode)
}

node_has_children <- function(node, onlyNode = TRUE) {
    .Call(`_xml2_node_has_children`, node, onlyNode)
}

node_parents <- function(node) {
    .Call(`_xml2_node_parents`, node)
}

node_siblings <- function(node, onlyNode = TRUE) {
    .Call(`_xml2_node_siblings`, node, onlyNode)
}

node_parent <- function(n) {
    .Call(`_xml2_node_parent`, n)
}

node_path <- function(n) {
    .Call(`_xml2_node_path`, n)
}

nodes_duplicated <- function(nodes) {
    .Call(`_xml2_nodes_duplicated`, nodes)
}

node_type <- function(node) {
    .Call(`_xml2_node_type`, node)
}

node_copy <- function(node) {
    .Call(`_xml2_node_copy`, node)
}

node_set_content <- function(node, content) {
    invisible(.Call(`_xml2_node_set_content`, node, content))
}

node_append_content <- function(node, content) {
    invisible(.Call(`_xml2_node_append_content`, node, content))
}

node_append_child <- function(parent, cur) {
    .Call(`_xml2_node_append_child`, parent, cur)
}

node_prepend_child <- function(parent, cur) {
    .Call(`_xml2_node_prepend_child`, parent, cur)
}

node_prepend_sibling <- function(cur, elem) {
    .Call(`_xml2_node_prepend_sibling`, cur, elem)
}

node_append_sibling <- function(cur, elem) {
    .Call(`_xml2_node_append_sibling`, cur, elem)
}

node_replace <- function(old, cur) {
    .Call(`_xml2_node_replace`, old, cur)
}

node_remove <- function(cur, free) {
    invisible(.Call(`_xml2_node_remove`, cur, free))
}

node_new <- function(name) {
    .Call(`_xml2_node_new`, name)
}

node_cdata_new <- function(doc, content) {
    .Call(`_xml2_node_cdata_new`, doc, content)
}

node_comment_new <- function(content) {
    .Call(`_xml2_node_comment_new`, content)
}

node_new_ns <- function(name, ns) {
    .Call(`_xml2_node_new_ns`, name, ns)
}

node_null <- function() {
    .Call(`_xml2_node_null`)
}

node_set_namespace_uri <- function(doc, node, uri) {
    invisible(.Call(`_xml2_node_set_namespace_uri`, doc, node, uri))
}

node_set_namespace_prefix <- function(doc, node, prefix) {
    invisible(.Call(`_xml2_node_set_namespace_prefix`, doc, node, prefix))
}

node_new_dtd <- function(doc, name = "", eid = "", sid = "") {
    invisible(.Call(`_xml2_node_new_dtd`, doc, name, eid, sid))
}

xml_save_options <- function() {
    .Call(`_xml2_xml_save_options`)
}

doc_write_file <- function(x, path, encoding = "UTF-8", options = 1L) {
    invisible(.Call(`_xml2_doc_write_file`, x, path, encoding, options))
}

doc_write_connection <- function(x, connection, encoding = "UTF-8", options = 1L) {
    invisible(.Call(`_xml2_doc_write_connection`, x, connection, encoding, options))
}

doc_write_character <- function(x, encoding = "UTF-8", options = 1L) {
    .Call(`_xml2_doc_write_character`, x, encoding, options)
}

node_write_file <- function(x, path, encoding = "UTF-8", options = 1L) {
    invisible(.Call(`_xml2_node_write_file`, x, path, encoding, options))
}

node_write_connection <- function(x, connection, encoding = "UTF-8", options = 1L) {
    invisible(.Call(`_xml2_node_write_connection`, x, connection, encoding, options))
}

node_write_character <- function(x, encoding = "UTF-8", options = 1L) {
    .Call(`_xml2_node_write_character`, x, encoding, options)
}

doc_validate <- function(doc, schema) {
    .Call(`_xml2_doc_validate`, doc, schema)
}

#' Convert between relative and absolute urls.
#'
#' @param x A character vector of urls relative to that base
#' @param base A string giving a base url.
#' @return A character vector of urls
#' @seealso \code{\link{xml_url}} to retrieve the URL associated with a document
#' @export
#' @examples
#' url_absolute(c(".", "..", "/", "/x"), "http://hadley.nz/a/b/c/d")
#'
#' url_relative("http://hadley.nz/a/c", "http://hadley.nz")
#' url_relative("http://hadley.nz/a/c", "http://hadley.nz/")
#' url_relative("http://hadley.nz/a/c", "http://hadley.nz/a/b")
#' url_relative("http://hadley.nz/a/c", "http://hadley.nz/a/b/")
url_absolute <- function(x, base) {
    .Call(`_xml2_url_absolute`, x, base)
}

#' @export
#' @rdname url_absolute
url_relative <- function(x, base) {
    .Call(`_xml2_url_relative`, x, base)
}

#' Parse a url into its component pieces.
#'
#' @param x A character vector of urls.
#' @return A dataframe with one row for each element of \code{x} and
#'   columns: scheme, server, port, user, path, query, fragment.
#' @export
#' @examples
#' url_parse("http://had.co.nz/")
#' url_parse("http://had.co.nz:1234/")
#' url_parse("http://had.co.nz:1234/?a=1&b=2")
#' url_parse("http://had.co.nz:1234/?a=1&b=2#def")
url_parse <- function(x) {
    .Call(`_xml2_url_parse`, x)
}

#' Escape and unescape urls.
#'
#' @param x A character vector of urls.
#' @param reserved A string containing additional characters to avoid escaping.
#' @export
#' @examples
#' url_escape("a b c")
#' url_escape("a b c", "")
#'
#' url_unescape("a%20b%2fc")
#' url_unescape("%C2%B5")
url_escape <- function(x, reserved = "") {
    .Call(`_xml2_url_escape`, x, reserved)
}

#' @export
#' @rdname url_escape
url_unescape <- function(x) {
    .Call(`_xml2_url_unescape`, x)
}

xpath_search <- function(node, doc, xpath, nsMap, num_results) {
    .Call(`_xml2_xpath_search`, node, doc, xpath, nsMap, num_results)
}

test_that("url_absolute", {
  expect_equal(
    url_absolute(c(".", "..", "/", "/x"), "http://hadley.nz/a/b/c/d"),
    c("http://hadley.nz/a/b/c/", "http://hadley.nz/a/b/", "http://hadley.nz/", "http://hadley.nz/x")
  )

  expect_error(
    url_absolute(c(".", "..", "/", "/x"), c("http://hadley.nz/a/b/c/d", "http://foo.bar")),
    "Base URL must be length 1"
  )
})

test_that("url_relative", {
  # The behavior of libxml2 with relative paths is fragile so we skip this test
  skip("libxml2-dependent")

  expect_equal(
    url_relative("http://hadley.nz/a/c", "http://hadley.nz"),
    "/a/c"
  )

  expect_equal(
    url_relative("http://hadley.nz/a/c", "http://hadley.nz/"),
    "a/c"
  )

  expect_equal(
    url_relative("http://hadley.nz/a/c", "http://hadley.nz/a/b"),
    "c"
  )

  expect_equal(
    url_relative("http://hadley.nz/a/c", "http://hadley.nz/a/b/"),
    "../c"
  )

  expect_error(
    url_relative("http://hadley.nz/a/c", c("http://hadley.nz/a/b/c/d", "http://foo.bar")),
    "Base URL must be length 1"
  )
})

test_that("url_parse", {
  expect_equal(
    url_parse("http://had.co.nz/"),
    data.frame(
      scheme = "http", server = "had.co.nz", port = NA_integer_,
      user = "", path = "/", query = "", fragment = "", stringsAsFactors = FALSE
    )
  )

  expect_equal(
    url_parse("http://had.co.nz:1234/"),
    data.frame(
      scheme = "http", server = "had.co.nz", port = 1234L,
      user = "", path = "/", query = "", fragment = "", stringsAsFactors = FALSE
    )
  )

  expect_equal(
    url_parse("http://had.co.nz:1234/?a=1&b=2"),
    data.frame(
      scheme = "http", server = "had.co.nz", port = 1234L,
      user = "", path = "/", query = "a=1&b=2", fragment = "", stringsAsFactors = FALSE
    )
  )

  expect_equal(
    url_parse("http://had.co.nz:1234/?a=1&b=2#def"),
    data.frame(
      scheme = "http", server = "had.co.nz", port = 1234L,
      user = "", path = "/", query = "a=1&b=2", fragment = "def", stringsAsFactors = FALSE
    )
  )
})

test_that("url_escape", {
  expect_error(
    url_escape("a b c", reserved = c("a", "b")),
    "`reserved` must be character vector of length 1"
  )

  expect_equal(
    url_escape("a b c"),
    "a%20b%20c"
  )

  expect_equal(
    url_escape("a b c", " "),
    "a b c"
  )

  expect_equal(
    url_unescape("a%20b%2fc"),
    "a b/c"
  )

  expect_equal(
    url_unescape("%C2%B5"),
    "\u00B5"
  )
})

context("url")

test_that("url_absolute", {
  expect_equal(
    url_absolute(c(".", "..", "/", "/x"), "http://hadley.nz/a/b/c/d"),
    c("http://hadley.nz/a/b/c/", "http://hadley.nz/a/b/", "http://hadley.nz/", "http://hadley.nz/x"))

    expect_error(url_absolute(c(".", "..", "/", "/x"), c("http://hadley.nz/a/b/c/d", "http://foo.bar")),
      "Base URL must be length 1")
})

test_that("url_relative", {

  expect_equal(
    url_relative("http://hadley.nz/a/c", "http://hadley.nz"),
    "/a/c")

  expect_equal(
    url_relative("http://hadley.nz/a/c", "http://hadley.nz/"),
    "../a/c")

  expect_equal(
    url_relative("http://hadley.nz/a/c", "http://hadley.nz/a/b"),
    "c")

  expect_equal(
    url_relative("http://hadley.nz/a/c", "http://hadley.nz/a/b/"),
    "../c")

    expect_error(url_relative("http://hadley.nz/a/c", c("http://hadley.nz/a/b/c/d", "http://foo.bar")),
      "Base URL must be length 1")
})

test_that("url_parse", {

  expect_equal(
    url_parse("http://had.co.nz/"),
    data.frame(scheme = "http", server = "had.co.nz", port = NA_integer_,
      user = "", path = "/", query = "", fragment = "", stringsAsFactors = FALSE))

  expect_equal(
    url_parse("http://had.co.nz:1234/"),
    data.frame(scheme = "http", server = "had.co.nz", port = 1234L,
      user = "", path = "/", query = "", fragment = "", stringsAsFactors = FALSE))

  expect_equal(
    url_parse("http://had.co.nz:1234/?a=1&b=2"),
    data.frame(scheme = "http", server = "had.co.nz", port = 1234L,
      user = "", path = "/", query = "a=1&b=2", fragment = "", stringsAsFactors = FALSE))

  expect_equal(
    url_parse("http://had.co.nz:1234/?a=1&b=2#def"),
    data.frame(scheme = "http", server = "had.co.nz", port = 1234L,
      user = "", path = "/", query = "a=1&b=2", fragment = "def", stringsAsFactors = FALSE))
})

test_that("url_escape", {
  expect_error(url_escape("a b c", reserved = c("a", "b")),
    "`reserved` must be character vector of length 1")

  expect_equal(
    url_escape("a b c"),
    "a%20b%20c")

  expect_equal(
    url_escape("a b c", " "),
    "a b c")

  expect_equal(
    url_unescape("a%20b%2fc"),
    "a b/c")

  expect_equal(
    url_unescape("%C2%B5"),
    "\u00B5")
})

test_that("url_absolute", {
  expect_equal(
    url_absolute(c(".", "..", "/", "/x"), "http://hadley.nz/a/b/c/d"),
    c(
      "http://hadley.nz/a/b/c/",
      "http://hadley.nz/a/b/",
      "http://hadley.nz/",
      "http://hadley.nz/x"
    )
  )

  expect_error(
    url_absolute(
      c(".", "..", "/", "/x"),
      c("http://hadley.nz/a/b/c/d", "http://foo.bar")
    ),
    "Base URL must be length 1"
  )
})

test_that("url_relative", {
  # The behavior of libxml2 with relative paths is fragile so we skip this test
  skip("libxml2-dependent")

  expect_equal(
    url_relative("http://hadley.nz/a/c", "http://hadley.nz"),
    "/a/c"
  )

  expect_equal(
    url_relative("http://hadley.nz/a/c", "http://hadley.nz/"),
    "a/c"
  )

  expect_equal(
    url_relative("http://hadley.nz/a/c", "http://hadley.nz/a/b"),
    "c"
  )

  expect_equal(
    url_relative("http://hadley.nz/a/c", "http://hadley.nz/a/b/"),
    "../c"
  )

  expect_error(
    url_relative(
      "http://hadley.nz/a/c",
      c("http://hadley.nz/a/b/c/d", "http://foo.bar")
    ),
    "Base URL must be length 1"
  )
})

test_that("url_parse", {
  expect_equal(
    url_parse("http://had.co.nz/"),
    data.frame(
      scheme = "http",
      server = "had.co.nz",
      port = NA_integer_,
      user = "",
      path = "/",
      query = "",
      fragment = "",
      stringsAsFactors = FALSE
    )
  )

  expect_equal(
    url_parse("http://had.co.nz:1234/"),
    data.frame(
      scheme = "http",
      server = "had.co.nz",
      port = 1234L,
      user = "",
      path = "/",
      query = "",
      fragment = "",
      stringsAsFactors = FALSE
    )
  )

  expect_equal(
    url_parse("http://had.co.nz:1234/?a=1&b=2"),
    data.frame(
      scheme = "http",
      server = "had.co.nz",
      port = 1234L,
      user = "",
      path = "/",
      query = "a=1&b=2",
      fragment = "",
      stringsAsFactors = FALSE
    )
  )

  expect_equal(
    url_parse("http://had.co.nz:1234/?a=1&b=2#def"),
    data.frame(
      scheme = "http",
      server = "had.co.nz",
      port = 1234L,
      user = "",
      path = "/",
      query = "a=1&b=2",
      fragment = "def",
      stringsAsFactors = FALSE
    )
  )
})

test_that("url_escape", {
  expect_error(
    url_escape("a b c", reserved = c("a", "b")),
    "`reserved` must be character vector of length 1"
  )

  expect_equal(
    url_escape("a b c"),
    "a%20b%20c"
  )

  expect_equal(
    url_escape("a b c", " "),
    "a b c"
  )

  expect_equal(
    url_unescape("a%20b%2fc"),
    "a b/c"
  )

  expect_equal(
    url_unescape("%C2%B5"),
    "\u00B5"
  )
})

nodeset_apply <- function(x, fun, ...) UseMethod("nodeset_apply")

#' @export
nodeset_apply.xml_missing <- function(x, fun, ...) {
  xml_nodeset()
}

#' @export
nodeset_apply.xml_nodeset <- function(x, fun, ...) {
  if (length(x) == 0) {
    return(xml_nodeset())
  }

  is_missing <- is.na(x)
  res <- list(length(x))

  res[is_missing] <- list(xml_missing())
  if (any(!is_missing)) {
    res[!is_missing] <- lapply(x[!is_missing], function(x) fun(x$node, ...))
  }

  make_nodeset(res, x[[1]]$doc)
}

#' @export
nodeset_apply.xml_node <- function(x, fun, ...) {
  nodes <- fun(x$node, ...)
  xml_nodeset(lapply(nodes, xml_node, doc = x$doc))
}

#' @export
nodeset_apply.xml_document <- function(x, fun, ...) {
  if (inherits(x, "xml_node")) {
    NextMethod()
  } else {
    xml_nodeset()
  }
}

#' Serializing XML objects to connections.
#'
#' @inheritParams base::serialize
#' @param ... Additional arguments passed to \code{\link{read_xml}}.
#' @inherit base::serialize return
#' @examples
#' library(xml2)
#' x <- read_xml("<a>
#'   <b><c>123</c></b>
#'   <b><c>456</c></b>
#' </a>")
#'
#' b <- xml_find_all(x, "//b")
#' out <- xml_serialize(b, NULL)
#' xml_unserialize(out)
#' @export
xml_serialize <- function(object, connection, ...) UseMethod("xml_serialize")

#' @export
xml_serialize.xml_document <- function(object, connection, ...) {
  if (is.character(connection)) {
    connection <- file(connection, "w", raw = TRUE)
    on.exit(close(connection))
  }
  serialize(structure(as.character(object, ...), class = "xml_serialized_document"), connection)
}

#' @export
xml_serialize.xml_node <- function(object, connection, ...) {
  if (is.character(connection)) {
    connection <- file(connection, "w", raw = TRUE)
    on.exit(close(connection))
  }
  x <- as_xml_document(object)
  serialize(structure(as.character(x, ...), class = "xml_serialized_node"), connection)
}

#' @export
xml_serialize.xml_nodeset <- function(object, connection, ...) {
  if (is.character(connection)) {
    connection <- file(connection, "w", raw = TRUE)
    on.exit(close(connection))
  }
  x <- as_xml_document(object, "root")
  serialize(structure(as.character(x, ...), class = "xml_serialized_nodeset"), connection)
}

#' @rdname xml_serialize
#' @export
xml_unserialize <- function(connection, ...) {
  if (is.character(connection)) {
    connection <- file(connection, "r", raw = TRUE)
    on.exit(close(connection))
  }
  object <- unserialize(connection)
  if (inherits(object, "xml_serialized_nodeset")) {
    x <- read_xml(unclass(object), ...)

    # Select only the direct children of the root
    res <- xml_find_all(x, "/*/node()")
  } else if (inherits(object, "xml_serialized_node")) {
    x <- read_xml(unclass(object), ...)

    # Select only the root
    res <- xml_find_first(x, "/node()")
  } else if (inherits(object, "xml_serialized_document")) {
    res <- read_xml(unclass(object), ...)
  } else {
    stop("Not a serialized xml2 object", call. = FALSE)
  }
  res
}

context("xml_serialize")

x <- read_xml("<a>
  <b><c>123</c></b>
  <b><c>456</c></b>
  </a>")
test_that("xml_serialize and xml_unserialize work with xml_document input", {
  out <- xml_unserialize(xml_serialize(x, NULL))
  expect_identical(as.character(x), as.character(out))

  f <- tempfile()
  on.exit(unlink(f))

  xml_serialize(x, f)
  expect_identical(as.character(xml_unserialize(f)), as.character(x))
})

test_that("xml_serialize and xml_unserialize work with xml_node input", {
  b <- xml_find_first(x, "//b")
  out <- xml_unserialize(xml_serialize(b, NULL))
  expect_identical(as.character(b), as.character(out))

  f <- tempfile()
  on.exit(unlink(f))

  xml_serialize(b, f)
  expect_identical(as.character(xml_unserialize(f)), as.character(b))
})

test_that("xml_serialize and xml_unserialize work with xml_nodeset input", {
  b <- xml_find_all(x, "//b")
  out <- xml_unserialize(xml_serialize(b, NULL))
  expect_identical(as.character(b), as.character(out))

  f <- tempfile()
  on.exit(unlink(f))

  xml_serialize(b, f)
  expect_identical(as.character(xml_unserialize(f)), as.character(b))
})

test_that("xml_unserialize throws an error if given a invalid object", {
  expect_error(xml_unserialize(serialize(1, NULL)), "Not a serialized xml2 object")
})

#' Find nodes that match an xpath expression.
#'
#' Xpath is like regular expressions for trees - it's worth learning if
#' you're trying to extract nodes from arbitrary locations in a document.
#' Use `xml_find_all` to find all matches - if there's no match you'll
#' get an empty result. Use `xml_find_first` to find a specific match -
#' if there's no match you'll get an `xml_missing` node.
#'
#' @section Deprecated functions:
#' `xml_find_one()` has been deprecated. Instead use
#' `xml_find_first()`.

#' @param xpath A string containing an xpath (1.0) expression.
#' @inheritParams xml_name
#' @param ... Further arguments passed to or from other methods.
#' @return `xml_find_all` returns a nodeset if applied to a node, and a nodeset
#'   or a list of nodesets if applied to a nodeset. If there are no matches,
#'   the nodeset(s) will be empty. Within each nodeset, the result will always
#'   be unique; repeated nodes are automatically de-duplicated.
#'
#'   `xml_find_first` returns a node if applied to a node, and a nodeset
#'   if applied to a nodeset. The output is *always* the same size as
#'   the input. If there are no matches, `xml_find_first` will return a
#'   missing node; if there are multiple matches, it will return the first
#'   only.
#'
#'   `xml_find_num`, `xml_find_chr`, `xml_find_lgl` return
#'   numeric, character and logical results respectively.
#' @export
#' @seealso [xml_ns_strip()] to remove the default namespaces
#' @examples
#' x <- read_xml("<foo><bar><baz/></bar><baz/></foo>")
#' xml_find_all(x, ".//baz")
#' xml_path(xml_find_all(x, ".//baz"))
#'
#' # Note the difference between .// and //
#' # //  finds anywhere in the document (ignoring the current node)
#' # .// finds anywhere beneath the current node
#' (bar <- xml_find_all(x, ".//bar"))
#' xml_find_all(bar, ".//baz")
#' xml_find_all(bar, "//baz")
#'
#' # Find all vs find one -----------------------------------------------------
#' x <- read_xml("<body>
#'   <p>Some <b>text</b>.</p>
#'   <p>Some <b>other</b> <b>text</b>.</p>
#'   <p>No bold here!</p>
#' </body>")
#' para <- xml_find_all(x, ".//p")
#'
#' # By default, if you apply xml_find_all to a nodeset, it finds all matches,
#' # de-duplicates them, and returns as a single nodeset. This means you
#' # never know how many results you'll get
#' xml_find_all(para, ".//b")
#'
#' # If you set flatten to FALSE, though, xml_find_all will return a list of
#' # nodesets, where each nodeset contains the matches for the corresponding
#' # node in the original nodeset.
#' xml_find_all(para, ".//b", flatten = FALSE)
#'
#' # xml_find_first only returns the first match per input node. If there are 0
#' # matches it will return a missing node
#' xml_find_first(para, ".//b")
#' xml_text(xml_find_first(para, ".//b"))
#'
#' # Namespaces ---------------------------------------------------------------
#' # If the document uses namespaces, you'll need use xml_ns to form
#' # a unique mapping between full namespace url and a short prefix
#' x <- read_xml('
#'  <root xmlns:f = "http://foo.com" xmlns:g = "http://bar.com">
#'    <f:doc><g:baz /></f:doc>
#'    <f:doc><g:baz /></f:doc>
#'  </root>
#' ')
#' xml_find_all(x, ".//f:doc")
#' xml_find_all(x, ".//f:doc", xml_ns(x))
xml_find_all <- function(x, xpath, ns = xml_ns(x), ...) {
  UseMethod("xml_find_all")
}

#' @export
xml_find_all.xml_missing <- function(x, xpath, ns = xml_ns(x), ...) {
  xml_nodeset()
}

#' @export
xml_find_all.xml_node <- function(x, xpath, ns = xml_ns(x), ...) {
  nodes <- .Call(xpath_search, x$node, x$doc, xpath, ns, Inf)
  xml_nodeset(nodes)
}

#' @param flatten A logical indicating whether to return a single, flattened
#'   nodeset or a list of nodesets.
#' @export
#' @rdname xml_find_all
xml_find_all.xml_nodeset <- function(
  x,
  xpath,
  ns = xml_ns(x),
  flatten = TRUE,
  ...
) {
  if (length(x) == 0) {
    return(xml_nodeset())
  }

  res <- lapply(x, function(x) {
    .Call(xpath_search, x$node, x$doc, xpath, ns, Inf)
  })

  if (isTRUE(flatten)) {
    return(xml_nodeset(unlist(recursive = FALSE, res)))
  }

  res[] <- lapply(res, xml_nodeset)
  res
}

#' @export
#' @rdname xml_find_all
xml_find_first <- function(x, xpath, ns = xml_ns(x)) {
  UseMethod("xml_find_first")
}

#' @export
xml_find_first.xml_missing <- function(x, xpath, ns = xml_ns(x)) {
  xml_missing()
}

#' @export
xml_find_first.xml_node <- function(x, xpath, ns = xml_ns(x)) {
  res <- .Call(xpath_search, x$node, x$doc, xpath, ns, 1)
  if (length(res) == 1) {
    res[[1]]
  } else {
    res
  }
}

#' @export
xml_find_first.xml_nodeset <- function(x, xpath, ns = xml_ns(x)) {
  if (length(x) == 0) {
    return(xml_nodeset())
  }

  xml_nodeset(
    lapply(
      x,
      function(x) {
        xml_find_first(x, xpath = xpath, ns = ns)
      }
    ),
    deduplicate = FALSE
  )
}


#' @export
#' @rdname xml_find_all
xml_find_num <- function(x, xpath, ns = xml_ns(x)) {
  UseMethod("xml_find_num")
}

#' @export
xml_find_num.xml_node <- function(x, xpath, ns = xml_ns(x)) {
  res <- .Call(xpath_search, x$node, x$doc, xpath, ns, Inf)
  if (is.numeric(res) && is.nan(res)) {
    return(res)
  }

  check_number_decimal(res, arg = I(paste0("Element at path `", xpath, "`")))
  res
}

#' @export
xml_find_num.xml_nodeset <- function(x, xpath, ns = xml_ns(x)) {
  if (length(x) == 0) {
    return(numeric())
  }

  vapply(x, function(x) xml_find_num(x, xpath = xpath, ns = ns), numeric(1))
}

#' @export
xml_find_num.xml_missing <- function(x, xpath, ns = xml_ns(x)) {
  numeric(0)
}

#' @export
#' @rdname xml_find_all
xml_find_int <- function(x, xpath, ns = xml_ns(x)) {
  UseMethod("xml_find_int")
}

#' @export
xml_find_int.xml_node <- function(x, xpath, ns = xml_ns(x)) {
  res <- .Call(xpath_search, x$node, x$doc, xpath, ns, Inf)
  check_number_whole(res, arg = I(paste0("Element at path `", xpath, "`")))
  as.integer(res)
}

#' @export
xml_find_int.xml_nodeset <- function(x, xpath, ns = xml_ns(x)) {
  if (length(x) == 0) {
    return(integer())
  }

  vapply(x, function(x) xml_find_int(x, xpath = xpath, ns = ns), integer(1))
}

#' @export
xml_find_int.xml_missing <- function(x, xpath, ns = xml_ns(x)) {
  integer(0)
}

#' @export
#' @rdname xml_find_all
xml_find_chr <- function(x, xpath, ns = xml_ns(x)) {
  UseMethod("xml_find_chr")
}

#' @export
xml_find_chr.xml_node <- function(x, xpath, ns = xml_ns(x)) {
  res <- .Call(xpath_search, x$node, x$doc, xpath, ns, Inf)
  check_string(res, arg = I(paste0("Element at path `", xpath, "`")))
  res
}

#' @export
xml_find_chr.xml_nodeset <- function(x, xpath, ns = xml_ns(x)) {
  if (length(x) == 0) {
    return(character())
  }

  vapply(x, function(x) xml_find_chr(x, xpath = xpath, ns = ns), character(1))
}

#' @export
xml_find_chr.xml_missing <- function(x, xpath, ns = xml_ns(x)) {
  character(0)
}

#' @export
#' @rdname xml_find_all
xml_find_lgl <- function(x, xpath, ns = xml_ns(x)) {
  UseMethod("xml_find_lgl")
}

#' @export
xml_find_lgl.xml_node <- function(x, xpath, ns = xml_ns(x)) {
  res <- .Call(xpath_search, x$node, x$doc, xpath, ns, Inf)
  check_bool(res, arg = I(paste0("Element at path `", xpath, "`")))
  res
}

#' @export
xml_find_lgl.xml_nodeset <- function(x, xpath, ns = xml_ns(x)) {
  if (length(x) == 0) {
    return(logical())
  }

  vapply(x, function(x) xml_find_lgl(x, xpath = xpath, ns = ns), logical(1))
}

#' @export
xml_find_lgl.xml_missing <- function(x, xpath, ns = xml_ns(x)) {
  logical(0)
}

# Deprecated functions ----------------------------------------------------

#' @rdname xml_find_all
#' @usage NULL
#' @export
xml_find_one <- function(x, xpath, ns = xml_ns(x)) {
  .Deprecated("xml_find_first")
  UseMethod("xml_find_first")
}

# Find one ---------------------------------------------------------------------

test_that("xml_find_first returns a missing object if no match", {
  x <- read_xml("<x><y/></x>")
  expect_equal(xml_find_first(x, ".//z"), xml_missing())
})

test_that("xml_find_first returns the first match if more than one match", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_first(x, ".//y")
  expect_s3_class(y, "xml_node")
})

test_that("xml_find_first does not deduplicate identical results", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_all(x, ".//y")
  z <- xml_find_first(y, "..")
  expect_s3_class(z, "xml_nodeset")
  expect_length(z, 2)
})

# Find all ---------------------------------------------------------------------

test_that("unqualified names don't look in default ns", {
  x <- read_xml(test_path("ns-multiple-default.xml"))

  expect_length(xml_find_all(x, "//bar"), 0)
})

test_that("qualified names matches to namespace", {
  x <- read_xml(test_path("ns-multiple-default.xml"))
  ns <- xml_ns(x)

  expect_length(xml_find_all(x, "//d1:bar", ns), 1)
  expect_length(xml_find_all(x, "//d2:bar", ns), 1)
})

test_that("warning if unknown namespace", {
  x <- read_xml("<foo><bar /></foo>")
  maybe_error(
    expect_warning(xml_find_all(x, "//g:bar"), "Undefined namespace prefix"),
    "evaluation failed"
  )
})

test_that("no matches returns empty nodeset", {
  x <- read_xml("<foo><bar /></foo>")
  expect_length(xml_find_all(x, "//baz"), 0)
})

test_that("xml_find_all returns nodeset or list of nodesets based on flatten", {
  x <- read_xml("<body><p>Some <b>text</b>.</p>
                 <p>Some <b>other</b> <b>text</b>.</p>
                 <p>No bold here!</p></body>")
  y <- xml_find_all(x, ".//p")
  z <- xml_find_all(y, ".//b", flatten = FALSE)
  expect_s3_class(xml_find_all(y, ".//b"), "xml_nodeset")
  expect_type(z, "list")
  expect_s3_class(z[[1L]], "xml_nodeset")
})

# Find num ---------------------------------------------------------------------
test_that("xml_find_num errors with non numeric results", {
  x <- read_xml("<x><y/><y/></x>")
  expect_snapshot(error = TRUE, {
    xml_find_num(x, "//z")
    xml_find_num(x, "//y")
    xml_find_num(x, "1=1")
    xml_find_num(x, "string(5)")
  })
})

test_that("xml_find_num returns a numeric result", {
  x <- read_xml("<x><y>1</y><y/></x>")
  expect_equal(xml_find_num(x, "1 div 0"), Inf)

  expect_equal(xml_find_num(x, "-1 div 0"), -Inf)

  expect_equal(xml_find_num(x, "0 div 0"), NaN)

  expect_equal(xml_find_num(x, "1 div floor(-0.1)"), -1)

  expect_equal(xml_find_num(x, "1 div floor(0)"), Inf)

  expect_equal(xml_find_num(x, "1 div floor(-0)"), -Inf)
})

# Find int ---------------------------------------------------------------------
test_that("xml_find_int errors with non integer results", {
  x <- read_xml("<x><y/><y/></x>")
  expect_snapshot(error = TRUE, {
    xml_find_int(x, "//z")
    xml_find_int(x, "//y")
    xml_find_int(x, "number(1.1)")
  })
})

test_that("xml_find_int returns a integer result", {
  x <- read_xml("<x><y>1</y><y/></x>")

  expect_identical(xml_find_int(x, "1 div floor(-0.1)"), -1L)
  expect_identical(xml_find_int(x, "number(//y)"), 1L)
  expect_identical(xml_find_int(x, "string-length(string('abcd'))"), 4L)
})

# Find chr ---------------------------------------------------------------------
test_that("xml_find_chr errors with non character results", {
  x <- read_xml("<x><y/><y/></x>")
  expect_snapshot(error = TRUE, {
    xml_find_chr(x, "//z")
    xml_find_chr(x, "//y")
    xml_find_chr(x, "1=1")
    xml_find_chr(x, "1+1")
  })
})

test_that("xml_find_chr returns a character result", {
  x <- read_xml("<x><y>1</y><y/></x>")
  expect_equal(xml_find_chr(x, "string(5)"), "5")

  expect_equal(xml_find_chr(x, "string(0.5)"), "0.5")

  expect_equal(xml_find_chr(x, "string(-0.5)"), "-0.5")

  expect_equal(xml_find_chr(x, "concat(\"titi\", \"toto\")"), "tititoto")

  expect_equal(xml_find_chr(x, "substring(\"12345\", 2, 3)"), "234")

  expect_equal(xml_find_chr(x, "substring(\"12345\", 2)"), "2345")

  expect_equal(xml_find_chr(x, "substring(\"12345\", -4)"), "12345")
})

# Find lgl ---------------------------------------------------------------------
test_that("xml_find_lgl errors with non logical results", {
  x <- read_xml("<x><y/><y/></x>")
  expect_snapshot(error = TRUE, {
    xml_find_lgl(x, "//z")
    xml_find_lgl(x, "//y")
    xml_find_lgl(x, "string(5)")
    xml_find_lgl(x, "1+1")
  })
})

test_that("xml_find_lgl returns a logical result", {
  x <- read_xml("<x><y>1</y><y/></x>")

  expect_true(xml_find_lgl(x, "1=1"))

  expect_false(xml_find_lgl(x, "1!=1"))

  expect_true(xml_find_lgl(x, "true()=true()"))

  expect_false(xml_find_lgl(x, "true()=false()"))

  expect_true(xml_find_lgl(x, "'test'='test'"))
})

test_that("Searches with empty inputs retain type stability", {
  empty <- xml_nodeset()

  expect_equal(xml_find_num(empty, "1 div 0"), integer())

  expect_equal(xml_find_chr(empty, "string(0.5)"), character())

  expect_equal(xml_find_lgl(empty, "1=1"), logical())
})

test_that("Searches with entities work (#241)", {
  res <- read_xml(test_path("records.xml"), options = "DTDVALID")

  field1 <- xml_find_all(res, "//field1")

  expect_equal(xml_text(field1), "foo bar Quantitative Consultancy")
})

test_that("A helpful error is given from non-string xpath in xml_find_first/all", {
  node <- read_xml("<a>1</a>")

  expect_error(xml_find_all(node, 1), "XPath must be a string", fixed = TRUE)
  expect_error(xml_find_first(node, 1), "XPath must be a string", fixed = TRUE)
})

test_that("methods work on empty nodesets", {
  x <- read_xml("<a><b/></a>")
  empty <- xml_find_all(x, "//c")

  expect_error(empty[[1]], "subscript out of bounds")
  expect_identical(empty[1], empty)
  test <- empty

  xml_attr(test, "test") <- 1
  expect_identical(test, empty)

  xml_attrs(test) <- c("test" = 1)
  expect_identical(test, empty)

  xml_name(test) <- "test"
  expect_identical(test, empty)

  xml_text(test) <- "test"
  expect_identical(test, empty)

  expect_identical(as.character(empty), character(0))
  expect_identical(as_list(empty), list())
  expect_identical(nodeset_apply(empty, identical), empty)
  expect_output(print(empty), "\\{xml_nodeset \\(0\\)\\}")
  expect_silent(tree_structure(empty))

  xml_add_child(test, "test")
  expect_identical(test, empty)

  xml_add_sibling(test, "test")
  expect_identical(test, empty)

  expect_identical(xml_attr(empty, "test"), character())
  expect_identical(xml_attrs(empty), list())
  expect_identical(xml_double(empty), numeric())
  expect_identical(xml_find_all(empty), empty)
  expect_identical(xml_find_chr(empty), character())
  expect_identical(xml_find_first(empty), empty)
  expect_identical(xml_find_lgl(empty), logical())
  expect_identical(xml_find_num(empty), numeric())
  expect_identical(xml_integer(empty), integer())
  expect_identical(xml_length(empty), 0L)
  expect_identical(xml_name(empty), character())
  expect_identical(xml_ns(empty), character())
  expect_identical(xml_parent(empty), empty)
  expect_identical(xml_path(empty), character())

  xml_remove(test)
  expect_identical(test, empty)

  xml_replace(test)
  expect_identical(test, empty)

  xml_set_attr(test, "test", 1)
  expect_identical(test, empty)

  xml_set_attrs(test, c("test" = 1))
  expect_identical(test, empty)

  xml_set_name(test, "test")
  expect_identical(test, empty)

  xml_set_text(test, "test")
  expect_identical(test, empty)

  expect_identical(xml_siblings(empty), empty)
  expect_silent(xml_structure(empty))

  expect_identical(xml_text(empty), character())
  expect_identical(xml_url(empty), character())
})

test_that("print method is correct", {
  skip_if(getOption("width") < 20L, "Screen too narrow")

  x <- read_html(test_path("lego.html.bz2"))
  body <- xml_find_first(x, "//body")
  divs <- xml_find_all(body, ".//div")[1:10]
  expect_snapshot(print(divs))

  # double-substring() logic
  s <- c(
    "123456789\\", # always too wide, '\' never encoded
    "12345",       # always fits
    "12\\45"       # doesn't fit when '\' is encoded
  )
  # embed as text on nodes <a>,<b>,<c>
  s <- sprintf("<%1$s>%2$s</%1$s>", letters[1:3], s)
  x <- read_xml(sprintf("<doc>%s</doc>", paste(s, collapse="")))
  expect_snapshot({
    print(x, width = 13L)
    print(x, width = 14L)
  })
})

x <- read_xml("<body>
  <p>Some <b>text</b>.</p>
  <p>Some <b>other</b>.</p>
  <p>No bold text</p>
  </body>")
para <- xml_find_all(x, ".//p")
b <- xml_find_first(para, ".//b")
mss <- b[[3]]

test_that("xml_find returns nodes of class 'xml_missing' for missing nodes", {
  expect_length(b, 3L)
  expect_equal(lengths(b), c(2L, 2L, 0L))
  expect_s3_class(mss, "xml_missing")
})

test_that("xml_missing methods return properly for all S3 methods", {
  expect_equal(as.character(mss), NA_character_)
  expect_equal(as_list(mss), list())
  expect_equal(nodeset_apply(mss), xml_nodeset())
  expect_output(print(mss), "\\{xml_missing\\}\n<NA>")
  expect_equal(tree_structure(mss), NA_character_)
  expect_error(write_xml(mss), "Missing data cannot be written")
  expect_error(write_html(mss), "Missing data cannot be written")
  expect_equal(xml_attr(mss, "dummy_attr"), NA_character_)
  expect_equal(xml_attrs(mss), NA_character_)
  expect_equal(xml_find_all(mss), xml_nodeset())
  expect_equal(xml_find_chr(mss), character())
  expect_equal(xml_find_lgl(mss), logical())
  expect_equal(xml_find_num(mss), numeric())
  expect_equal(xml_find_first(mss), xml_missing())
  expect_equal(xml_length(mss), 0L)
  expect_equal(xml_name(mss), NA_character_)
  expect_equal(xml_parent(mss), xml_missing())
  expect_equal(xml_path(mss), NA_character_)
  expect_equal(xml_text(mss), NA_character_)
  expect_equal(xml_url(mss), NA_character_)
})

test_that("is.na() should return TRUE for xml_missing", {
  expect_true(is.na(xml_missing()))
})

context("xml_nodeset")

test_that("methods work on empty nodesets", {

  x <- read_xml("<a><b/></a>")
  empty <- xml_find_all(x, "//c")

  expect_error(empty[[1]], "subscript out of bounds")
  expect_identical(empty[1], empty)
  test <- empty

  xml_attr(test, "test") <- 1
  expect_identical(test, empty)

  xml_attrs(test) <- c("test" = 1)
  expect_identical(test, empty)

  xml_name(test) <- "test"
  expect_identical(test, empty)

  xml_text(test) <- "test"
  expect_identical(test, empty)

  expect_identical(as.character(empty), character(0))
  expect_identical(as_list(empty), list())
  expect_identical(nodeset_apply(empty, identical), empty)
  expect_output(print(empty), "\\{xml_nodeset \\(0\\)\\}")
  expect_output(tree_structure(empty), NA)

  xml_add_child(test, "test")
  expect_identical(test, empty)

  xml_add_sibling(test, "test")
  expect_identical(test, empty)

  expect_identical(xml_attr(empty, "test"), character())
  expect_identical(xml_attrs(empty), list())
  expect_identical(xml_double(empty), numeric())
  expect_identical(xml_find_all(empty), empty)
  expect_identical(xml_find_chr(empty), character())
  expect_identical(xml_find_first(empty), empty)
  expect_identical(xml_find_lgl(empty), logical())
  expect_identical(xml_find_num(empty), numeric())
  expect_identical(xml_integer(empty), integer())
  expect_identical(xml_length(empty), 0L)
  expect_identical(xml_name(empty), character())
  expect_identical(xml_ns(empty), character())
  expect_identical(xml_parent(empty), empty)
  expect_identical(xml_path(empty), character())

  xml_remove(test)
  expect_identical(test, empty)

  xml_replace(test)
  expect_identical(test, empty)

  xml_set_attr(test, "test", 1)
  expect_identical(test, empty)

  xml_set_attrs(test, c("test" = 1))
  expect_identical(test, empty)

  xml_set_name(test, "test")
  expect_identical(test, empty)

  xml_set_text(test, "test")
  expect_identical(test, empty)

  expect_identical(xml_siblings(empty), empty)
  expect_output(xml_structure(empty), NA)

  expect_identical(xml_text(empty), character())
  expect_identical(xml_type(empty), character())
  expect_identical(xml_url(empty), character())
})

# Find one ---------------------------------------------------------------------

test_that("xml_find_first returns a missing object if no match", {
  x <- read_xml("<x><y/></x>")
  expect_equal(xml_find_first(x, ".//z"), xml_missing())
})

test_that("xml_find_first returns the first match if more than one match", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_first(x, ".//y")
  expect_s3_class(y, "xml_node")
})

test_that("xml_find_first does not deduplicate identical results", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_all(x, ".//y")
  z <- xml_find_first(y, "..")
  expect_s3_class(z, "xml_nodeset")
  expect_length(z, 2)
})

# Find all ---------------------------------------------------------------------

test_that("unqualified names don't look in default ns", {
  x <- read_xml(test_path("ns-multiple-default.xml"))

  expect_length(xml_find_all(x, "//bar"), 0)
})

test_that("qualified names matches to namespace", {
  x <- read_xml(test_path("ns-multiple-default.xml"))
  ns <- xml_ns(x)

  expect_length(xml_find_all(x, "//d1:bar", ns), 1)
  expect_length(xml_find_all(x, "//d2:bar", ns), 1)
})

test_that("warning if unknown namespace", {
  x <- read_xml("<foo><bar /></foo>")
  maybe_error(
    expect_warning(xml_find_all(x, "//g:bar"), "Undefined namespace prefix"),
    "evaluation failed"
  )
})

test_that("no matches returns empty nodeset", {
  x <- read_xml("<foo><bar /></foo>")
  expect_length(xml_find_all(x, "//baz"), 0)
})

test_that("xml_find_all returns nodeset or list of nodesets based on flatten", {
  x <- read_xml(
    "<body><p>Some <b>text</b>.</p>
                 <p>Some <b>other</b> <b>text</b>.</p>
                 <p>No bold here!</p></body>"
  )
  y <- xml_find_all(x, ".//p")
  z <- xml_find_all(y, ".//b", flatten = FALSE)
  expect_s3_class(xml_find_all(y, ".//b"), "xml_nodeset")
  expect_type(z, "list")
  expect_s3_class(z[[1L]], "xml_nodeset")
})

# Find num ---------------------------------------------------------------------
test_that("xml_find_num errors with non numeric results", {
  x <- read_xml("<x><y/><y/></x>")
  expect_snapshot(error = TRUE, {
    xml_find_num(x, "//z")
    xml_find_num(x, "//y")
    xml_find_num(x, "1=1")
    xml_find_num(x, "string(5)")
  })
})

test_that("xml_find_num returns a numeric result", {
  x <- read_xml("<x><y>1</y><y/></x>")
  expect_equal(xml_find_num(x, "1 div 0"), Inf)

  expect_equal(xml_find_num(x, "-1 div 0"), -Inf)

  expect_equal(xml_find_num(x, "0 div 0"), NaN)

  expect_equal(xml_find_num(x, "1 div floor(-0.1)"), -1)

  expect_equal(xml_find_num(x, "1 div floor(0)"), Inf)

  expect_equal(xml_find_num(x, "1 div floor(-0)"), -Inf)
})

# Find int ---------------------------------------------------------------------
test_that("xml_find_int errors with non integer results", {
  x <- read_xml("<x><y/><y/></x>")
  expect_snapshot(error = TRUE, {
    xml_find_int(x, "//z")
    xml_find_int(x, "//y")
    xml_find_int(x, "number(1.1)")
  })
})

test_that("xml_find_int returns a integer result", {
  x <- read_xml("<x><y>1</y><y/></x>")

  expect_identical(xml_find_int(x, "1 div floor(-0.1)"), -1L)
  expect_identical(xml_find_int(x, "number(//y)"), 1L)
  expect_identical(xml_find_int(x, "string-length(string('abcd'))"), 4L)
})

# Find chr ---------------------------------------------------------------------
test_that("xml_find_chr errors with non character results", {
  x <- read_xml("<x><y/><y/></x>")
  expect_snapshot(error = TRUE, {
    xml_find_chr(x, "//z")
    xml_find_chr(x, "//y")
    xml_find_chr(x, "1=1")
    xml_find_chr(x, "1+1")
  })
})

test_that("xml_find_chr returns a character result", {
  x <- read_xml("<x><y>1</y><y/></x>")
  expect_equal(xml_find_chr(x, "string(5)"), "5")

  expect_equal(xml_find_chr(x, "string(0.5)"), "0.5")

  expect_equal(xml_find_chr(x, "string(-0.5)"), "-0.5")

  expect_equal(xml_find_chr(x, "concat(\"titi\", \"toto\")"), "tititoto")

  expect_equal(xml_find_chr(x, "substring(\"12345\", 2, 3)"), "234")

  expect_equal(xml_find_chr(x, "substring(\"12345\", 2)"), "2345")

  expect_equal(xml_find_chr(x, "substring(\"12345\", -4)"), "12345")
})

# Find lgl ---------------------------------------------------------------------
test_that("xml_find_lgl errors with non logical results", {
  x <- read_xml("<x><y/><y/></x>")
  expect_snapshot(error = TRUE, {
    xml_find_lgl(x, "//z")
    xml_find_lgl(x, "//y")
    xml_find_lgl(x, "string(5)")
    xml_find_lgl(x, "1+1")
  })
})

test_that("xml_find_lgl returns a logical result", {
  x <- read_xml("<x><y>1</y><y/></x>")

  expect_true(xml_find_lgl(x, "1=1"))

  expect_false(xml_find_lgl(x, "1!=1"))

  expect_true(xml_find_lgl(x, "true()=true()"))

  expect_false(xml_find_lgl(x, "true()=false()"))

  expect_true(xml_find_lgl(x, "'test'='test'"))
})

test_that("Searches with empty inputs retain type stability", {
  empty <- xml_nodeset()

  expect_equal(xml_find_num(empty, "1 div 0"), integer())

  expect_equal(xml_find_chr(empty, "string(0.5)"), character())

  expect_equal(xml_find_lgl(empty, "1=1"), logical())
})

test_that("Searches with entities work (#241)", {
  res <- read_xml(test_path("records.xml"), options = "DTDVALID")

  field1 <- xml_find_all(res, "//field1")

  expect_equal(xml_text(field1), "foo bar Quantitative Consultancy")
})

test_that("A helpful error is given from non-string xpath in xml_find_first/all", {
  node <- read_xml("<a>1</a>")

  expect_error(xml_find_all(node, 1), "XPath must be a string", fixed = TRUE)
  expect_error(xml_find_first(node, 1), "XPath must be a string", fixed = TRUE)
})

test_that("write_xml errors for incorrect directory and with invalid inputs", {
  x <- read_xml("<x/>")
  filename <- "does_not_exist/test.xml"
  expect_error(
    write_xml(x, filename),
    "'does_not_exist' does not exist in current working directory"
  )

  expect_snapshot(error = TRUE, {
    write_xml(x, c("test.xml", "foo"))
  })
})

test_that("write_xml works with relative file paths", {
  x <- read_xml("<x/>")

  filename <- "../test.xml"
  on.exit(unlink(filename))
  write_xml(x, filename, options = "no_declaration")
  expect_identical(readChar(filename, 1000L), "<x/>\n")
})

test_that("write_xml works with no options", {
  x <- read_xml("<x/>")

  filename <- "../test.xml"
  on.exit(unlink(filename))
  write_xml(x, filename, options = NULL)
  expect_identical(
    readChar(filename, 1000L),
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<x/>\n"
  )
})

test_that("write_xml works with an explicit connections", {
  x <- read_xml("<x/>")

  filename <- "../test.xml"
  file <- file(filename, "wb")
  on.exit(unlink(filename))
  write_xml(x, file, options = "no_declaration")
  close(file)
  expect_identical(readChar(filename, 1000L), "<x/>\n")
})

test_that("write_xml works with an implicit connections", {
  x <- read_xml("<x/>")

  filename <- "../test.xml.gz"
  write_xml(x, filename, options = "no_declaration")
  file <- gzfile(filename, "rb")
  on.exit({
    unlink(filename)
    close(file)
  })
  expect_identical(readChar(file, 1000L), "<x/>\n")
})

test_that("write_xml works with nodeset input and files", {
  x <- read_xml("<x><y/><y><z/></y></x>")
  y <- xml_find_all(x, "//y")

  filename <- "../test.xml"
  on.exit(unlink(filename))
  expect_error(
    write_xml(y, filename, options = "no_declaration"),
    "Can only save length 1 node sets"
  )

  write_xml(y[1], filename, options = "no_declaration")
  expect_identical(readChar(filename, 1000L), "<y/>")
})

test_that("write_xml works with nodeset input and connections", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_all(x, "//y")

  filename <- "../test.xml.gz"
  expect_error(
    write_xml(y, filename, options = "no_declaration"),
    "Can only save length 1 node sets"
  )

  expect_snapshot(error = TRUE, {
    write_xml(y[1], c(filename, "foo"))
  })

  write_xml(y[1], filename, options = "no_declaration")
  file <- gzfile(filename, "rb")
  on.exit({
    unlink(filename)
    close(file)
  })
  expect_identical(readChar(file, 1000L), "<y/>")
})

test_that("write_xml works with node input and files", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_first(x, "//y")

  filename <- "../test.xml"
  expect_snapshot(error = TRUE, write_xml(y, c(filename, "foo")))

  write_xml(y, filename, options = "no_declaration")
  on.exit(unlink(filename))
  expect_identical(readChar(filename, 1000L), "<y/>")
})

test_that("write_xml works with node input and connections", {
  x <- read_xml("<x><y/><y/></x>")
  y <- xml_find_first(x, "//y")

  filename <- "../test.xml.gz"
  write_xml(y, filename, options = "no_declaration")
  file <- gzfile(filename, "rb")
  on.exit({
    unlink(filename)
    close(file)
  })
  expect_identical(readChar(file, 1000L), "<y/>")
})

test_that("write_xml returns invisibly", {
  x <- read_xml("<x>foo</x>")
  tf <- tempfile()
  on.exit(unlink(tf))

  res <- withVisible(write_xml(x, tf))

  expect_null(res$value)
  expect_false(res$visible)
})

context("xml_missing")
x <- read_xml("<body>
  <p>Some <b>text</b>.</p>
  <p>Some <b>other</b>.</p>
  <p>No bold text</p>
  </body>")
para <- xml_find_all(x, ".//p")
b <- xml_find_first(para, ".//b")
mss <- b[[3]]

test_that("xml_find returns nodes of class 'xml_missing' for missing nodes", {
    expect_equal(length(b), 3L)
    expect_equal(vapply(b, length, integer(1)), c(2L, 2L, 0L))
    expect_is(mss, "xml_missing")
})

test_that("xml_missing methods return properly for all S3 methods", {

  expect_equal(as.character(mss), NA_character_)
  expect_equal(as_list(mss), list())
  expect_equal(nodeset_apply(mss), xml_nodeset())
  expect_output(print(mss), "\\{xml_missing\\}\n<NA>")
  expect_equal(tree_structure(mss), NA_character_)
  expect_error(write_xml(mss), "Missing data cannot be written")
  expect_error(write_html(mss), "Missing data cannot be written")
  expect_equal(xml_attr(mss), NA_character_)
  expect_equal(xml_attrs(mss), NA_character_)
  expect_equal(xml_find_all(mss), xml_nodeset())
  expect_equal(xml_find_chr(mss), character())
  expect_equal(xml_find_lgl(mss), logical())
  expect_equal(xml_find_num(mss), numeric())
  expect_equal(xml_find_first(mss), xml_missing())
  expect_equal(xml_length(mss), 0L)
  expect_equal(xml_name(mss), NA_character_)
  expect_equal(xml_parent(mss), xml_missing())
  expect_equal(xml_path(mss), NA_character_)
  expect_equal(xml_text(mss), NA_character_)
  expect_equal(xml_type(mss), NA_character_)
  expect_equal(xml_url(mss), NA_character_)
})

test_that("is.na() should return TRUE for xml_missing",
  expect_true(is.na(xml_missing()))
)

context("print")

x <- read_html("lego.html.bz2")
body <- xml_find_first(x, "//body")
divs <- xml_find_all(body, ".//div")[1:10]
quicklinks <- xml_find_first(body, "//div[contains(@div, 'quicklinks')]")

test_that("print method is correct", {
  expect_output_file(print(x), "output/print-xml_document.txt", update = FALSE)

  body <- xml_find_first(x, "//body")
  expect_output_file(print(body), "output/print-xml_node.txt", update = FALSE)

  divs <- xml_find_all(body, ".//div")[1:10]
  expect_output_file(print(divs), "output/print-xml_nodeset.txt", update = FALSE)
})

test_that("xml_structure is correct", {
  quicklinks <- xml_find_first(x, "//div[contains(@div, 'quicklinks')]")
  expect_output_file(html_structure(quicklinks), "output/html_structure.txt", update = FALSE)
})

#' Determine the type of a node.
#'
#' @inheritParams xml_name
#' @export
#' @examples
#' x <- read_xml("<foo> a <b /> <![CDATA[ blah]]></foo>")
#' xml_type(x)
#' xml_type(xml_contents(x))
xml_type <- function(x) {
  UseMethod("xml_type")
}

#' @export
xml_type.xml_missing <- function(x) {
  NA_character_
}

#' @export
xml_type.xml_node <- function(x) {
  xmlElementType[node_type(x$node)]
}

#' @export
xml_type.xml_nodeset <- function(x) {
  types <- vapply(x, function(x) node_type(x$node), integer(1))
  xmlElementType[types]
}

xmlElementType <- c(
  "element",
  "attribute",
  "text",
  "cdata",
  "entity_ref",
  "entity",
  "pi",
  "comment",
  "document",
  "document_type",
  "document_frag",
  "notation",
  "html_document",
  "dtd",
  "element_decl",
  "attribute_decl",
  "entity_decl",
  "namespace_decl",
  "xinclude_start",
  "xinclude_end",
  "docb_document"
)

#' Extract or modify the text
#'
#' \code{xml_text} returns a character vector, \code{xml_double} returns a
#' numeric vector, \code{xml_integer} returns an integer vector.
#' @inheritParams xml_name
#' @param trim If \code{TRUE} will trim leading and trailing spaces.
#' @return A character vector, the same length as x.
#' @examples
#' x <- read_xml("<p>This is some text. This is <b>bold!</b></p>")
#' xml_text(x)
#' xml_text(xml_children(x))
#'
#' x <- read_xml("<x>This is some text. <x>This is some nested text.</x></x>")
#' xml_text(x)
#' xml_text(xml_find_all(x, "//x"))
#'
#' x <- read_xml("<p>   Some text    </p>")
#' xml_text(x, trim = TRUE)
#'
#' # xml_double() and xml_integer() are useful for extracting numeric
#' attributes
#' x <- read_xml("<plot><point x='1' y='2' /><point x='2' y='1' /></plot>")
#' xml_integer(xml_find_all(x, "//@x"))
#' @export
xml_text <- function(x, trim = FALSE) {
  UseMethod("xml_text")
}

#' @export
xml_text.xml_missing <- function(x, trim = FALSE) {
  NA_character_
}

#' @export
xml_text.xml_node <- function(x, trim = FALSE) {
  res <- node_text(x$node)
  if (isTRUE(trim)) {
    res <- sub("^[[:space:]\u00a0]+", "", res)
    res <- sub("[[:space:]\u00a0]+$", "", res)
  }
  res
}

#' @export
xml_text.xml_nodeset <- function(x, trim = FALSE) {
  vapply(x, xml_text, trim = trim, FUN.VALUE = character(1))
}

#' @rdname xml_text
#' @param value character vector with replacement text.
#' @export
`xml_text<-` <- function(x, value) {
  UseMethod("xml_text<-")
}

#' @export
`xml_text<-.xml_nodeset` <- function(x, value) {
  if (length(x) == 0) {
    return(x)
  }
  # We need to do the modification in reverse order as the modification can
  # potentially delete nodes
  Map(`xml_text<-`, rev(x), rev(value))

  # what to return here, setting the text could invalidate some nodes in
  # the nodeset having pointers to free'd memory.
  x
}

#' @export
`xml_text<-.xml_node` <- function(x, value) {
  if (xml_type(x) != "text") {
    text_child <- xml_find_first(x, ".//text()[1]", ns = character())
    if (inherits(text_child, "xml_missing")) {
      node_append_content(x$node, value)
    } else {
      node_set_content(text_child$node, value)
    }
  } else {
    node_set_content(x$node, value)
  }

  x
}

#' @export
`xml_text<-.xml_missing` <- function(x, value) {
  NA_character_
}

#' @export
#' @rdname xml_text
`xml_set_text` <- `xml_text<-`

#' @rdname xml_text
#' @export
xml_double <- function(x) {
  UseMethod("xml_double")
}

#' @export
xml_double.xml_missing <- function(x) {
  NA_real_
}

#' @export
xml_double.xml_node <- function(x) {
  as.numeric(xml_text(x))
}

#' @export
xml_double.xml_nodeset <- function(x) {
  vapply(x, xml_double, numeric(1))
}

#' @export
xml_integer <- function(x) {
  UseMethod("xml_integer")
}

#' @export
xml_integer.xml_missing <- function(x) {
  NA_integer_
}

#' @rdname xml_text
#' @export
xml_integer <- function(x) {
  UseMethod("xml_integer")
}

#' @export
xml_integer.xml_node <- function(x) {
  as.integer(xml_text(x))
}

#' @export
xml_integer.xml_nodeset <- function(x) {
  vapply(x, xml_integer, integer(1))
}

test_that("xml_text returns only text without markup", {
  x <- read_xml("<p>This is some text. This is <b>bold!</b></p>")

  expect_identical(xml_text(x), "This is some text. This is bold!")

  expect_identical(xml_text(xml_children(x)), "bold!")
})

test_that("xml_text works properly with xml_nodeset objects", {
  x <- read_xml("<body>
    <p>Some <b>text</b>.</p>
    <p>Some <i>other</i>.</p>
    <p>No bold text</p>
    </body>")

  children <- xml_children(x)
  x <- xml_find_first(children, ".//b|.//i")

  expect_identical(
    xml_text(x),
    c("text", "other", NA)
  )
})

test_that("xml_text<- and xml_set_text work properly with xml_nodeset objects", {
  x <- read_xml("<x>This is some text. <x>This is some nested text.</x></x>")

  expect_identical(xml_text(x), "This is some text. This is some nested text.")

  xml_text(x) <- "test"
  expect_identical(xml_text(x), "testThis is some nested text.")
  xml_set_text(x, "test2")
  expect_identical(xml_text(x), "test2This is some nested text.")
})

test_that("xml_text trims whitespace if requested, including non-breaking spaces", {
  x <- read_html("<p>   Some text &euro;  &nbsp;</p>")
  expect_identical(
    xml_text(x),
    "   Some text \u20ac  \u00a0"
  )

  expect_identical(
    xml_text(x, trim = TRUE),
    "Some text \u20ac"
  )

  x2 <- read_html("<body><p>   Some text &euro;  &nbsp;</p><p>  and more &euro; text   &nbsp;</body>")
  expect_identical(
    xml_text(xml_find_all(x2, ".//p"), trim = TRUE),
    c("Some text \u20ac", "and more \u20ac text")
  )
})

test_that("xml_integer() returns an integer vector", {
  x <- read_xml("<plot><point x='1' y='2' /><point x='2' y='1' /></plot>")

  expect_identical(
    xml_integer(xml_find_all(x, "//@x")),
    c(1L, 2L)
  )
})


test_that("xml_double() returns a numeric vector", {
  x <- read_xml("<earth><point latitude = '42.3466456' longitude = '-71.0390351' /><point latitude = '-36.8523378' longitude = '174.7691073' /></earth>")

  expect_identical(xml_double(xml_find_all(x, "//@latitude")), c(42.3466456, -36.8523378))
})

test_that("xml_text returns only text without markup", {
  x <- read_xml("<p>This is some text. This is <b>bold!</b></p>")

  expect_identical(xml_text(x), "This is some text. This is bold!")

  expect_identical(xml_text(xml_children(x)), "bold!")
})

test_that("xml_text works properly with xml_nodeset objects", {
  x <- read_xml(
    "<body>
    <p>Some <b>text</b>.</p>
    <p>Some <i>other</i>.</p>
    <p>No bold text</p>
    </body>"
  )

  children <- xml_children(x)
  x <- xml_find_first(children, ".//b|.//i")

  expect_identical(
    xml_text(x),
    c("text", "other", NA)
  )
})

test_that("xml_text<- and xml_set_text work properly with xml_nodeset objects", {
  x <- read_xml("<x>This is some text. <x>This is some nested text.</x></x>")

  expect_identical(xml_text(x), "This is some text. This is some nested text.")

  xml_text(x) <- "test"
  expect_identical(xml_text(x), "testThis is some nested text.")
  xml_set_text(x, "test2")
  expect_identical(xml_text(x), "test2This is some nested text.")
})

test_that("xml_text trims whitespace if requested, including non-breaking spaces", {
  x <- read_html("<p>   Some text &euro;  &nbsp;</p>")
  expect_identical(
    xml_text(x),
    "   Some text \u20ac  \u00a0"
  )

  expect_identical(
    xml_text(x, trim = TRUE),
    "Some text \u20ac"
  )

  x2 <- read_html(
    "<body><p>   Some text &euro;  &nbsp;</p><p>  and more &euro; text   &nbsp;</body>"
  )
  expect_identical(
    xml_text(xml_find_all(x2, ".//p"), trim = TRUE),
    c("Some text \u20ac", "and more \u20ac text")
  )
})

test_that("xml_integer() returns an integer vector", {
  x <- read_xml("<plot><point x='1' y='2' /><point x='2' y='1' /></plot>")

  expect_identical(
    xml_integer(xml_find_all(x, "//@x")),
    c(1L, 2L)
  )
})


test_that("xml_double() returns a numeric vector", {
  x <- read_xml(
    "<earth><point latitude = '42.3466456' longitude = '-71.0390351' /><point latitude = '-36.8523378' longitude = '174.7691073' /></earth>"
  )

  expect_identical(
    xml_double(xml_find_all(x, "//@latitude")),
    c(42.3466456, -36.8523378)
  )
})

# node -------------------------------------------------------------------------

xml_node <- function(node = NULL, doc = NULL) {
  if (inherits(node, "xml_node")) {
    node
  } else {
    out <- list(node = node, doc = doc)
    class(out) <- "xml_node"
    out
  }
}

#' @export
as.character.xml_node <- function(
  x,
  ...,
  options = "format",
  encoding = "UTF-8"
) {
  options <- parse_options(options, xml_save_options())
  .Call(node_write_character, x$node, encoding, options)
}

#' @export
print.xml_node <- function(x, width = getOption("width"), max_n = 20, ...) {
  cat("{", doc_type(x), "_node}\n", sep = "")
  cat(format(x), "\n", sep = "")
  show_nodes(xml_children(x), width = width, max_n = max_n)
}

#' @export
is.na.xml_node <- function(x) {
  FALSE
}

if (!file.exists("../windows/libxml2/include/libxml2/libxml")) {
  unlink("../windows", recursive = TRUE)
  url <- if (grepl("aarch", R.version$platform)) {
    "https://github.com/r-windows/bundles/releases/download/libxml2-2.11.5/libxml2-2.11.5-clang-aarch64.tar.xz"
  } else if (grepl("clang", Sys.getenv('R_COMPILED_BY'))) {
    "https://github.com/r-windows/bundles/releases/download/libxml2-2.11.5/libxml2-2.11.5-clang-x86_64.tar.xz"
  } else if (getRversion() >= "4.2") {
    "https://github.com/r-windows/bundles/releases/download/libxml2-2.11.5/libxml2-2.11.5-ucrt-x86_64.tar.xz"
  } else {
    "https://github.com/rwinlib/libxml2/archive/v2.10.3.tar.gz"
  }
  download.file(url, basename(url), quiet = TRUE)
  dir.create("../windows", showWarnings = FALSE)
  untar(basename(url), exdir = "../windows", tar = 'internal')
  unlink(basename(url))
  setwd("../windows")
  file.rename(list.files(), 'libxml2')
}

path_to_connection <- function(path, check = c("file", "dir")) {
  check <- match.arg(check)

  if (!is.character(path) || length(path) != 1L)
    return(path)

  if (is_url(path)) {
    if (requireNamespace("curl", quietly = TRUE)) {
      return(curl::curl(path))
    } else {
      return(url(path))
    }
  }

  if (check == "file") {
    path <- check_path(path)
  } else {
    path <- file.path(check_path(dirname(path)), basename(path))
  }
  switch(tools::file_ext(path),
    gz = gzfile(path, ""),
    bz2 = bzfile(path, ""),
    xz = xzfile(path, ""),
    zip = zipfile(path, ""),
    path
  )
}

is_url <- function(path) {
  grepl("^(http|ftp)s?://", path)
}

check_path <- function(path) {
  if (file.exists(path))
    return(normalizePath(path, "/", mustWork = FALSE))

  stop("'", path, "' does not exist",
    if (!is_absolute_path(path))
      paste0(" in current working directory ('", getwd(), "')"),
    ".",
    call. = FALSE
  )
}

is_absolute_path <- function(path) {
  grepl("^(/|[A-Za-z]:|\\\\|~)", path)
}

zipfile <- function(path, open = "r") {
  files <- utils::unzip(path, list = TRUE)
  file <- files$Name[[1]]

  if (nrow(files) > 1) {
    message("Multiple files in zip: reading '", file, "'")
  }

  unz(path, file, open = open)
}

#' Show the structure of an html/xml document.
#'
#' Show the structure of an html/xml document without displaying any of
#' the values. This is useful if you want to get a high level view of the
#' way a document is organised. Compared to `xml_structure`,
#' `html_structure` prints the id and class attributes.
#'
#' @param x HTML/XML document (or part there of)
#' @param indent Number of spaces to ident
#' @inheritParams base::cat
#' @export
#' @examples
#' xml_structure(read_xml("<a><b><c/><c/></b><d/></a>"))
#'
#' rproj <- read_html(system.file("extdata", "r-project.html", package = "xml2"))
#' xml_structure(rproj)
#' xml_structure(xml_find_all(rproj, ".//p"))
#'
#' h <- read_html("<body><p id = 'a'></p><p class = 'c d'></p></body>")
#' html_structure(h)
xml_structure <- function(x, indent = 2, file = "") {
  cat(file = file)
  tree_structure(x, indent = indent, html = FALSE, file = file)
}

#' @export
#' @rdname xml_structure
html_structure <- function(x, indent = 2, file = "") {
  cat(file = file)
  tree_structure(x, indent = indent, html = TRUE, file = file)
}

tree_structure <- function(x, indent = 2, html = FALSE, file = "") {
  UseMethod("tree_structure")
}

#' @export
tree_structure.xml_missing <- function(x, indent = 2, html = FALSE, file = "") {
  NA_character_
}

#' @export
tree_structure.xml_nodeset <- function(x, indent = 2, html = FALSE, file = "") {
  for (i in seq_along(x)) {
    cat("[[", i, "]]\n", sep = "", file = file, append = TRUE)
    print_xml_structure(x[[i]], indent = indent, html = html, file = file)
    cat("\n", file = file, append = TRUE)
  }

  invisible()
}

#' @export
tree_structure.xml_node <- function(x, indent = 2, html = FALSE, file = "") {
  print_xml_structure(x, indent = indent, html = html, file = file)
  invisible()
}

print_xml_structure <- function(
  x,
  prefix = 0,
  indent = 2,
  html = FALSE,
  file = ""
) {
  padding <- paste(rep(" ", prefix), collapse = "")
  type <- xml_type(x)

  if (type == "element") {
    attr <- xml_attrs(x)
    if (html) {
      html_attrs <- list()
      if ("id" %in% names(attr)) {
        html_attrs$id <- paste0("#", attr[["id"]])
        attr <- attr[setdiff(names(attr), "id")]
      }

      if ("class" %in% names(attr)) {
        html_attrs$class <- paste0(".", gsub(" ", ".", attr[["class"]]))
        attr <- attr[setdiff(names(attr), "class")]
      }

      attr_str <- paste(unlist(html_attrs), collapse = " ")
    } else {
      attr_str <- ""
    }

    if (length(attr) > 0) {
      attr_str <- paste0(
        attr_str,
        " [",
        paste0(names(attr), collapse = ", "),
        "]"
      )
    }

    node <- paste0("<", xml_name(x), attr_str, ">")

    cat(padding, node, "\n", sep = "", file = file, append = TRUE)
    lapply(
      xml_contents(x),
      print_xml_structure,
      prefix = prefix + indent,
      indent = indent,
      html = html,
      file = file
    )
  } else {
    cat(padding, "{", type, "}\n", sep = "", file = file, append = TRUE)
  }
}

#' Write XML or HTML to disk.
#'
#' This writes out both XML and normalised HTML. The default behavior will
#' output the same format which was read. If you want to force output pass
#' `option = "as_xml"` or `option = "as_html"` respectively.
#'
#' @param x A document or node to write to disk. It's not possible to
#'   save nodesets containing more than one node.
#' @param file Path to file or connection to write to.
#' @param encoding The character encoding to use in the document. The default
#' encoding is \sQuote{UTF-8}. Available encodings are specified at
#' <http://xmlsoft.org/html/libxml-encoding.html#xmlCharEncoding>.
#' @param options default: \sQuote{format}. Zero or more of
#' \Sexpr[results=rd, stage=build]{xml2:::describe_options(xml2:::xml_save_options())}
#' @param ... additional arguments passed to methods.
#' @export
#' @examples
#' h <- read_html("<p>Hi!</p>")
#'
#' tmp <- tempfile(fileext = ".xml")
#' write_xml(h, tmp, options = "format")
#' readLines(tmp)
#'
#' # write formatted HTML output
#' write_html(h, tmp, options = "format")
#' readLines(tmp)
write_xml <- function(x, file, ...) {
  UseMethod("write_xml")
}

#' @export
write_xml.xml_missing <- function(x, file, ...) {
  cli::cli_abort("Missing data cannot be written.")
}

#' @rdname write_xml
#' @export
write_xml.xml_document <- function(
  x,
  file,
  ...,
  options = "format",
  encoding = "UTF-8"
) {
  options <- parse_options(options, xml_save_options())
  file <- path_to_connection(file, check = "dir")

  if (inherits(file, "connection")) {
    if (!isOpen(file)) {
      open(file, "wb")
      on.exit(close(file))
    }
    .Call(doc_write_connection, x$doc, file, encoding, options)
  } else {
    check_string(file)
    .Call(doc_write_file, x$doc, file, encoding, options)
  }

  invisible()
}

#' @export
write_xml.xml_nodeset <- function(
  x,
  file,
  ...,
  options = "format",
  encoding = "UTF-8"
) {
  if (length(x) != 1) {
    cli::cli_abort("Can only save length 1 node sets.")
  }

  options <- parse_options(options, xml_save_options())
  file <- path_to_connection(file, check = "dir")

  if (inherits(file, "connection")) {
    if (!isOpen(file)) {
      open(file, "wb")
      on.exit(close(file))
    }
    .Call(node_write_connection, x[[1]]$node, file, encoding, options)
  } else {
    check_string(file)
    .Call(node_write_file, x[[1]]$node, file, encoding, options)
  }

  invisible()
}

#' @export
write_xml.xml_node <- function(
  x,
  file,
  ...,
  options = "format",
  encoding = "UTF-8"
) {
  options <- parse_options(options, xml_save_options())

  file <- path_to_connection(file, check = "dir")
  if (inherits(file, "connection")) {
    if (!isOpen(file)) {
      open(file, "wb")
      on.exit(close(file))
    }
    .Call(node_write_connection, x$node, file, encoding, options)
  } else {
    check_string(file)
    .Call(node_write_file, x$node, file, encoding, options)
  }

  invisible()
}


#' @export
#' @rdname write_xml
write_html <- function(x, file, ...) {
  UseMethod("write_html")
}

#' @export
write_html.xml_missing <- function(x, file, ...) {
  cli::cli_abort("Missing data cannot be written.")
}

#' @rdname write_xml
#' @export
write_html.xml_document <- write_xml.xml_document

#' @export
write_html.xml_nodeset <- write_xml.xml_nodeset

#' @export
write_html.xml_node <- write_xml.xml_node

#' The URL of an XML document
#'
#' This is useful for interpreting relative urls with [url_relative()].
#'
#' @param x A node or document.
#' @return A character vector of length 1. Returns `NA` if the name is
#'   not set.
#' @export
#' @examples
#' catalog <- read_xml(xml2_example("cd_catalog.xml"))
#' xml_url(catalog)
#'
#' x <- read_xml("<foo/>")
#' xml_url(x)
xml_url <- function(x) {
  UseMethod("xml_url")
}

#' @export
xml_url.xml_missing <- function(x) {
  NA_character_
}

#' @export
xml_url.xml_node <- function(x) {
  .Call(doc_url, x$doc)
}

#' @export
xml_url.xml_nodeset <- function(x) {
  vapply(x, function(x) .Call(doc_url, x), character(1))
}

#' Convert between relative and absolute urls.
#'
#' @param x A character vector of urls relative to that base
#' @param base A string giving a base url.
#' @return A character vector of urls
#' @seealso \code{\link{xml_url}} to retrieve the URL associated with a document
#' @export
#' @examples
#' url_absolute(c(".", "..", "/", "/x"), "http://hadley.nz/a/b/c/d")
#'
#' url_relative("http://hadley.nz/a/c", "http://hadley.nz")
#' url_relative("http://hadley.nz/a/c", "http://hadley.nz/")
#' url_relative("http://hadley.nz/a/c", "http://hadley.nz/a/b")
#' url_relative("http://hadley.nz/a/c", "http://hadley.nz/a/b/")
#' @export
url_absolute <- function(x, base) {
  .Call(url_absolute_, x, base)
}

#' @rdname url_absolute
#' @export
url_relative <- function(x, base) {
  .Call(url_relative_, x, base)
}

#' Escape and unescape urls.
#'
#' @param x A character vector of urls.
#' @param reserved A string containing additional characters to avoid escaping.
#' @export
#' @examples
#' url_escape("a b c")
#' url_escape("a b c", "")
#'
#' url_unescape("a%20b%2fc")
#' url_unescape("%C2%B5")
url_escape <- function(x, reserved = "") {
  .Call(url_escape_, x, reserved)
}

#' @rdname url_escape
#' @export
url_unescape <- function(x) {
  .Call(url_unescape_, x)
}

#' Parse a url into its component pieces.
#'
#' @param x A character vector of urls.
#' @return A dataframe with one row for each element of \code{x} and
#'   columns: scheme, server, port, user, path, query, fragment.
#' @export
#' @examples
#' url_parse("http://had.co.nz/")
#' url_parse("http://had.co.nz:1234/")
#' url_parse("http://had.co.nz:1234/?a=1&b=2")
#' url_parse("http://had.co.nz:1234/?a=1&b=2#def")
url_parse <- function(x) {
  .Call(url_parse_, x)
}

#' Retrieve an attribute.
#'
#' `xml_attrs()` retrieves all attributes values as a named character
#' vector, `xml_attrs() <-` or `xml_set_attrs()` sets all attribute
#' values. `xml_attr()` retrieves the value of single attribute and
#' `xml_attr() <-` or `xml_set_attr()` modifies its value. If the
#' attribute doesn't exist, it will return `default`, which defaults to
#' `NA`. `xml_has_attr()` tests if an attribute is present.
#'
#' @inheritParams xml_name
#' @param attr Name of attribute to extract.
#' @param default Default value to use when attribute is not present.
#' @return `xml_attr()` returns a character vector. `NA` is used
#'  to represent of attributes that aren't defined.
#'
#'  `xml_has_attr()` returns a logical vector.
#'
#'  `xml_attrs()` returns a named character vector if `x` x is single
#'  node, or a list of character vectors if given a nodeset
#' @export
#' @examples
#' x <- read_xml("<root id='1'><child id ='a' /><child id='b' d='b'/></root>")
#' xml_attr(x, "id")
#' xml_attr(x, "apple")
#' xml_attrs(x)
#'
#' kids <- xml_children(x)
#' kids
#' xml_attr(kids, "id")
#' xml_has_attr(kids, "id")
#' xml_attrs(kids)
#'
#' # Missing attributes give missing values
#' xml_attr(xml_children(x), "d")
#' xml_has_attr(xml_children(x), "d")
#'
#' # If the document has a namespace, use the ns argument and
#' # qualified attribute names
#' x <- read_xml('
#'  <root xmlns:b="http://bar.com" xmlns:f="http://foo.com">
#'    <doc b:id="b" f:id="f" id="" />
#'  </root>
#' ')
#' doc <- xml_children(x)[[1]]
#' ns <- xml_ns(x)
#'
#' xml_attrs(doc)
#' xml_attrs(doc, ns)
#'
#' # If you don't supply a ns spec, you get the first matching attribute
#' xml_attr(doc, "id")
#' xml_attr(doc, "b:id", ns)
#' xml_attr(doc, "id", ns)
#'
#' # Can set a single attribute with `xml_attr() <-` or `xml_set_attr()`
#' xml_attr(doc, "id") <- "one"
#' xml_set_attr(doc, "id", "two")
#'
#' # Or set multiple attributes with `xml_attrs()` or `xml_set_attrs()`
#' xml_attrs(doc) <- c("b:id" = "one", "f:id" = "two", "id" = "three")
#' xml_set_attrs(doc, c("b:id" = "one", "f:id" = "two", "id" = "three"))
xml_attr <- function(x, attr, ns = character(), default = NA_character_) {
  .Call(node_attr, x, attr, as.character(default), ns)
}

#' @export
#' @rdname xml_attr
xml_has_attr <- function(x, attr, ns = character()) {
  !is.na(xml_attr(x, attr, ns = ns))
}

#' @export
#' @rdname xml_attr
xml_attrs <- function(x, ns = character()) {
  .Call(node_attrs, x, nsMap = ns)
}

#' @param value character vector of new value.
#' @rdname xml_attr
#' @export
`xml_attr<-` <- function(x, attr, ns = character(), value) {
  UseMethod("xml_attr<-")
}

#' @export
`xml_attr<-.xml_node` <- function(x, attr, ns = character(), value) {
  if (is.null(value)) {
    .Call(node_remove_attr, x$node, attr, ns)
  } else {
    value <- as.character(value)
    .Call(node_set_attr, x$node, attr, value, ns)
  }
  x
}

#' @export
`xml_attr<-.xml_nodeset` <- function(x, attr, ns = character(), value) {
  if (length(x) == 0) {
    return(x)
  }

  if (length(value) == 0) {
    value <- list(value)
  }

  mapply(
    `xml_attr<-`,
    x,
    attr = attr,
    value = value,
    SIMPLIFY = FALSE,
    MoreArgs = list(ns = ns)
  )
  x
}

#' @export
`xml_attr<-.xml_missing` <- function(x, attr, ns = character(), value) {
  x
}

#' @rdname xml_attr
#' @export
xml_set_attr <- function(x, attr, value, ns = character()) {
  UseMethod("xml_set_attr")
}

# This function definition is used for all methods, we need to rearrange the `ns`
# argument to be at the end of the set function
set_attr_fun <- function(x, attr, value, ns = character()) {
  xml_attr(x = x, attr = attr, ns = ns) <- value
}

#' @export
xml_set_attr.xml_node <- set_attr_fun

#' @export
xml_set_attr.xml_nodeset <- set_attr_fun

#' @export
xml_set_attr.xml_missing <- set_attr_fun

#' @rdname xml_attr
#' @export
`xml_attrs<-` <- function(x, ns = character(), value) {
  UseMethod("xml_attrs<-")
}

#' @export
`xml_attrs<-.xml_node` <- function(x, ns = character(), value) {
  if (!is_named(value)) {
    cli::cli_abort("{.arg value} must be a named character vector or `NULL`")
  }

  attrs <- names(value)

  # as.character removes all attributes (including names)
  value <- stats::setNames(as.character(value), attrs)

  current_attrs <- names(xml_attrs(x, ns = ns))

  existing <- intersect(current_attrs, attrs)
  new <- setdiff(attrs, current_attrs)
  removed <- setdiff(current_attrs, attrs)

  # replace existing attributes and add new ones
  Map(
    function(attr, val) {
      xml_attr(x, attr, ns) <- val
    },
    attr = c(existing, new),
    value[c(existing, new)]
  )

  # Remove attributes which no longer exist
  Map(
    function(attr) {
      xml_attr(x, attr, ns) <- NULL
    },
    attr = removed
  )

  x
}

#' @export
`xml_attrs<-.xml_nodeset` <- function(x, ns = character(), value) {
  if (length(x) == 0) {
    return(x)
  }
  if (!is.list(ns)) {
    ns <- list(ns)
  }
  if (!is.list(value)) {
    value <- list(value)
  }
  if (!all(vapply(value, is_named, logical(1)))) {
    cli::cli_abort("{.arg {value}} must be a list of named character vectors.")
  }

  Map(`xml_attrs<-`, x, ns, value)

  x
}

#' @export
`xml_attrs<-.xml_missing` <- function(x, ns = character(), value) {
  x
}

#' @rdname xml_attr
#' @export
xml_set_attrs <- function(x, value, ns = character()) {
  UseMethod("xml_set_attrs")
}

# This function definition is used for all methods, we need to rearrange the `ns`
# argument to be at the end of the set function
set_attrs_fun <- function(x, value, ns = character()) {
  xml_attrs(x = x, ns = ns) <- value
}

#' @export
xml_set_attrs.xml_node <- set_attrs_fun

#' @export
xml_set_attrs.xml_nodeset <- set_attrs_fun

#' @export
xml_set_attrs.xml_missing <- set_attrs_fun

test_that("missing attributes returned as NA by default", {
  x <- read_xml("<x/>")
  expect_equal(xml_attr(x, "id"), NA_character_)
})

test_that("missing attributes returned as NA", {
  x <- read_xml("<x/>")
  expect_equal(xml_attr(x, "id", default = 1), "1")
})

test_that("attributes are correctly found", {
  x <- read_xml("<x id='1' />")

  expect_true(xml_has_attr(x, "id"))

  expect_false(xml_has_attr(x, "id2"))
})

test_that("returning an attribute node prints properly", {
  x <- read_xml("<a><b c='1' /></a>")

  t1 <- xml_find_first(x, "//@c")

  expect_equal(format(t1), "<c>")
})

# Namespaces -------------------------------------------------------------------

# Default namespace doesn't apply to attributes

test_that("qualified names returned when ns given", {
  x <- read_xml(test_path("ns-multiple.xml"))
  ns <- xml_ns(x)

  bars <- xml_children(xml_children(x))
  attr <- xml_attrs(bars, ns)

  expect_named(attr[[1]], "f:id")
  expect_named(attr[[2]], "g:id")
})


x <- read_xml('
 <root xmlns:b="http://bar.com" xmlns:f="http://foo.com">
   <doc b:id="b" f:id="f" id="" />
   <doc b:id="b" f:id="f" id="" />
 </root>
')
doc <- xml_children(x)[[1]]
docs <- xml_find_all(x, "//doc")
ns <- xml_ns(x)

test_that("qualified attributes get own values", {
  expect_equal(xml_attrs(doc, ns), c("b:id" = "b", "f:id" = "f", "id" = ""))
})

test_that("unqualified name gets unnamespace attribute", {
  expect_equal(xml_attr(doc, "id", ns), "")
})

test_that("namespace names gets namespaced attribute", {
  expect_equal(xml_attr(doc, "b:id", ns), "b")
  expect_equal(xml_attr(doc, "f:id", ns), "f")
})

test_that("xml_attr<- modifies properties", {
  xml_attr(doc, "id", ns) <- "test"
  expect_equal(xml_attr(doc, "id", ns), "test")

  xml_attr(doc, "b:id", ns) <- "b_test"
  expect_equal(xml_attr(doc, "b:id", ns), "b_test")

  xml_attr(doc, "f:id", ns) <- "f_test"
  expect_equal(xml_attr(doc, "f:id", ns), "f_test")

  xml_attr(docs, "f:id", ns) <- "f_test2"
  expect_equal(xml_attr(docs, "f:id", ns), c("f_test2", "f_test2"))

  xml_attr(docs, "f:id", ns) <- NULL
  expect_equal(xml_attr(docs, "f:id", ns), c(NA_character_, NA_character_))
})

test_that("xml_attr<- recycles values", {
  x <- read_xml("<x><a b='c'/><a b='d'/></x>")
  a <- xml_find_all(x, "a")
  xml_attr(a, "b") <- c("e", "f")
  expect_equal(xml_attr(a, "b"), c("e", "f"))
})

test_that("xml_attrs<- modifies all attributes", {
  expect_error(xml_attrs(doc) <- 1, "`value` must be a named character vector or `NULL`")
  expect_error(xml_attrs(doc) <- "test", "`value` must be a named character vector or `NULL`")

  xml_attrs(doc, ns) <- c("b:id" = "b", "f:id" = "f", "id" = "test")
  expect_equal(xml_attrs(doc, ns), c("b:id" = "b", "id" = "test", "f:id" = "f"))

  xml_attrs(doc, ns) <- c("b:id" = "b", "f:id" = "f")
  expect_equal(xml_attrs(doc, ns), c("b:id" = "b", "f:id" = "f"))

  xml_attrs(doc, ns) <- c("b:id" = "b", "id" = "test")
  expect_equal(xml_attrs(doc, ns), c("b:id" = "b", "id" = "test"))

  expect_snapshot(error = TRUE, xml_attrs(docs) <- "test")

  xml_attrs(docs, ns) <- c("b:id" = "b", "id" = "test")
  expect_equal(
    xml_attrs(docs, ns),
    list(
      c("b:id" = "b", "id" = "test"),
      c("b:id" = "b", "id" = "test")
    )
  )

  xml_attrs(docs, ns) <- NULL
  expect_equal(xml_attrs(docs, ns), list(setNames(character(0), character()), setNames(character(0), character())))
})

test_that("xml_attr<- accepts non-character values", {
  x <- read_xml("<svg><rect /></svg>")
  svg <- xml_root(x)

  xml_attr(svg, "width") <- 8L
  expect_equal(xml_attr(svg, "width"), "8")

  xml_attr(svg, "height") <- 12.5
  expect_equal(xml_attr(svg, "height"), "12.5")

  expect_equal(xml_attrs(svg), c(width = "8", height = "12.5"))

  xml_attrs(svg) <- c(width = 14L, height = 23.45)
  expect_equal(xml_attrs(svg), c(width = "14", height = "23.45"))
})

test_that("xml_attr<- can set empty strings, and removes attributes with NULL", {
  x <- read_xml("<x><y /></x>")
  xml_attr(x, "test") <- ""
  expect_equal(xml_attr(x, "test"), "")

  xml_attr(x, "test") <- NULL
  expect_equal(xml_attr(x, "test"), NA_character_)
})

test_that("xml_attr<- removes namespaces if desired", {
  xml_attr(x, "xmlns:b") <- NULL

  expect_equal(xml_attrs(x), c("xmlns:f" = "http://foo.com"))
})

test_that("xml_attr<- removes namespaces if desired", {
  x <- read_xml("<a xmlns = 'tag:foo'><b/></a>")

  # cannot find //b with a default namespace
  expect_length(xml_find_all(x, "//b"), 0)

  # unless we specify it explicitly
  expect_length(xml_find_all(x, "//b"), 0)
  expect_length(xml_find_all(x, "//d1:b", xml_ns(x)), 1)

  # but can find it once we remove the namespace
  xml_attr(x, "xmlns") <- NULL
  expect_length(xml_find_all(x, "//b"), 1)

  # and add the old namespace back
  xml_attr(x, "xmlns") <- "tag:foo"
  expect_equal(xml_attr(x, "xmlns"), "tag:foo")
  expect_length(xml_find_all(x, "//b"), 0)
  expect_length(xml_find_all(x, "//d1:b", xml_ns(x)), 1)

  expect_equal(xml_attr(x, "xmlns"), "tag:foo")
})

test_that("xml_attr<- removes prefixed namespaces if desired", {
  x <- read_xml("<a xmlns:pre = 'tag:foo'><pre:b/></a>")

  # cannot find //b with a prefixed namespace
  expect_length(xml_find_all(x, "//b"), 0)

  # unless we specify it explicitly
  expect_length(xml_find_all(x, "//b"), 0)
  expect_length(xml_find_all(x, "//pre:b", xml_ns(x)), 1)

  # but can find it once we remove the namespace
  xml_attr(x, "xmlns:pre") <- NULL
  expect_length(xml_find_all(x, "//b"), 1)

  # and add the old namespace back
  xml_attr(x, "xmlns:pre") <- "tag:foo"
  xml_set_namespace(xml_children(x)[[1]], "pre")
  expect_equal(xml_attr(x, "xmlns:pre"), "tag:foo")
  expect_length(xml_find_all(x, "//b"), 0)
  expect_length(xml_find_all(x, "//pre:b", xml_ns(x)), 1)

  expect_equal(xml_attr(x, "xmlns:pre"), "tag:foo")
})

test_that("xml_set_attr works identically to xml_attr<-", {
  content <- "<a><b><c/></b><b/></a>"
  x <- read_xml(content)
  y <- read_xml(content)

  xml_attr(x, "a") <- "test"
  xml_set_attr(y, "a", "test")

  expect_equal(as.character(x), as.character(y))

  bx <- xml_find_all(x, "//b")
  by <- xml_find_all(y, "//b")

  xml_attr(bx, "b") <- "test2"
  xml_set_attr(by, "b", "test2")

  expect_equal(as.character(x), as.character(y))

  # No errors for xml_missing
  mss <- xml_find_first(bx, "./c")
  expect_no_error(xml_attr(mss[[2]], "b") <- "blah")
  expect_no_error(xml_set_attr(mss[[2]], "b", "blah"))
})

test_that("xml_set_attrs works identically to xml_attrs<-", {
  content <- "<a><b><c/></b><b/></a>"
  x <- read_xml(content)
  y <- read_xml(content)

  xml_attrs(x) <- c(a = "test")
  xml_set_attrs(y, c(a = "test"))

  expect_equal(as.character(x), as.character(y))

  bx <- xml_find_all(x, "//b")
  by <- xml_find_all(y, "//b")

  xml_attrs(bx) <- c(b = "test2")
  xml_set_attrs(by, c(b = "test2"))

  expect_equal(as.character(x), as.character(y))

  # No errors for xml_missing
  mss <- xml_find_first(bx, "./c")
  expect_no_error(xml_attrs(mss[[2]]) <- c("b" = "blah"))
  expect_no_error(xml_set_attrs(mss[[2]], c("b" = "blah")))
})

test_that("xml_set_attr can set the same namespace multiple times", {
  doc <- xml_new_root("foo")
  xml_set_attr(doc, "xmlns:bar", "http://a/namespace")
  xml_set_attr(doc, "xmlns:bar", "http://b/namespace")
  expect_equal(xml_attr(doc, "xmlns:bar"), "http://b/namespace")
})

context("xml_attrs")

test_that("missing attributes returned as NA by default", {
  x <- read_xml("<x/>")
  expect_equal(xml_attr(x, "id"), NA_character_)
})

test_that("missing attributes returned as NA", {
  x <- read_xml("<x/>")
  expect_equal(xml_attr(x, "id", default = 1), "1")
})

test_that("attributes are correctly found", {
  x <- read_xml("<x id='1' />")

  expect_true(xml_has_attr(x, "id"))

  expect_false(xml_has_attr(x, "id2"))
})

test_that("returning an attribute node prints properly", {
  x <- read_xml("<a><b c='1' /></a>")

  t1 <- xml_find_first(x, "//@c")

  expect_equal(format(t1), "<c>")
})

# Namespaces -------------------------------------------------------------------

# Default namespace doesn't apply to attributes

test_that("qualified names returned when ns given", {
  x <- read_xml("ns-multiple.xml")
  ns <- xml_ns(x)

  bars <- xml_children(xml_children(x))
  attr <- xml_attrs(bars, ns)

  expect_equal(names(attr[[1]]), "f:id")
  expect_equal(names(attr[[2]]), "g:id")
})


x <- read_xml('
 <root xmlns:b="http://bar.com" xmlns:f="http://foo.com">
   <doc b:id="b" f:id="f" id="" />
   <doc b:id="b" f:id="f" id="" />
 </root>
')
doc <- xml_children(x)[[1]]
docs <- xml_find_all(x, "//doc")
ns <- xml_ns(x)

test_that("qualified attributes get own values", {
  expect_equal(xml_attrs(doc, ns), c("b:id" = "b", "f:id" = "f", "id" = ""))
})

test_that("unqualified name gets unnamespace attribute", {
  expect_equal(xml_attr(doc, "id", ns), "")
})

test_that("namespace names gets namespaced attribute", {
  expect_equal(xml_attr(doc, "b:id", ns), "b")
  expect_equal(xml_attr(doc, "f:id", ns), "f")
})

test_that("xml_attr<- modifies properties", {
  xml_attr(doc, "id", ns) <- "test"
  expect_equal(xml_attr(doc, "id", ns), "test")

  xml_attr(doc, "b:id", ns) <- "b_test"
  expect_equal(xml_attr(doc, "b:id", ns), "b_test")

  xml_attr(doc, "f:id", ns) <- "f_test"
  expect_equal(xml_attr(doc, "f:id", ns), "f_test")

  xml_attr(docs, "f:id", ns) <- "f_test2"
  expect_equal(xml_attr(docs, "f:id", ns), c("f_test2", "f_test2"))

  xml_attr(docs, "f:id", ns) <- NULL
  expect_equal(xml_attr(docs, "f:id", ns), c(NA_character_, NA_character_))
})

test_that("xml_attrs<- modifies all attributes", {
  expect_error(xml_attrs(doc) <- 1, "`value` must be a named character vector or `NULL`")
  expect_error(xml_attrs(doc) <- "test", "`value` must be a named character vector or `NULL`")

  xml_attrs(doc, ns) <- c("b:id" = "b", "f:id" = "f", "id" = "test")
  expect_equal(xml_attrs(doc, ns), c("b:id" = "b", "id" = "test", "f:id" = "f"))

  xml_attrs(doc, ns) <- c("b:id" = "b", "f:id" = "f")
  expect_equal(xml_attrs(doc, ns), c("b:id" = "b", "f:id" = "f"))

  xml_attrs(doc, ns) <- c("b:id" = "b", "id" = "test")
  expect_equal(xml_attrs(doc, ns), c("b:id" = "b", "id" = "test"))

  expect_error(xml_attrs(docs) <- "test", "`value` must be a list of named character vectors")

  xml_attrs(docs, ns) <- c("b:id" = "b", "id" = "test")
  expect_equal(xml_attrs(docs, ns),
    list(
      c("b:id" = "b", "id" = "test"),
      c("b:id" = "b", "id" = "test")))

  xml_attrs(docs, ns) <- NULL
  expect_equivalent(xml_attrs(docs, ns), list(character(0), character(0)))
})

test_that("xml_attr<- accepts non-character values", {
  x <- read_xml('<svg><rect /></svg>')
  svg <- xml_root(x)

  xml_attr(svg, "width") <- 8L
  expect_that(xml_attr(svg, "width"), equals("8"))

  xml_attr(svg, "height") <- 12.5
  expect_that(xml_attr(svg, "height"), equals("12.5"))

  expect_that(xml_attrs(svg), equals(c(width = "8", height = "12.5")))

  xml_attrs(svg) <- c(width = 14L, height = 23.45)
  expect_that(xml_attrs(svg), equals(c(width = "14", height = "23.45")))
})

test_that("xml_attr<- can set empty strings, and removes attributes with NULL", {
  x <- read_xml("<x><y /></x>")
  xml_attr(x, "test") <- ""
  expect_equal(xml_attr(x, "test"), "")

  xml_attr(x, "test") <- NULL
  expect_equal(xml_attr(x, "test"), NA_character_)
})

test_that("xml_attr<- removes namespaces if desired", {
  xml_attr(x, "xmlns:b") <- NULL

  expect_equal(xml_attrs(x), c("xmlns:f" = "http://foo.com"))
})

test_that("xml_attr<- removes namespaces if desired", {
  x <- read_xml("<a xmlns = 'tag:foo'><b/></a>")

  # cannot find //b with a default namespace
  expect_equal(length(xml_find_all(x, "//b")), 0)

  # unless we specify it explicitly
  expect_equal(length(xml_find_all(x, "//b")), 0)
  expect_equal(length(xml_find_all(x, "//d1:b", xml_ns(x))), 1)

  # but can find it once we remove the namespace
  xml_attr(x, "xmlns") <- NULL
  expect_equal(length(xml_find_all(x, "//b")), 1)

  # and add the old namespace back
  xml_attr(x, "xmlns") <- "tag:foo"
  expect_equal(xml_attr(x, "xmlns"), "tag:foo")
  expect_equal(length(xml_find_all(x, "//b")), 0)
  expect_equal(length(xml_find_all(x, "//d1:b", xml_ns(x))), 1)

  expect_equal(xml_attr(x, "xmlns"), "tag:foo")
})

test_that("xml_attr<- removes prefixed namespaces if desired", {
  x <- read_xml("<a xmlns:pre = 'tag:foo'><pre:b/></a>")

  # cannot find //b with a prefixed namespace
  expect_equal(length(xml_find_all(x, "//b")), 0)

  # unless we specify it explicitly
  expect_equal(length(xml_find_all(x, "//b")), 0)
  expect_equal(length(xml_find_all(x, "//pre:b", xml_ns(x))), 1)

  # but can find it once we remove the namespace
  xml_attr(x, "xmlns:pre") <- NULL
  expect_equal(length(xml_find_all(x, "//b")), 1)

  # and add the old namespace back
  xml_attr(x, "xmlns:pre") <- "tag:foo"
  xml_set_namespace(xml_children(x)[[1]], "pre")
  expect_equal(xml_attr(x, "xmlns:pre"), "tag:foo")
  expect_equal(length(xml_find_all(x, "//b")), 0)
  expect_equal(length(xml_find_all(x, "//pre:b", xml_ns(x))), 1)

  expect_equal(xml_attr(x, "xmlns:pre"), "tag:foo")
})

test_that("xml_set_attr works identically to xml_attr<-", {
  content <- "<a><b><c/></b><b/></a>"
  x <- read_xml(content)
  y <- read_xml(content)

  xml_attr(x, "a") <- "test"
  xml_set_attr(y, "a", "test")

  expect_equal(as.character(x), as.character(y))

  bx <- xml_find_all(x, "//b")
  by <- xml_find_all(y, "//b")

  xml_attr(bx, "b") <- "test2"
  xml_set_attr(by, "b", "test2")

  expect_equal(as.character(x), as.character(y))

  # No errors for xml_missing
  mss <- xml_find_first(bx, "./c")
  expect_error(xml_attr(mss[[2]], "b") <- "blah", NA)
  expect_error(xml_set_attr(mss[[2]], "b", "blah"), NA)
})

test_that("xml_set_attrs works identically to xml_attrs<-", {
  content <- "<a><b><c/></b><b/></a>"
  x <- read_xml(content)
  y <- read_xml(content)

  xml_attrs(x) <- c(a = "test")
  xml_set_attrs(y, c(a = "test"))

  expect_equal(as.character(x), as.character(y))

  bx <- xml_find_all(x, "//b")
  by <- xml_find_all(y, "//b")

  xml_attrs(bx) <- c(b = "test2")
  xml_set_attrs(by, c(b = "test2"))

  expect_equal(as.character(x), as.character(y))

  # No errors for xml_missing
  mss <- xml_find_first(bx, "./c")
  expect_error(xml_attrs(mss[[2]]) <- c("b" = "blah"), NA)
  expect_error(xml_set_attrs(mss[[2]], c("b" = "blah")), NA)
})

#' Retrieve the xpath to a node
#'
#' This is useful when you want to figure out where nodes matching an
#' xpath expression live in a document.
#'
#' @inheritParams xml_name
#' @return A character vector.
#' @export
#' @examples
#' x <- read_xml("<foo><bar><baz /></bar><baz /></foo>")
#' xml_path(xml_find_all(x, ".//baz"))
xml_path <- function(x) {
  .Call(node_path, x)
}

libxml2_version <- function() {
  as.numeric_version(.Call(libxml2_version_))
}

xml_parse_options <- function() {
  .Call(xml_parse_options_)
}

xml_save_options <- function() {
  .Call(xml_save_options_)
}

download_file_callback <- function(url){
  tmp <- tempfile()
  on.exit(unlink(tmp))
  utils::download.file(url, tmp, quiet = TRUE)
  readBin(tmp, raw(), file.info(tmp)$size)
}

#' @useDynLib xml2, .registration = TRUE
NULL

#' Construct a cdata node
#' @param content The CDATA content, does not include `<![CDATA[`
#' @examples
#' x <- xml_new_root("root")
#' xml_add_child(x, xml_cdata("<d/>"))
#' as.character(x)
#' @export
xml_cdata <- function(content) {
  class(content) <- "xml_cdata"
  content
}

#' Construct a comment node
#' @param content The comment content
#' @examples
#' x <- xml_new_document()
#' r <- xml_add_child(x, "root")
#' xml_add_child(r, xml_comment("Hello!"))
#' as.character(x)
#' @export
xml_comment <- function(content) {
  class(content) <- "xml_comment"
  content
}

#' Construct a document type definition
#'
#' This is used to create simple document type definitions. If you need to
#' create a more complicated definition with internal subsets it is recommended
#' to parse a string directly with `read_xml()`.
#' @param name The name of the declaration
#' @param external_id The external ID of the declaration
#' @param system_id The system ID of the declaration
#' @examples
#' r <- xml_new_root(
#'   xml_dtd(
#'     "html",
#'     "-//W3C//DTD XHTML 1.0 Transitional//EN",
#'     "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"
#'   )
#' )
#'
#' # Use read_xml directly for more complicated DTD
#' d <- read_xml(
#'   '<!DOCTYPE doc [
#' <!ELEMENT doc (#PCDATA)>
#' <!ENTITY foo " test ">
#' ]>
#' <doc>This is a valid document &foo; !</doc>'
#' )
#' @export
xml_dtd <- function(name = "", external_id = "", system_id = "") {
  out <- list(name = name, external_id = external_id, system_id = system_id)
  class(out) <- "xml_dtd"
  out
}

test_that("CDATA creation works", {
  x <- xml_new_root("root")
  xml_add_child(x, xml_cdata("<d/>"))
  expect_identical(as.character(x), "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<root><![CDATA[<d/>]]></root>\n")
})

test_that("Comment creation works", {
  x <- xml_new_root("root")
  xml_add_child(x, xml_comment("Hello!"))
  expect_identical("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<root><!--Hello!--></root>\n", as.character(x, options = ""))
})

test_that("xml_dtd works", {
  r <- xml_new_root(xml_dtd(name = "html", external_id = "-//W3C//DTD XHTML 1.0 Transitional//EN", system_id = "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"))
  expect_identical("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n", as.character(r))

  no_name <- xml_new_root(xml_dtd(external_id = "-//W3C//DTD XHTML 1.0 Transitional//EN", system_id = "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"))
  expect_identical("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE  PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n", as.character(no_name))

  no_name_external_id <- xml_new_root(xml_dtd(system_id = "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"))
  expect_identical("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE  SYSTEM \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n", as.character(no_name_external_id))

  no_name_external_id_internal_id <- xml_new_root(xml_dtd())
  expect_identical("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE >\n", as.character(no_name_external_id_internal_id))
})

test_that("CDATA creation works", {
  x <- xml_new_root("root")
  xml_add_child(x, xml_cdata("<d/>"))
  expect_identical(
    as.character(x),
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<root><![CDATA[<d/>]]></root>\n"
  )
})

test_that("Comment creation works", {
  x <- xml_new_root("root")
  xml_add_child(x, xml_comment("Hello!"))
  expect_identical(
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<root><!--Hello!--></root>\n",
    as.character(x, options = "")
  )
})

test_that("xml_dtd works", {
  r <- xml_new_root(xml_dtd(
    name = "html",
    external_id = "-//W3C//DTD XHTML 1.0 Transitional//EN",
    system_id = "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"
  ))
  expect_identical(
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n",
    as.character(r)
  )

  no_name <- xml_new_root(xml_dtd(
    external_id = "-//W3C//DTD XHTML 1.0 Transitional//EN",
    system_id = "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"
  ))
  expect_identical(
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE  PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n",
    as.character(no_name)
  )

  no_name_external_id <- xml_new_root(xml_dtd(
    system_id = "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"
  ))
  expect_identical(
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE  SYSTEM \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n",
    as.character(no_name_external_id)
  )

  no_name_external_id_internal_id <- xml_new_root(xml_dtd())
  expect_identical(
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE >\n",
    as.character(no_name_external_id_internal_id)
  )
})

#!/bin/sh

# Create the test.xml file
echo '<?xml version="1.0" encoding="UTF-8"?>' > $AUTOPKGTEST_TMP/test.xml
echo '<root>' >> $AUTOPKGTEST_TMP/test.xml
echo '    <element attribute="value">Content</element>' >> $AUTOPKGTEST_TMP/test.xml
echo '    <element>Another Content</element>' >> $AUTOPKGTEST_TMP/test.xml
echo '</root>' >> $AUTOPKGTEST_TMP/test.xml

# Convert XML file to HTML using xml2
xml2 < $AUTOPKGTEST_TMP/test.xml > $AUTOPKGTEST_TMP/test.html

# Check if the test.html file was created correctly
if [ ! -s $AUTOPKGTEST_TMP/test.html ]; then
    echo "Test failed: $AUTOPKGTEST_TMP/test.html file is empty."
    exit 1
else
    echo "Test passed: the $AUTOPKGTEST_TMP/test.html file was created successfully."
    exit 0
fi

---
title: "Node Modification"
author: "Jim Hester"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Node Modification}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE, message = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
library(xml2)
library(magrittr)
```

# Modifying Existing XML

Modifying existing XML can be done in xml2 by using the replacement functions
of the accessors. They all have methods for both individual `xml_node` objects
as well as `xml_nodeset` objects. If a vector of values is provided it is
applied piecewise over the nodeset, otherwise the value is recycled.

## Text Modification ##

Text modification only happens on text nodes. If a given node has more than one
text node only the first will be affected. If you want to modify additional
text nodes you need to select them explicitly with `/text()`.

```{r}
x <- read_xml("<p>This is some <b>text</b>. This is more.</p>")
xml_text(x)

xml_text(x) <- "This is some other text."
xml_text(x)

# You can avoid this by explicitly selecting the text node.
x <- read_xml("<p>This is some text. This is <b>bold!</b></p>")
text_only <- xml_find_all(x, "//text()")

xml_text(text_only) <- c("This is some other text. ", "Still bold!")
xml_text(x)
xml_structure(x)
```

## Attribute and Namespace Definition Modification ##

Attributes and namespace definitions are modified one at a time with
`xml_attr()` or all at once with `xml_attrs()`. In both cases using `NULL` as
the value will remove the attribute completely.

```{r}
x <- read_xml("<a href='invalid!'>xml2</a>")
xml_attr(x, "href")

xml_attr(x, "href") <- "https://github.com/r-lib/xml2"
xml_attr(x, "href")

xml_attrs(x) <- c(id = "xml2", href = "https://github.com/r-lib/xml2")
xml_attrs(x)
x

xml_attrs(x) <- NULL
x

# Namespaces are added with as a xmlns or xmlns:prefix attribute
xml_attr(x, "xmlns") <- "http://foo"
x

xml_attr(x, "xmlns:bar") <- "http://bar"
x
```

## Name Modification ##

Node names are modified with `xml_name()`.

```{r}
x <- read_xml("<a><b/></a>")
x
xml_name(x)
xml_name(x) <- "c"
x
```

# Node modification #
All of these functions have a `.copy` argument. If this is set to `FALSE` they
will remove the new node from its location before inserting it into the new
location. Otherwise they make a copy of the node before insertion.

## Replacing existing nodes ##
```{r}
x <- read_xml("<parent><child>1</child><child>2<child>3</child></child></parent>")
children <- xml_children(x)
t1 <- children[[1]]
t2 <- children[[2]]
t3 <- xml_children(children[[2]])[[1]]

xml_replace(t1, t3)
x
```

## Add a sibling ##
```{r}
x <- read_xml("<parent><child>1</child><child>2<child>3</child></child></parent>")
children <- xml_children(x)
t1 <- children[[1]]
t2 <- children[[2]]
t3 <- xml_children(children[[2]])[[1]]

xml_add_sibling(t1, t3)
x

xml_add_sibling(t3, t1, where = "before")
x
```

## Add a child ##
```{r}
x <- read_xml("<parent><child>1</child><child>2<child>3</child></child></parent>")
children <- xml_children(x)
t1 <- children[[1]]
t2 <- children[[2]]
t3 <- xml_children(children[[2]])[[1]]

xml_add_child(t1, t3)
x

xml_add_child(t1, read_xml("<test/>"))
x
```

## Removing nodes ##
The `xml_remove()` can be used to remove a node (and its children) from a
tree. The default behavior is to unlink the node from the tree, but does _not_
free the memory for the node, so R objects pointing to the node are still
valid.

This allows code like the following to work without crashing R

```{r}
x <- read_xml("<foo><bar><baz/></bar></foo>")
x1 <- x %>%
  xml_children() %>%
  .[[1]]
x2 <- x1 %>%
  xml_children() %>%
  .[[1]]

xml_remove(x1)
rm(x1)
gc()

x2
```
If you are not planning on referencing these nodes again this memory is wasted.
Calling `xml_remove(free = TRUE)` will remove the nodes _and_ free the memory
used to store them.  **Note** In this case _any_ node which previously pointed
to the node or its children will instead be pointing to free memory and may
cause R to crash. xml2 can't figure this out for you, so it's your
responsibility to remove any objects which are no longer valid.

In particular `xml_find_*()` results are easy to overlook, for example

```{r}
x <- read_xml("<a><b /><b><b /></b></a>")
bees <- xml_find_all(x, "//b")
xml_remove(xml_child(x), free = TRUE)
# bees[[1]] is no longer valid!!!
rm(bees)
gc()
```

## Namespaces ##

We want to construct a document with the following namespace layout. (From
<https://stackoverflow.com/questions/32939229/creating-xml-in-r-with-namespaces/32941524#32941524>).
```xml
<?xml version = "1.0" encoding="UTF-8"?>
<sld xmlns="http://www.opengis.net/sld"
     xmlns:ogc="http://www.opengis.net/ogc"
     xmlns:se="http://www.opengis.net/se"
     version="1.1.0" >
<layer>
<se:Name>My Layer</se:Name>
</layer>
</sld>
```

```{r}
d <- xml_new_root("sld",
  "xmlns" = "http://www.opengis.net/sld",
  "xmlns:ogc" = "http://www.opengis.net/ogc",
  "xmlns:se" = "http://www.opengis.net//se",
  version = "1.1.0"
) %>%
  xml_add_child("layer") %>%
  xml_add_child("se:Name", "My Layer") %>%
  xml_root()

d
```

# print method is correct

    Code
      print(divs)
    Output
      {xml_nodeset (10)}
       [1] <div id="wrapper">\n            <div id="root" class="redesign">\n<scrip ...
       [2] <div id="root" class="redesign">\n<script>\n    if (typeof uet == 'funct ...
       [3] <div id="nb20" class="navbarSprite">\n<div id="supertab">\t\n\t<!-- begi ...
       [4] <div id="supertab">\t\n\t<!-- begin TOP_AD -->\n<div id="top_ad_wrapper" ...
       [5] <div id="top_ad_wrapper" class="dfp_slot">\n<script type="text/javascrip ...
       [6] <div id="top_ad_reflow_helper"></div>
       [7] <div id="navbar" class="navbarSprite">\n<noscript>\n  <link rel="stylesh ...
       [8] <div id="nb_search">\n    <noscript><div id="more_if_no_javascript"><a h ...
       [9] <div id="more_if_no_javascript"><a href="/search/">More</a></div>
      [10] <div class="magnifyingglass navbarSprite"></div>

---

    Code
      print(x, width = 13L)
    Output
      {xml_document}
      <doc>
      [1] <a>123 ...
      [2] <b>123 ...
      [3] <c>12\ ...
    Code
      print(x, width = 14L)
    Output
      {xml_document}
      <doc>
      [1] <a>1234 ...
      [2] <b>1234 ...
      [3] <c>12\\ ...

# xml_find_num errors with non numeric results

    Code
      xml_find_num(x, "//z")
    Condition
      Error in `xml_find_num()`:
      ! Element at path `//z` must be a number, not a <xml_missing> object.
    Code
      xml_find_num(x, "//y")
    Condition
      Error in `xml_find_num()`:
      ! Element at path `//y` must be a number, not a list.
    Code
      xml_find_num(x, "1=1")
    Condition
      Error in `xml_find_num()`:
      ! Element at path `1=1` must be a number, not `TRUE`.
    Code
      xml_find_num(x, "string(5)")
    Condition
      Error in `xml_find_num()`:
      ! Element at path `string(5)` must be a number, not the string "5".

# xml_find_int errors with non integer results

    Code
      xml_find_int(x, "//z")
    Condition
      Error in `xml_find_int()`:
      ! Element at path `//z` must be a whole number, not a <xml_missing> object.
    Code
      xml_find_int(x, "//y")
    Condition
      Error in `xml_find_int()`:
      ! Element at path `//y` must be a whole number, not a list.
    Code
      xml_find_int(x, "number(1.1)")
    Condition
      Error in `xml_find_int()`:
      ! Element at path `number(1.1)` must be a whole number, not the number 1.1.

# xml_find_chr errors with non character results

    Code
      xml_find_chr(x, "//z")
    Condition
      Error in `xml_find_chr()`:
      ! Element at path `//z` must be a single string, not a <xml_missing> object.
    Code
      xml_find_chr(x, "//y")
    Condition
      Error in `xml_find_chr()`:
      ! Element at path `//y` must be a single string, not a list.
    Code
      xml_find_chr(x, "1=1")
    Condition
      Error in `xml_find_chr()`:
      ! Element at path `1=1` must be a single string, not `TRUE`.
    Code
      xml_find_chr(x, "1+1")
    Condition
      Error in `xml_find_chr()`:
      ! Element at path `1+1` must be a single string, not the number 2.

# xml_find_lgl errors with non logical results

    Code
      xml_find_lgl(x, "//z")
    Condition
      Error in `xml_find_lgl()`:
      ! Element at path `//z` must be `TRUE` or `FALSE`, not a <xml_missing> object.
    Code
      xml_find_lgl(x, "//y")
    Condition
      Error in `xml_find_lgl()`:
      ! Element at path `//y` must be `TRUE` or `FALSE`, not a list.
    Code
      xml_find_lgl(x, "string(5)")
    Condition
      Error in `xml_find_lgl()`:
      ! Element at path `string(5)` must be `TRUE` or `FALSE`, not the string "5".
    Code
      xml_find_lgl(x, "1+1")
    Condition
      Error in `xml_find_lgl()`:
      ! Element at path `1+1` must be `TRUE` or `FALSE`, not the number 2.

---
title: xml2 1.1.1
output:
  html_fragment
---

```{r setup, include=FALSE}
library(xml2)
library(xslt)
library(magrittr)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```
<meta charset="utf-8">

Today we are pleased to release version 1.1.1 of xml2.  xml2 makes it easy to read, create, and modify XML with R. You can install it with:

```{r, eval = FALSE}
install.packages("xml2")
```

As well as fixing many bugs, this release:

* Makes it easier to create an modify XML
* Improves roundtrip support between XML and lists
* Adds support for XML validation and XSLT transformations. 

You can see a  full list of changes in the [release notes](https://github.com/r-lib/xml2/releases/tag/v1.1.1). This is the first release maintained by [Jim Hester](https://github.com/jimhester).

## Creating and modifying XML

xml2 has been overhauled with a set of methods to make generating and modfying XML easier:

*   `xml_new_root()` can be used to create a new document and root node 
    simultaneously.

    ```{r}
    xml_new_root("x") %>%
      xml_add_child("y") %>%
      xml_root()
    ```

*   New `xml_set_text()`, `xml_set_name()`, `xml_set_attr()`, and 
    `xml_set_attrs()` make it easy to modify nodes within a pipeline.
  
    ```{r}
    x <- read_xml("<a>
        <b />
        <c><b/></c>
      </a>")
    x
    
    x %>% 
      xml_find_all(".//b") %>% 
      xml_set_name("banana") %>% 
      xml_set_attr("oldname", "b")
    x
    ```

* New `xml_add_parent()` makes it easy to insert a node as the parent of 
  an existing node.

* You can create more esoteric node types with `xml_comment()` (comments),
  `xml_cdata()` (CDATA nodes), and `xml_dtd()` (DTDs).

## Coercion to and from R Lists

xml2 1.1.1 improves support for converting to and from R lists, thanks in part
to work by [Peter Foley](https://github.com/peterfoley) and [Jenny
Bryan](https://github.com/jennybc). In particular xml2 now supports preserving the root
node name as well as saving all xml2 attributes as R attributes. These changes
allows you to convert most XML documents to and from R lists with `as_list()`
and `as_xml_document()` without loss of data.

```{r}
x <- read_xml("<fruits><apple color = 'red' /></fruits>")
x
as_list(x)
as_xml_document(as_list(x))
```

## XML validation and xslt

xml2 1.1.1 also adds support for XML validation, thanks to [Jeroen
Ooms](https://github.com/jeroenooms).
Simply read the document and schema files and call `xml_validate()`.

```{r}
doc <- read_xml(system.file("extdata/order-doc.xml", package = "xml2"))
schema <- read_xml(system.file("extdata/order-schema.xml", package = "xml2"))
xml_validate(doc, schema)
```

Jeroen also released the first xml2 extension package in conjunction with xml2
1.1.1, [xslt](https://cran.r-project.org/package=xslt). xslt allows one to
apply [XSLT (Extensible Stylesheet Language
Transformations)](https://en.wikipedia.org/wiki/XSLT) to XML documents, which
are great for transforming XML data into other formats such as HTML.

# xml_structure is correct

    Code
      html_structure(quicklinks)
    Output
      <div.aux-content-widget-3.links.subnav [div]>
        {text}
        <h3>
          {text}
        {text}
        <div#maindetails_quicklinks>
          {text}
          <div.split_0>
            {text}
            <ul.quicklinks>
              <li.subnav_item_main>
                {text}
                <a.link [href]>
                  {text}
                {text}
              {text}
              <li.subnav_item_main>
                {text}
                <a.link [href]>
                  {text}
                {text}
              {text}
              <li.subnav_item_main>
                {text}
                <a.link [href]>
                  {text}
                {text}
              {text}
              <li.subnav_item_main>
                {text}
                <a.link [href]>
                  {text}
                {text}
              {text}
              <li.subnav_item_main>
                {text}
                <a.link [href]>
                  {text}
                {text}
              {text}
          {text}
          <div.split_1>
            {text}
            <ul.quicklinks>
              <li.subnav_item_main>
                {text}
                <a.link [href]>
                  {text}
                {text}
              {text}
              <li.subnav_item_main>
                {text}
                <a.link [href]>
                  {text}
                {text}
              {text}
              <li.subnav_item_main>
                {text}
                <a.link [href]>
                  {text}
                {text}
              {text}
              <li.subnav_item_main>
                {text}
                <a.link [href]>
                  {text}
                {text}
              {text}
              <li.subnav_item_main>
                {text}
                <a.link [href]>
                  {text}
                {text}
              {text}
          {text}
        {text}
        <hr>
        <div#full_subnav>
          {text}
          <h4>
            {text}
          {text}
          <ul.quicklinks>
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link.ghost [href]>
                {text}
              {text}
            {text}
          <h4>
            {text}
          {text}
          <ul.quicklinks>
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
          <h4>
            {text}
          {text}
          <ul.quicklinks>
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link.ghost [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
          <h4>
            {text}
          {text}
          <ul.quicklinks>
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
          <h4>
            {text}
          {text}
          <ul.quicklinks>
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link.ghost [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
          <h4>
            {text}
          {text}
          <ul.quicklinks>
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
          <h4>
            {text}
          {text}
          <ul.quicklinks>
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link.ghost [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
          <h4>
            {text}
          {text}
          <ul.quicklinks>
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
            <li.subnav_item_main>
              {text}
              <a.link [href]>
                {text}
              {text}
            {text}
          <hr>
        {text}
        <div.show_more>
          <span.titlePageSprite.arrows.show>
          {text}
        {text}
        <div.show_less>
          <span.titlePageSprite.arrows.hide>
          {text}
        {text}

# read_xml errors with an empty document

    Code
      read_xml(character())
    Condition
      Error in `read_xml()`:
      ! `x` must be a single string, not an empty character vector.

# parse_options errors when given an invalid option

    Code
      read_html(test_path("lego.html.bz2"), options = "INVALID")
    Condition
      Error in `read_html()`:
      x `options` "INVALID" is not a valid option.
      i Valid options are one of "RECOVER", "NOENT", "DTDLOAD", "DTDATTR", "DTDVALID", "NOERROR", "NOWARNING", "PEDANTIC", "NOBLANKS", "SAX1", "XINCLUDE", "NONET", "NODICT", "NSCLEAN", "NOCDATA", "NOXINCNODE", "COMPACT", "OLD10", ..., "IGNORE_ENC", or "BIG_LINES".
      i See read_html (`?xml2::read_html()`) for all options.

# read_xml and read_html fail with > 1 input

    Code
      read_xml(c("foo", "bar"))
    Condition
      Error in `read_xml()`:
      ! `x` must be a single string, not a character vector.
    Code
      read_html(c("foo", "bar"))
    Condition
      Error in `read_xml()`:
      ! `x` must be a single string, not a character vector.

# print method is correct

    Code
      print(body)
    Output
      {html_node}
      <body id="styleguide-v2" class="fixed">
       [1] <script>\n    if (typeof uet == 'function') {\n      uet("bb");\n    }\n ...
       [2] <script>\n    if ('csm' in window) {\n      csm.measure('csm_body_delive ...
       [3] <div id="wrapper">\n            <div id="root" class="redesign">\n<scrip ...
       [4] <script type="text/javascript" src="http://ia.media-imdb.com/images/G/01 ...
       [5] <script type="text/imdblogin-js" id="login">\njQuery(document).ready(fun ...
       [6] <script type="text/javascript">\n                jQuery(\n               ...
       [7] <iframe id="sis_pixel_sitewide" width="1" height="1" frameborder="0" mar ...
       [8] <script>\n    setTimeout(function(){\n        try{\n            //sis3.0 ...
       [9] <script type="text/javascript" src="http://ia.media-imdb.com/images/G/01 ...
      [10] <script type="text/javascript">\nif(window.COMSCORE){\nCOMSCORE.beacon({ ...
      [11] <noscript>\n<img src="http://b.scorecardresearch.com/p?c1=2&amp;c2=60349 ...
      [12] <script>\n    doWithAds(function(){\n        (new Image()).src = "http:/ ...
      [13] <script>\n(function(){\n    var readyTimeout = setInterval(function(){\n ...
      [14] <div id="servertime" time="235"></div>
      [15] <script>\n    if (typeof uet == 'function') {\n      uet("be");\n    }\n ...

# print method is correct

    Code
      print(x)
    Output
      {html_document}
      <html xmlns:og="http://ogp.me/ns#" xmlns:fb="http://www.facebook.com/2008/fbml">
      [1] <head>\n<script type="text/javascript">var ue_t0=window.ue_t0||+new Date( ...
      [2] <body id="styleguide-v2" class="fixed">\n<script>\n    if (typeof uet ==  ...

# write_xml errors for incorrect directory and with invalid inputs

    Code
      write_xml(x, c("test.xml", "foo"))
    Condition
      Error in `write_xml()`:
      ! `file` must be a single string, not a character vector.

# write_xml works with nodeset input and connections

    Code
      write_xml(y[1], c(filename, "foo"))
    Condition
      Error in `write_xml()`:
      ! `file` must be a single string, not a character vector.

# write_xml works with node input and files

    Code
      write_xml(y, c(filename, "foo"))
    Condition
      Error in `write_xml()`:
      ! `file` must be a single string, not a character vector.

# write_xml errors for incorrect directory and with invalid inputs

    `file` must be a single string, not a character vector.

# write_xml works with nodeset input and connections

    `file` must be a single string, not a character vector.

# write_xml works with node input and files

    `file` must be a single string, not a character vector.

# xml_find_num errors with non numeric results

    Element at path `//z` must be a number, not a <xml_missing> object.

# xml_find_int errors with non integer results

    Code
      xml_find_int(x, "//z")
    Condition
      Error in `xml_find_int()`:
      ! Element at path `//z` must be a whole number, not a <xml_missing> object.
    Code
      xml_find_int(x, "//y")
    Condition
      Error in `xml_find_int()`:
      ! Element at path `//y` must be a whole number, not a list.
    Code
      xml_find_int(x, "number(1.1)")
    Condition
      Error in `xml_find_int()`:
      ! Element at path `number(1.1)` must be a whole number, not the number 1.1.

# xml_find_chr errors with non character results

    Element at path `//z` must be a single string, not a <xml_missing> object.

# xml_find_lgl errors with non logical results

    Element at path `//z` must be `TRUE` or `FALSE`, not a <xml_missing> object.

# read_xml errors with an empty document

    `x` must be a single string, not an empty character vector.

# read_xml and read_html fail with > 1 input

    `x` must be a single string, not a character vector.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_children.R
\name{xml_children}
\alias{xml_children}
\alias{xml_child}
\alias{xml_contents}
\alias{xml_parents}
\alias{xml_siblings}
\alias{xml_parent}
\alias{xml_length}
\alias{xml_root}
\title{Navigate around the family tree.}
\usage{
xml_children(x)

xml_child(x, search = 1, ns = xml_ns(x))

xml_contents(x)

xml_parents(x)

xml_siblings(x)

xml_parent(x)

xml_length(x, only_elements = TRUE)

xml_root(x)
}
\arguments{
\item{x}{A document, node, or node set.}

\item{search}{For \code{xml_child}, either the child number to return (by
position), or the name of the child node to return. If there are multiple
child nodes with the same name, the first will be returned}

\item{ns}{Optionally, a named vector giving prefix-url pairs, as produced
by \code{\link{xml_ns}}. If provided, all names will be explicitly
qualified with the ns prefix, i.e. if the element \code{bar} is defined
in namespace \code{foo}, it will be called \code{foo:bar}. (And
similarly for atttributes). Default namespaces must be given an explicit
name. The ns is ignored when using \code{\link{xml_name<-}} and
\code{\link{xml_set_name}}.}

\item{only_elements}{For \code{xml_length}, should it count all children,
or just children that are elements (the default)?}
}
\value{
A node or nodeset (possibly empty). Results are always de-duplicated.
}
\description{
\code{xml_children} returns only elements, \code{xml_contents} returns
all nodes. \code{xml_length} returns the number of children.
\code{xml_parent} returns the parent node, \code{xml_parents}
returns all parents up to the root. \code{xml_siblings} returns all nodes
at the same level. \code{xml_child} makes it easy to specify a specific
child to return.
}
\examples{
x <- read_xml("<foo> <bar><boo /></bar> <baz/> </foo>")
xml_children(x)
xml_children(xml_children(x))
xml_siblings(xml_children(x)[[1]])

# Note the each unique node only appears once in the output
xml_parent(xml_children(x))

# Mixed content
x <- read_xml("<foo> a <b/> c <d>e</d> f</foo>")
# Childen gets the elements, contents gets all node types
xml_children(x)
xml_contents(x)

xml_length(x)
xml_length(x, only_elements = FALSE)

# xml_child makes it easier to select specific children
xml_child(x)
xml_child(x, 2)
xml_child(x, "baz")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_find.R
\name{xml_find_all}
\alias{xml_find_all}
\alias{xml_find_first}
\alias{xml_find_num}
\alias{xml_find_chr}
\alias{xml_find_lgl}
\alias{xml_find_one}
\title{Find nodes that match an xpath expression.}
\usage{
xml_find_all(x, xpath, ns = xml_ns(x))

xml_find_first(x, xpath, ns = xml_ns(x))

xml_find_num(x, xpath, ns = xml_ns(x))

xml_find_chr(x, xpath, ns = xml_ns(x))

xml_find_lgl(x, xpath, ns = xml_ns(x))
}
\arguments{
\item{x}{A document, node, or node set.}

\item{xpath}{A string containing a xpath (1.0) expression.}

\item{ns}{Optionally, a named vector giving prefix-url pairs, as produced
by \code{\link{xml_ns}}. If provided, all names will be explicitly
qualified with the ns prefix, i.e. if the element \code{bar} is defined
in namespace \code{foo}, it will be called \code{foo:bar}. (And
similarly for atttributes). Default namespaces must be given an explicit
name. The ns is ignored when using \code{\link{xml_name<-}} and
\code{\link{xml_set_name}}.}
}
\value{
\code{xml_find_all} always returns a nodeset: if there are no matches
  the nodeset will be empty. The result will always be unique; repeated
  nodes are automatically de-duplicated.

  \code{xml_find_first} returns a node if applied to a node, and a nodeset
  if applied to a nodeset. The output is \emph{always} the same size as
  the input. If there are no matches, \code{xml_find_first} will return a
  missing node; if there are multiple matches, it will return the first
  only.

  \code{xml_find_num}, \code{xml_find_chr}, \code{xml_find_lgl} return
  numeric, character and logical results respectively.
}
\description{
Xpath is like regular expressions for trees - it's worth learning if
you're trying to extract nodes from arbitrary locations in a document.
Use \code{xml_find_all} to find all matches - if there's no match you'll
get an empty result. Use \code{xml_find_first} to find a specific match -
if there's no match you'll get an \code{xml_missing} node.
}
\section{Deprecated functions}{

\code{xml_find_one()} has been deprecated. Instead use
\code{xml_find_first()}.
}

\examples{
x <- read_xml("<foo><bar><baz/></bar><baz/></foo>")
xml_find_all(x, ".//baz")
xml_path(xml_find_all(x, ".//baz"))

# Note the difference between .// and //
# //  finds anywhere in the document (ignoring the current node)
# .// finds anywhere beneath the current node
(bar <- xml_find_all(x, ".//bar"))
xml_find_all(bar, ".//baz")
xml_find_all(bar, "//baz")

# Find all vs find one -----------------------------------------------------
x <- read_xml("<body>
  <p>Some <b>text</b>.</p>
  <p>Some <b>other</b> <b>text</b>.</p>
  <p>No bold here!</p>
</body>")
para <- xml_find_all(x, ".//p")

# If you apply xml_find_all to a nodeset, it finds all matches,
# de-duplicates them, and returns as a single list. This means you
# never know how many results you'll get
xml_find_all(para, ".//b")

# xml_find_first only returns the first match per input node. If there are 0
# matches it will return a missing node
xml_find_first(para, ".//b")
xml_text(xml_find_first(para, ".//b"))

# Namespaces ---------------------------------------------------------------
# If the document uses namespaces, you'll need use xml_ns to form
# a unique mapping between full namespace url and a short prefix
x <- read_xml('
 <root xmlns:f = "http://foo.com" xmlns:g = "http://bar.com">
   <f:doc><g:baz /></f:doc>
   <f:doc><g:baz /></f:doc>
 </root>
')
xml_find_all(x, ".//f:doc")
xml_find_all(x, ".//f:doc", xml_ns(x))
}
\seealso{
\code{\link{xml_ns_strip}} to remove the default namespaces
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_xml_document.R
\name{as_xml_document}
\alias{as_xml_document}
\title{Coerce a R list to xml nodes.}
\usage{
as_xml_document(x, ...)
}
\arguments{
\item{x}{A document, node, or node set.}

\item{...}{Needed for compatibility with generic. Unused.}
}
\description{
This turns an R list into the equivalent XML document. Not all R lists will
produce valid XML, in particular there can only be one root node and all
child nodes need to be named (or empty) lists. R attributes become XML
attributes and R names become XML node names.
}
\examples{
as_xml_document(list(x = list()))

# Nesting multiple nodes
as_xml_document(list(foo = list(bar = list(baz = list()))))

# attributes are stored as R attributes
as_xml_document(list(foo = structure(list(), id = "a")))
as_xml_document(list(foo = list(
     bar = structure(list(), id = "a"),
     bar = structure(list(), id = "b"))))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_serialize.R
\name{xml_serialize}
\alias{xml_serialize}
\alias{xml_unserialize}
\title{Serializing XML objects to connections.}
\usage{
xml_serialize(object, connection, ...)

xml_unserialize(connection, ...)
}
\arguments{
\item{object}{\R object to serialize.}

\item{connection}{an open \link{connection} or (for \code{serialize})
    \code{NULL} or (for \code{unserialize}) a raw vector
    (see \sQuote{Details}).}

\item{...}{Additional arguments passed to \code{\link{read_xml}}.}
}
\value{
For \code{serialize}, \code{NULL} unless \code{connection = NULL}, when
  the result is returned in a raw vector.

  For \code{unserialize} an \R object.
}
\description{
Serializing XML objects to connections.
}
\examples{
library(xml2)
x <- read_xml("<a>
  <b><c>123</c></b>
  <b><c>456</c></b>
</a>")

b <- xml_find_all(x, "//b")
out <- xml_serialize(b, NULL)
xml_unserialize(out)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_parse.R
\name{read_xml}
\alias{read_xml}
\alias{read_html}
\alias{read_xml.character}
\alias{read_xml.raw}
\alias{read_xml.connection}
\title{Read HTML or XML.}
\usage{
read_xml(x, encoding = "", ..., as_html = FALSE, options = "NOBLANKS")

read_html(
  x,
  encoding = "",
  ...,
  options = c("RECOVER", "NOERROR", "NOBLANKS", "HUGE")
)

\method{read_xml}{character}(x, encoding = "", ..., as_html = FALSE, options = "NOBLANKS")

\method{read_xml}{raw}(
  x,
  encoding = "",
  base_url = "",
  ...,
  as_html = FALSE,
  options = "NOBLANKS"
)

\method{read_xml}{connection}(
  x,
  encoding = "",
  n = 64 * 1024,
  verbose = FALSE,
  ...,
  base_url = "",
  as_html = FALSE,
  options = "NOBLANKS"
)
}
\arguments{
\item{x}{A string, a connection, or a raw vector.

A string can be either a path, a url or literal xml. Urls will
be converted into connections either using \code{base::url} or, if
installed, \code{curl::curl}. Local paths ending in \code{.gz},
\code{.bz2}, \code{.xz}, \code{.zip} will be automatically uncompressed.

If a connection, the complete connection is read into a raw vector before
being parsed.}

\item{encoding}{Specify a default encoding for the document. Unless
otherwise specified XML documents are assumed to be in UTF-8 or
UTF-16. If the document is not UTF-8/16, and lacks an explicit
encoding directive, this allows you to supply a default.}

\item{...}{Additional arguments passed on to methods.}

\item{as_html}{Optionally parse an xml file as if it's html.}

\item{options}{Set parsing options for the libxml2 parser. Zero or more of
\Sexpr[results=rd, stage=build]{xml2:::describe_options(xml2:::xml_parse_options())}}

\item{base_url}{When loading from a connection, raw vector or literal
html/xml, this allows you to specify a base url for the document. Base
urls are used to turn relative urls into absolute urls.}

\item{n}{If \code{file} is a connection, the number of bytes to read per
iteration. Defaults to 64kb.}

\item{verbose}{When reading from a slow connection, this prints some
output on every iteration so you know its working.}
}
\value{
An XML document. HTML is normalised to valid XML - this may not
be exactly the same transformation performed by the browser, but it's
a reasonable approximation.
}
\description{
Read HTML or XML.
}
\section{Setting the "user agent" header}{


When performing web scraping tasks it is both good practice --- and often required ---
to set the \href{https://en.wikipedia.org/wiki/User_agent}{user agent} request header
to a specific value. Sometimes this value is assigned to emulate a browser in order
to have content render in a certain way (e.g. \verb{Mozilla/5.0 (Windows NT 5.1; rv:52.0) Gecko/20100101 Firefox/52.0} to emulate more recent Windows browsers). Most often,
this value should be set to provide the web resource owner information on who you are
and the intent of your actions like this Google scraping bot user agent identifier:
\verb{Googlebot/2.1 (+http://www.google.com/bot.html)}.

You can set the HTTP user agent for URL-based requests using \code{\link[httr:set_config]{httr::set_config()}} and \code{\link[httr:user_agent]{httr::user_agent()}}:

\code{httr::set_config(httr::user_agent("me@example.com; +https://example.com/info.html"))}

\code{\link[httr:set_config]{httr::set_config()}} changes the configuration globally,
\code{\link[httr:with_config]{httr::with_config()}} can be used to change configuration temporarily.
}

\examples{
# Literal xml/html is useful for small examples
read_xml("<foo><bar /></foo>")
read_html("<html><title>Hi<title></html>")
read_html("<html><title>Hi")

# From a local path
read_html(system.file("extdata", "r-project.html", package = "xml2"))

\dontrun{
# From a url
cd <- read_xml(xml2_example("cd_catalog.xml"))
me <- read_html("http://had.co.nz")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{xml_missing}
\alias{xml_missing}
\title{Construct an missing xml object}
\usage{
xml_missing()
}
\description{
Construct an missing xml object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_name.R
\name{xml_name}
\alias{xml_name}
\alias{xml_name<-}
\alias{xml_set_name}
\title{The (tag) name of an xml element.}
\usage{
xml_name(x, ns = character())

xml_name(x, ns = character()) <- value

xml_set_name(x, value, ns = character())
}
\arguments{
\item{x}{A document, node, or node set.}

\item{ns}{Optionally, a named vector giving prefix-url pairs, as produced
by \code{\link{xml_ns}}. If provided, all names will be explicitly
qualified with the ns prefix, i.e. if the element \code{bar} is defined
in namespace \code{foo}, it will be called \code{foo:bar}. (And
similarly for atttributes). Default namespaces must be given an explicit
name. The ns is ignored when using \code{\link{xml_name<-}} and
\code{\link{xml_set_name}}.}

\item{value}{a character vector with replacement name.}
}
\value{
A character vector.
}
\description{
The (tag) name of an xml element.

Modify the (tag) name of an element
}
\examples{
x <- read_xml("<bar>123</bar>")
xml_name(x)

y <- read_xml("<bar><baz>1</baz>abc<foo /></bar>")
z <- xml_children(y)
xml_name(xml_children(y))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_list.R
\name{as_list}
\alias{as_list}
\title{Coerce xml nodes to a list.}
\usage{
as_list(x, ns = character(), ...)
}
\arguments{
\item{x}{A document, node, or node set.}

\item{ns}{Optionally, a named vector giving prefix-url pairs, as produced
by \code{\link[=xml_ns]{xml_ns()}}. If provided, all names will be explicitly
qualified with the ns prefix, i.e. if the element \code{bar} is defined
in namespace \code{foo}, it will be called \code{foo:bar}. (And
similarly for attributes). Default namespaces must be given an explicit
name. The ns is ignored when using \code{\link[=xml_name<-]{xml_name<-()}} and
\code{\link[=xml_set_name]{xml_set_name()}}.}

\item{...}{Needed for compatibility with generic. Unused.}
}
\description{
This turns an XML document (or node or nodeset) into the equivalent R
list. Note that this is \code{as_list()}, not \code{as.list()}:
\code{lapply()} automatically calls \code{as.list()} on its inputs, so
we can't override the default.
}
\details{
\code{as_list} currently only handles the four most common types of
children that an element might have:

\itemize{
\item Other elements, converted to lists.
\item Attributes, stored as R attributes. Attributes that have special meanings in R
(\code{\link[=class]{class()}}, \code{\link[=comment]{comment()}}, \code{\link[=dim]{dim()}},
\code{\link[=dimnames]{dimnames()}}, \code{\link[=names]{names()}}, \code{\link[=row.names]{row.names()}} and
\code{\link[=tsp]{tsp()}}) are escaped with '.'
\item Text, stored as a character vector.
}
}
\examples{
as_list(read_xml("<foo> a <b /><c><![CDATA[<d></d>]]></c></foo>"))
as_list(read_xml("<foo> <bar><baz /></bar> </foo>"))
as_list(read_xml("<foo id = 'a'></foo>"))
as_list(read_xml("<foo><bar id='a'/><bar id='b'/></foo>"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_path.R
\name{xml_path}
\alias{xml_path}
\title{Retrieve the xpath to a node}
\usage{
xml_path(x)
}
\arguments{
\item{x}{A document, node, or node set.}
}
\value{
A character vector.
}
\description{
This is useful when you want to figure out where nodes matching an
xpath expression live in a document.
}
\examples{
x <- read_xml("<foo><bar><baz /></bar><baz /></foo>")
xml_path(xml_find_all(x, ".//baz"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_parse.R
\name{download_xml}
\alias{download_xml}
\alias{download_html}
\title{Download a HTML or XML file}
\usage{
download_xml(url, file = basename(url), quiet = TRUE, mode = "wb",
  handle = curl::new_handle())

download_html(url, file = basename(url), quiet = TRUE, mode = "wb",
  handle = curl::new_handle())
}
\arguments{
\item{url}{A character string naming the URL of a resource to be downloaded.}

\item{file}{A character string with the name where the downloaded file is
saved.}

\item{quiet}{If \code{TRUE}, suppress status messages (if any), and the
progress bar.}

\item{mode}{A character string specifying the mode with which to write the file.
Useful values are \code{"w"}, \code{"wb"} (binary), \code{"a"} (append)
and \code{"ab"}.}

\item{handle}{a curl handle object}
}
\value{
Path of downloaded file (invisibly).
}
\description{
Download a HTML or XML file
}
\details{
The main difference between \code{curl_download} and \code{curl_fetch_disk}
is that \code{curl_download} checks the http status code before starting the
download, and raises an error when status is non-successful. The behavior of
\code{curl_fetch_disk} on the other hand is to proceed as normal and write
the error page to disk in case of a non success response.
}
\examples{
\dontrun{
download_html("http://tidyverse.org/index.html")
}
}
\seealso{
\link[curl]{curl_download}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{xml_cdata}
\alias{xml_cdata}
\title{Construct a cdata node}
\usage{
xml_cdata(content)
}
\arguments{
\item{content}{The CDATA content, does not include \code{<![CDATA[}}
}
\description{
Construct a cdata node
}
\examples{
x <- xml_new_root("root")
xml_add_child(x, xml_cdata("<d/>"))
as.character(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_attr.R
\name{xml_attr}
\alias{xml_attr}
\alias{xml_has_attr}
\alias{xml_attrs}
\alias{xml_attr<-}
\alias{xml_set_attr}
\alias{xml_attrs<-}
\alias{xml_set_attrs}
\title{Retrieve an attribute.}
\usage{
xml_attr(x, attr, ns = character(), default = NA_character_)

xml_has_attr(x, attr, ns = character())

xml_attrs(x, ns = character())

xml_attr(x, attr, ns = character()) <- value

xml_set_attr(x, attr, value, ns = character())

xml_attrs(x, ns = character()) <- value

xml_set_attrs(x, value, ns = character())
}
\arguments{
\item{x}{A document, node, or node set.}

\item{attr}{Name of attribute to extract.}

\item{ns}{Optionally, a named vector giving prefix-url pairs, as produced
by \code{\link{xml_ns}}. If provided, all names will be explicitly
qualified with the ns prefix, i.e. if the element \code{bar} is defined
in namespace \code{foo}, it will be called \code{foo:bar}. (And
similarly for atttributes). Default namespaces must be given an explicit
name. The ns is ignored when using \code{\link{xml_name<-}} and
\code{\link{xml_set_name}}.}

\item{default}{Default value to use when attribute is not present.}

\item{value}{character vector of new value.}
}
\value{
\code{xml_attr()} returns a character vector. \code{NA} is used
 to represent of attributes that aren't defined.

 \code{xml_has_attr()} returns a logical vector.

 \code{xml_attrs()} returns a named character vector if \code{x} x is single
 node, or a list of character vectors if given a nodeset
}
\description{
\code{xml_attrs()} retrieves all attributes values as a named character
vector, \code{xml_attrs() <-} or \code{xml_set_attrs()} sets all attribute
values. \code{xml_attr()} retrieves the value of single attribute and
\code{xml_attr() <-} or \code{xml_set_attr()} modifies its value. If the
attribute doesn't exist, it will return \code{default}, which defaults to
\code{NA}. \code{xml_has_attr()} tests if an attribute is present.
}
\examples{
x <- read_xml("<root id='1'><child id ='a' /><child id='b' d='b'/></root>")
xml_attr(x, "id")
xml_attr(x, "apple")
xml_attrs(x)

kids <- xml_children(x)
kids
xml_attr(kids, "id")
xml_has_attr(kids, "id")
xml_attrs(kids)

# Missing attributes give missing values
xml_attr(xml_children(x), "d")
xml_has_attr(xml_children(x), "d")

# If the document has a namespace, use the ns argument and
# qualified attribute names
x <- read_xml('
 <root xmlns:b="http://bar.com" xmlns:f="http://foo.com">
   <doc b:id="b" f:id="f" id="" />
 </root>
')
doc <- xml_children(x)[[1]]
ns <- xml_ns(x)

xml_attrs(doc)
xml_attrs(doc, ns)

# If you don't supply a ns spec, you get the first matching attribute
xml_attr(doc, "id")
xml_attr(doc, "b:id", ns)
xml_attr(doc, "id", ns)

# Can set a single attribute with `xml_attr() <-` or `xml_set_attr()`
xml_attr(doc, "id") <- "one"
xml_set_attr(doc, "id", "two")

# Or set multiple attributes with `xml_attrs()` or `xml_set_attrs()`
xml_attrs(doc) <- c("b:id" = "one", "f:id" = "two", "id" = "three")
xml_set_attrs(doc, c("b:id" = "one", "f:id" = "two", "id" = "three"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{url_parse}
\alias{url_parse}
\title{Parse a url into its component pieces.}
\usage{
url_parse(x)
}
\arguments{
\item{x}{A character vector of urls.}
}
\value{
A dataframe with one row for each element of \code{x} and
  columns: scheme, server, port, user, path, query, fragment.
}
\description{
Parse a url into its component pieces.
}
\examples{
url_parse("http://had.co.nz/")
url_parse("http://had.co.nz:1234/")
url_parse("http://had.co.nz:1234/?a=1&b=2")
url_parse("http://had.co.nz:1234/?a=1&b=2#def")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_type.R
\name{xml_type}
\alias{xml_type}
\title{Determine the type of a node.}
\usage{
xml_type(x)
}
\arguments{
\item{x}{A document, node, or node set.}
}
\description{
Determine the type of a node.
}
\examples{
x <- read_xml("<foo> a <b /> <![CDATA[ blah]]></foo>")
xml_type(x)
xml_type(xml_contents(x))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_modify.R
\name{xml_new_document}
\alias{xml_new_document}
\alias{xml_new_root}
\title{Create a new document, possibly with a root node}
\usage{
xml_new_document(version = "1.0", encoding = "UTF-8")

xml_new_root(.value, ..., .copy = inherits(.value, "xml_node"),
  .version = "1.0", .encoding = "UTF-8")
}
\arguments{
\item{version}{The version number of the document.}

\item{encoding}{The character encoding to use in the document. The default
encoding is \sQuote{UTF-8}. Available encodings are specified at
\url{http://xmlsoft.org/html/libxml-encoding.html#xmlCharEncoding}.}

\item{.value}{node to insert.}

\item{...}{If named attributes or namespaces to set on the node, if unnamed
text to assign to the node.}

\item{.copy}{whether to copy the \code{.value} before replacing. If this is \code{FALSE}
then the node will be moved from it's current location.}

\item{.version}{The version number of the document, passed to \code{xml_new_document(version)}.}

\item{.encoding}{The encoding of the document, passed to \code{xml_new_document(encoding)}.}
}
\value{
A \code{xml_document} object.
}
\description{
\code{xml_new_document} creates only a new document without a root node. In
most cases you should instead use \code{xml_new_root}, which creates a new
document and assigns the root node in one step.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{url_escape}
\alias{url_escape}
\alias{url_unescape}
\title{Escape and unescape urls.}
\usage{
url_escape(x, reserved = "")

url_unescape(x)
}
\arguments{
\item{x}{A character vector of urls.}

\item{reserved}{A string containing additional characters to avoid escaping.}
}
\description{
Escape and unescape urls.
}
\examples{
url_escape("a b c")
url_escape("a b c", "")

url_unescape("a\%20b\%2fc")
url_unescape("\%C2\%B5")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_modify.R
\name{xml_set_namespace}
\alias{xml_set_namespace}
\title{Set the node's namespace}
\usage{
xml_set_namespace(.x, prefix = "", uri = "")
}
\arguments{
\item{.x}{a node}

\item{prefix}{The namespace prefix to use}

\item{uri}{The namespace URI to use}
}
\value{
the node (invisibly)
}
\description{
The namespace to be set must be already defined in one of the node's
ancestors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml2-package.R
\docType{package}
\name{xml2-package}
\alias{xml2}
\alias{xml2-package}
\title{xml2: Parse XML}
\description{
Bindings to 'libxml2' for working with XML data using a simple, consistent interface based on 'XPath' expressions. Also supports XML schema validation; for 'XSLT' transformations see the 'xslt' package.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://xml2.r-lib.org}
  \item \url{https://r-lib.r-universe.dev/xml2}
  \item Report bugs at \url{https://github.com/r-lib/xml2/issues}
}

}
\author{
\strong{Maintainer}: Jeroen Ooms \email{jeroenooms@gmail.com}

Authors:
\itemize{
  \item Hadley Wickham
  \item Jim Hester
}

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
  \item R Foundation (Copy of R-project homepage cached as example) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_schema.R
\name{xml_validate}
\alias{xml_validate}
\title{Validate XML schema}
\usage{
xml_validate(x, schema)
}
\arguments{
\item{x}{A document, node, or node set.}

\item{schema}{an XML document containing the schema}
}
\value{
TRUE or FALSE
}
\description{
Validate an XML document against an XML 1.0 schema.
}
\examples{
# Example from https://msdn.microsoft.com/en-us/library/ms256129(v=vs.110).aspx
doc <- read_xml(system.file("extdata/order-doc.xml", package = "xml2"))
schema <- read_xml(system.file("extdata/order-schema.xml", package = "xml2"))
xml_validate(doc, schema)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_url.R
\name{xml_url}
\alias{xml_url}
\title{The URL of an XML document}
\usage{
xml_url(x)
}
\arguments{
\item{x}{A node or document.}
}
\value{
A character vector of length 1. Returns \code{NA} if the name is
  not set.
}
\description{
This is useful for interpreting relative urls with \code{\link{url_relative}}.
}
\examples{
catalog <- read_xml(xml2_example("cd_catalog.xml"))
xml_url(catalog)

x <- read_xml("<foo/>")
xml_url(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_text.R
\name{xml_text}
\alias{xml_text}
\alias{xml_text<-}
\alias{xml_set_text}
\alias{xml_double}
\alias{xml_integer}
\title{Extract or modify the text}
\usage{
xml_text(x, trim = FALSE)

xml_text(x) <- value

xml_set_text(x, value)

xml_double(x)

xml_integer(x)
}
\arguments{
\item{x}{A document, node, or node set.}

\item{trim}{If \code{TRUE} will trim leading and trailing spaces.}

\item{value}{character vector with replacement text.}
}
\value{
A character vector, the same length as x.
}
\description{
\code{xml_text} returns a character vector, \code{xml_double} returns a
numeric vector, \code{xml_integer} returns an integer vector.
}
\examples{
x <- read_xml("<p>This is some text. This is <b>bold!</b></p>")
xml_text(x)
xml_text(xml_children(x))

x <- read_xml("<x>This is some text. <x>This is some nested text.</x></x>")
xml_text(x)
xml_text(xml_find_all(x, "//x"))

x <- read_xml("<p>   Some text    </p>")
xml_text(x, trim = TRUE)

# xml_double() and xml_integer() are useful for extracting numeric
attributes
x <- read_xml("<plot><point x='1' y='2' /><point x='2' y='1' /></plot>")
xml_integer(xml_find_all(x, "//@x"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{url_absolute}
\alias{url_absolute}
\alias{url_relative}
\title{Convert between relative and absolute urls.}
\usage{
url_absolute(x, base)

url_relative(x, base)
}
\arguments{
\item{x}{A character vector of urls relative to that base}

\item{base}{A string giving a base url.}
}
\value{
A character vector of urls
}
\description{
Convert between relative and absolute urls.
}
\examples{
url_absolute(c(".", "..", "/", "/x"), "http://hadley.nz/a/b/c/d")

url_relative("http://hadley.nz/a/c", "http://hadley.nz")
url_relative("http://hadley.nz/a/c", "http://hadley.nz/")
url_relative("http://hadley.nz/a/c", "http://hadley.nz/a/b")
url_relative("http://hadley.nz/a/c", "http://hadley.nz/a/b/")
}
\seealso{
\code{\link{xml_url}} to retrieve the URL associated with a document
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{xml2_example}
\alias{xml2_example}
\title{Get path to a xml2 example}
\usage{
xml2_example(path = NULL)
}
\arguments{
\item{path}{Name of file. If \code{NULL}, the example files will be listed.}
}
\description{
xml2 comes bundled with a number of sample files in its \sQuote{inst/extdata}
directory. This function makes them easy to access.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_modify.R
\name{xml_ns_strip}
\alias{xml_ns_strip}
\title{Strip the default namespaces from a document}
\usage{
xml_ns_strip(x)
}
\arguments{
\item{x}{A document, node, or node set.}
}
\description{
Strip the default namespaces from a document
}
\examples{
x <- read_xml(
 "<foo xmlns = 'http://foo.com'>
   <baz/>
   <bar xmlns = 'http://bar.com'>
     <baz/>
   </bar>
  </foo>")
# Need to specify the default namespaces to find the baz nodes
xml_find_all(x, "//d1:baz")
xml_find_all(x, "//d2:baz")

# After stripping the default namespaces you can find both baz nodes directly
xml_ns_strip(x)
xml_find_all(x, "//baz")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_modify.R
\name{xml_replace}
\alias{xml_replace}
\alias{xml_add_sibling}
\alias{xml_add_child}
\alias{xml_add_parent}
\alias{xml_remove}
\title{Modify a tree by inserting, replacing or removing nodes}
\usage{
xml_replace(.x, .value, ..., .copy = TRUE)

xml_add_sibling(.x, .value, ..., .where = c("after", "before"),
  .copy = TRUE)

xml_add_child(.x, .value, ..., .where = length(xml_children(.x)),
  .copy = TRUE)

xml_add_parent(.x, .value, ...)

xml_remove(.x, free = FALSE)
}
\arguments{
\item{.x}{a document, node or nodeset.}

\item{.value}{node to insert.}

\item{...}{If named attributes or namespaces to set on the node, if unnamed
text to assign to the node.}

\item{.copy}{whether to copy the \code{.value} before replacing. If this is \code{FALSE}
then the node will be moved from it's current location.}

\item{.where}{to add the new node, for \code{xml_add_child} the position
after which to add, use \code{0} for the first child. For
\code{xml_add_sibling} either \sQuote{"before"} or \sQuote{"after"}
indicating if the new node should be before or after \code{.x}.}

\item{free}{When removing the node also free the memory used for that node.
Note if you use this option you cannot use any existing objects pointing to
the node or its children, it is likely to crash R or return garbage.}
}
\description{
\code{xml_add_sibling()} and \code{xml_add_child()} are used to insert a node
as a sibling or a child. \code{xml_add_parent()} adds a new parent in
between the input node and the current parent. \code{xml_replace()}
replaces an existing node with a new node. \code{xml_remove()} removes a
node from the tree.
}
\details{
Care needs to be taken when using \code{xml_remove()},
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{xml_comment}
\alias{xml_comment}
\title{Construct a comment node}
\usage{
xml_comment(content)
}
\arguments{
\item{content}{The comment content}
}
\description{
Construct a comment node
}
\examples{
x <- xml_new_document()
r <- xml_add_child(x, "root")
xml_add_child(r, xml_comment("Hello!"))
as.character(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4.R
\name{xml_document-class}
\alias{xml_document-class}
\alias{xml_missing-class}
\alias{xml_node-class}
\alias{xml_nodeset-class}
\title{Register S4 classes}
\description{
Classes are exported so they can be re-used within S4 classes, see \code{\link[methods:setOldClass]{methods::setOldClass()}}.
\itemize{
\item \code{xml_document}: a complete document.
\item \code{xml_nodeset}: a \emph{set} of nodes within a document.
\item \code{xml_missing}: a missing object, e.g. for an empty result set.
\item \code{xml_node}: a single node in a document.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_namespaces.R
\name{xml_ns}
\alias{xml_ns}
\alias{xml_ns_rename}
\title{XML namespaces.}
\usage{
xml_ns(x)

xml_ns_rename(old, ...)
}
\arguments{
\item{x}{A document, node, or node set.}

\item{old, ...}{An existing xml_namespace object followed by name-value
(old prefix-new prefix) pairs to replace.}
}
\value{
A character vector with class \code{xml_namespace} so the
  default display is a little nicer.
}
\description{
\code{xml_ns} extracts all namespaces from a document, matching each
unique namespace url with the prefix it was first associated with. Default
namespaces are named \code{d1}, \code{d2} etc. Use \code{xml_ns_rename}
to change the prefixes. Once you have a namespace object, you can pass it to
other functions to work with fully qualified names instead of local names.
}
\examples{
x <- read_xml('
 <root>
   <doc1 xmlns = "http://foo.com"><baz /></doc1>
   <doc2 xmlns = "http://bar.com"><baz /></doc2>
 </root>
')
xml_ns(x)

# When there are default namespaces, it's a good idea to rename
# them to give informative names:
ns <- xml_ns_rename(xml_ns(x), d1 = "foo", d2 = "bar")
ns

# Now we can pass ns to other xml function to use fully qualified names
baz <- xml_children(xml_children(x))
xml_name(baz)
xml_name(baz, ns)

xml_find_all(x, "//baz")
xml_find_all(x, "//foo:baz", ns)

str(as_list(x))
str(as_list(x, ns))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_write.R
\name{write_xml}
\alias{write_xml}
\alias{write_xml.xml_document}
\alias{write_html}
\alias{write_html.xml_document}
\title{Write XML or HTML to disk.}
\usage{
write_xml(x, file, ...)

\method{write_xml}{xml_document}(x, file, ..., options = "format",
  encoding = "UTF-8")

write_html(x, file, ...)

\method{write_html}{xml_document}(x, file, ..., options = "format",
  encoding = "UTF-8")
}
\arguments{
\item{x}{A document or node to write to disk. It's not possible to
save nodesets containing more than one node.}

\item{file}{Path to file or connection to write to.}

\item{...}{additional arguments passed to methods.}

\item{options}{default: \sQuote{format}. Zero or more of
\Sexpr[results=rd]{xml2:::describe_options(xml2:::xml_save_options())}}

\item{encoding}{The character encoding to use in the document. The default
encoding is \sQuote{UTF-8}. Available encodings are specified at
\url{http://xmlsoft.org/html/libxml-encoding.html#xmlCharEncoding}.}
}
\description{
This writes out both XML and normalised HTML. The default behavior will
output the same format which was read. If you want to force output pass
\code{option = "as_xml"} or \code{option = "as_html"} respectively.
}
\examples{
h <- read_html("<p>Hi!</p>")

tmp <- tempfile(fileext = ".xml")
write_xml(h, tmp, options = "format")
readLines(tmp)

# write formatted HTML output
write_html(h, tmp, options = "format")
readLines(tmp)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{xml_dtd}
\alias{xml_dtd}
\title{Construct a document type definition}
\usage{
xml_dtd(name = "", external_id = "", system_id = "")
}
\arguments{
\item{name}{The name of the declaration}

\item{external_id}{The external ID of the declaration}

\item{system_id}{The system ID of the declaration}
}
\description{
This is used to create simple document type definitions. If you need to
create a more complicated definition with internal subsets it is recommended
to parse a string directly with \code{read_xml()}.
}
\examples{
r <- xml_new_root(
  xml_dtd("html",
    "-//W3C//DTD XHTML 1.0 Transitional//EN",
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"))

# Use read_xml directly for more complicated DTD
d <- read_xml(
'<!DOCTYPE doc [
<!ELEMENT doc (#PCDATA)>
<!ENTITY foo " test ">
]>
<doc>This is a valid document &foo; !</doc>')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xml_structure.R
\name{xml_structure}
\alias{xml_structure}
\alias{html_structure}
\title{Show the structure of an html/xml document.}
\usage{
xml_structure(x, indent = 2)

html_structure(x, indent = 2)
}
\arguments{
\item{x}{HTML/XML document (or part there of)}

\item{indent}{Number of spaces to ident}
}
\description{
Show the structure of an html/xml document without displaying any of
the values. This is useful if you want to get a high level view of the
way a document is organised. Compared to \code{xml_structure},
\code{html_structure} prints the id and class attributes.
}
\examples{
xml_structure(read_xml("<a><b><c/><c/></b><d/></a>"))

rproj <- read_html(system.file("extdata","r-project.html", package = "xml2"))
xml_structure(rproj)
xml_structure(xml_find_all(rproj, ".//p"))

h <- read_html("<body><p id = 'a'></p><p class = 'c d'></p></body>")
html_structure(h)
}
